/*
 * sisctrl - SiS Display Control Panel
 * for the SiS XFree86/X.org driver
 *
 * GUI part (gtk)
 *
 * (C) 2003-2005 Thomas Winischhofer <thomas@winischhofer.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "sisc.h"

#include "sisctrl_gui.h"

sisctrl_gui		gui;

extern gui_look		gl;

/**************************************************
 *                Clean up and exit               *
 **************************************************/

void
gui_clean_ups(void)
{
   /* Put GUI related clean-ups here */
   if(gui.ab_context) g_object_unref(gui.ab_context);
   gui.ab_context = NULL;
}

void
gui_main_quit(void)
{
   if(gui.window) gtk_main_quit();
}

/**************************************************
 *               Show error pop-ups               *
 **************************************************/

void
gui_showerror(char *message)
{
    GtkWidget *dialog, *label;

    dialog = gtk_dialog_new_with_buttons("Error", GTK_WINDOW(gui.window),
			GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_MODAL,
			GTK_STOCK_OK,
			GTK_RESPONSE_NONE,
			NULL);

    label = gtk_label_new((gchar *)message);

    g_signal_connect_swapped(GTK_OBJECT(dialog), "response",
			G_CALLBACK(gtk_widget_destroy), GTK_OBJECT(dialog));

    gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), label);

    gtk_widget_show_all(dialog);
}

Bool
gui_showerror_pregui(char *message, Bool yesno, Bool warning)
{
   GtkWidget *dialog;
   gint result;

   dialog = gtk_message_dialog_new(NULL,
              		   GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
			   warning ? GTK_MESSAGE_WARNING : GTK_MESSAGE_ERROR,
			   yesno ? GTK_BUTTONS_YES_NO : GTK_BUTTONS_OK,
	      		   (gchar *)message);

   result = gtk_dialog_run(GTK_DIALOG(dialog));

   gtk_widget_destroy(dialog);

   switch(result) {
   case GTK_RESPONSE_YES:
      return TRUE;
   default:
      return FALSE;
   }
}

static GtkWidget *
xpm_label_box(char **xpmdata, gchar *labeltext, GtkWidget *toactivate, int space,
				gchar *tooltip, gchar *tooltip2)
{
   GtkWidget *box, *label = NULL, *image, *align, *event;
   GdkPixmap *icon = NULL;
   GdkBitmap *mask;

   box = gtk_hbox_new(FALSE, 0);
   gtk_container_set_border_width(GTK_CONTAINER(box), 1);

   align = gtk_alignment_new(0.5, 0.5, 0, 0);

#if (GTK_MAJOR_VERSION > 2) || ((GTK_MAJOR_VERSION == 2) && (GTK_MINOR_VERSION >= 4))
   if((tooltip != NULL) && (strlen(tooltip) > 0)) {
      event = gtk_event_box_new();
      gtk_container_set_border_width(GTK_CONTAINER(event), 0);
      gtk_container_add(GTK_CONTAINER(event), box);
      gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), event, tooltip, tooltip2);
      gtk_event_box_set_visible_window(GTK_EVENT_BOX(event), FALSE);
      gtk_widget_show(event);
      gtk_container_add(GTK_CONTAINER(align), event);
   } else {
#endif
      gtk_container_add(GTK_CONTAINER(align), box);
#if (GTK_MAJOR_VERSION > 2) || ((GTK_MAJOR_VERSION == 2) && (GTK_MINOR_VERSION >= 4))
   }
#endif

   icon = gdk_pixmap_create_from_xpm_d(gui.window->window, &mask, NULL, xpmdata);
   image = gtk_image_new_from_pixmap(icon, mask);

   gtk_box_pack_start(GTK_BOX(box), image, FALSE, FALSE, space);

   if(labeltext) {
      label = gtk_label_new_with_mnemonic(labeltext);
      gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);
      gtk_box_pack_start(GTK_BOX(box), label, FALSE, FALSE, space);
      gtk_widget_show(label);
   }

   gtk_widget_show(box);
   gtk_widget_show(image);

   if((label) && (toactivate)) {
      gtk_label_set_mnemonic_widget(GTK_LABEL(label), toactivate);
   }

   return align;
}

/**************************************************
 *             Xinerama screen chooser            *
 **************************************************/


Bool
gui_xinerama_chooser(int maxscreens, int *newscreen)
{
   char *format = "Screen %d [SiS %s%s]";
   char *crtnoformat = ", CRT%d";
   int i, crtnum;
   char *chip;
   char buffer[64];
   char buffer2[16];
   Bool myret;
   gint ret;
   GtkWidget *dialog, *frame, *vbox, *label, *menu, *omenu, *menuitem;

   frame = gtk_frame_new("Xinerama");
   gtk_container_set_border_width(GTK_CONTAINER(frame), 10);

   vbox = gtk_vbox_new(FALSE, 8);
   gtk_container_add(GTK_CONTAINER(frame), vbox);
   gtk_container_set_border_width(GTK_CONTAINER(vbox), 10);

   label = gtk_label_new("Xinerama detected.\n\nPlease choose the screen to configure:");
   gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);

   omenu = gtk_option_menu_new();
   menu = gtk_menu_new();

   for(i = 0; i < maxscreens; i++) {
      if(sisctrl_xinerama_query_screen(i, &chip, &crtnum)) {
         buffer2[0] = 0;
	 if(crtnum != -1) sprintf(buffer2, crtnoformat, crtnum);
         sprintf(buffer, format, i, chip, buffer2);
	 menuitem = gtk_menu_item_new_with_label(buffer);
         gtk_menu_shell_append(GTK_MENU_SHELL(menu), menuitem);
      }
   }

   gtk_option_menu_set_menu(GTK_OPTION_MENU(omenu), menu);

   dialog = gtk_dialog_new_with_buttons("SiSCtrl",
				NULL,
				GTK_DIALOG_MODAL,
				GTK_STOCK_OK,
				GTK_RESPONSE_OK,
				GTK_STOCK_CANCEL,
				GTK_RESPONSE_CANCEL,
				NULL);

   gtk_container_add(GTK_CONTAINER(vbox), label);
   gtk_container_add(GTK_CONTAINER(vbox), omenu);

   gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), frame);

   gtk_widget_show_all(dialog);

   ret = gtk_dialog_run(GTK_DIALOG(dialog));

   switch(ret) {
   case GTK_RESPONSE_NONE:
   case GTK_RESPONSE_CANCEL:
   case GTK_RESPONSE_DELETE_EVENT:
      myret = FALSE;
      break;
   default:
      *newscreen = gtk_option_menu_get_history(GTK_OPTION_MENU(omenu));
      myret = TRUE;
   }

   gtk_widget_destroy(dialog);

   return myret;
}

/**************************************************
 *                Displaymode page                *
 **************************************************/

int
gui_dm_get_current_selected_mode(void)
{
   int i;

   if(!gtk_tree_selection_get_selected(gui.dm_select, NULL, NULL))
      return -1;

   for(i = 0; i < SISCTRL_MAXMODES; i++) {
       if(gtk_tree_selection_iter_is_selected(gui.dm_select, &gui.dm_iter[i])) {
	  return i;
       }
   }

   return -1;
}

#ifdef HAVE_RANDR
int
gui_dm_get_current_randr_index(void)
{
   int index = -1;

   if(gl.do_randr && gui.dm_omenu) {
      index = gtk_option_menu_get_history(GTK_OPTION_MENU(gui.dm_omenu));
   }

   return index;
}
#endif

void
gui_dm_create_gfx(void)
{
   if(!gui.pixbuf_lcd)
      gui.pixbuf_lcd = gdk_pixbuf_new_from_xpm_data((const char **)icon_lcd);

   if(!gui.pixbuf_lcda)
      gui.pixbuf_lcda = gdk_pixbuf_new_from_xpm_data((const char **)icon_lcda);

   if(!gui.pixbuf_tvpal)
      gui.pixbuf_tvpal = gdk_pixbuf_new_from_xpm_data((const char **)icon_tvpal);

   if(!gui.pixbuf_tvpalm)
      gui.pixbuf_tvpalm = gdk_pixbuf_new_from_xpm_data((const char **)icon_tvpalm);

   if(!gui.pixbuf_tvpaln)
      gui.pixbuf_tvpaln = gdk_pixbuf_new_from_xpm_data((const char **)icon_tvpaln);

   if(!gui.pixbuf_tvntsc)
      gui.pixbuf_tvntsc = gdk_pixbuf_new_from_xpm_data((const char **)icon_tvntsc);

   if(!gui.pixbuf_tvhivi)
      gui.pixbuf_tvhivi = gdk_pixbuf_new_from_xpm_data((const char **)icon_tvhivi);

   if(!gui.pixbuf_tv525i)
      gui.pixbuf_tv525i = gdk_pixbuf_new_from_xpm_data((const char **)icon_tv525i);

   if(!gui.pixbuf_tv525p)
      gui.pixbuf_tv525p = gdk_pixbuf_new_from_xpm_data((const char **)icon_tv525p);

   if(!gui.pixbuf_tv750p)
      gui.pixbuf_tv750p = gdk_pixbuf_new_from_xpm_data((const char **)icon_tv750p);

   if(!gui.pixbuf_tv1080i)
      gui.pixbuf_tv1080i = gdk_pixbuf_new_from_xpm_data((const char **)icon_tv1080i);

   if(!gui.pixbuf_vga2)
      gui.pixbuf_vga2 = gdk_pixbuf_new_from_xpm_data((const char **)icon_vga2);
}

void
gui_dm_set_mode_list_attributes(int listindex, unsigned short field)
{

   gtk_list_store_set(gui.dm_liststore, &gui.dm_iter[listindex], 1,
		(field & SiS_CRT2_CRT1LCDA) ? gui.pixbuf_lcda : NULL, -1);

   gtk_list_store_set(gui.dm_liststore, &gui.dm_iter[listindex], 2,
		(field & SiS_CRT2_LCD) ? gui.pixbuf_lcd : NULL, -1);

   gtk_list_store_set(gui.dm_liststore, &gui.dm_iter[listindex], 3,
		(field & SiS_CRT2_TVPAL) ? gui.pixbuf_tvpal : NULL, -1);

   gtk_list_store_set(gui.dm_liststore, &gui.dm_iter[listindex], 4,
		(field & SiS_CRT2_TVNTSC) ? gui.pixbuf_tvntsc : NULL, -1);

   gtk_list_store_set(gui.dm_liststore, &gui.dm_iter[listindex], 5,
		(field & SiS_CRT2_TVPALM) ? gui.pixbuf_tvpalm : NULL, -1);

   gtk_list_store_set(gui.dm_liststore, &gui.dm_iter[listindex], 6,
		(field & SiS_CRT2_TVPALN) ? gui.pixbuf_tvpaln : NULL, -1);

   gtk_list_store_set(gui.dm_liststore, &gui.dm_iter[listindex], 7,
		(field & SiS_CRT2_HIVISION) ? gui.pixbuf_tvhivi : NULL, -1);

   gtk_list_store_set(gui.dm_liststore, &gui.dm_iter[listindex], 8,
   		(field & SiS_CRT2_YPBPR525I) ? gui.pixbuf_tv525i : NULL, -1);

   gtk_list_store_set(gui.dm_liststore, &gui.dm_iter[listindex], 9,
   		(field & SiS_CRT2_YPBPR525P) ? gui.pixbuf_tv525p : NULL, -1);

   gtk_list_store_set(gui.dm_liststore, &gui.dm_iter[listindex], 10,
   		(field & SiS_CRT2_YPBPR750P) ? gui.pixbuf_tv750p : NULL, -1);

   gtk_list_store_set(gui.dm_liststore, &gui.dm_iter[listindex], 11,
   		(field & SiS_CRT2_YPBPR1080I) ? gui.pixbuf_tv1080i : NULL, -1);

   gtk_list_store_set(gui.dm_liststore, &gui.dm_iter[listindex], 12,
   		(field & SiS_CRT2_VGA2) ? gui.pixbuf_vga2 : NULL, -1);
}

void
gui_dm_set_current_entry(int index)
{
   gtk_tree_selection_unselect_all(gui.dm_select);
   gtk_tree_selection_select_iter(gui.dm_select, &gui.dm_iter[index]);
}

#ifdef HAVE_RANDR
void
gui_dm_update_leaveunchanged(char *newtext)
{
   GtkWidget *label;
   gint  current_menu;

   /* update the "leave unchanged at xxx" text. Due to GTK misbehavior,
      we need to change the menu index before changing the text if the
      "leave unchanged" item is currently selected
    */

   if(!gui.dm_omenu) return;

   current_menu = gtk_option_menu_get_history(GTK_OPTION_MENU(gui.dm_omenu));
   if(current_menu == 0) {
      gtk_option_menu_set_history(GTK_OPTION_MENU(gui.dm_omenu), 1);
   }
   label = gtk_bin_get_child(GTK_BIN(gui.dm_menuitem[0]));
   gtk_label_set_text(GTK_LABEL(label), (gchar *)newtext);
   if(current_menu == 0) {
      gtk_option_menu_set_history(GTK_OPTION_MENU(gui.dm_omenu), 0);
   }
}
#endif

/* Signal handler for displaymode list "selection changed" */

static void
dm_selection_changed(GtkTreeSelection *selection, gpointer data)
{
   char leftstring[64];
   char rightstring[64];
   char *crt1format = "CRT1 <span weight=\"bold\">%dx%d</span> %dHz%s";
   char *crt2format = "CRT2 <span weight=\"bold\">%dx%d</span> %dHz%s";
   int i, index = -1;
   GtkWidget *crt1label = gui.dm_mmdl_label;
   GtkWidget *crt2label = gui.dm_mmdr_label;

   if(!gl.do_mergedfb_details) return;

   index = gui_dm_get_current_selected_mode();

   if(index == -1) return;

   sisctrl_dm_selection_changed(index, leftstring, rightstring,
   					crt1format, crt2format);

   gtk_label_set_markup(GTK_LABEL(gui.dm_mmdl_label), (const gchar *)leftstring);
   gtk_label_set_markup(GTK_LABEL(gui.dm_mmdr_label), (const gchar *)rightstring);
}

/* Build page */

static void
build_display_mode_page(void)
{
    int i, rows = 6, randrrow = 5;

    gui.dm_frame = gtk_frame_new((gl.do_randr) ? "Display mode and desktop size" : "Display mode");
    gtk_container_set_border_width(GTK_CONTAINER(gui.dm_frame), 8);
    gtk_widget_show(gui.dm_frame);

    if(gl.do_randr || gl.do_mergedfb_details) {
       rows = 6;
       randrrow = 5;
       if(gl.do_randr && gl.do_mergedfb_details) {
          rows = 7;
	  randrrow = 6;
       }
       gui.dm_table = gtk_table_new(rows, 5, TRUE);
       gtk_table_set_row_spacings(GTK_TABLE(gui.dm_table), 0);
       gtk_table_set_col_spacings(GTK_TABLE(gui.dm_table), 0);
       gtk_container_set_border_width(GTK_CONTAINER(gui.dm_table), 10);
       gtk_container_add(GTK_CONTAINER(gui.dm_frame), gui.dm_table);
       gtk_widget_show(gui.dm_table);
    }

#ifdef HAVE_RANDR
    if(gl.do_randr) {

       gui.dm_omenu = gtk_option_menu_new();
       gui.dm_menu = gtk_menu_new();

       gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.dm_omenu,
       		"Select desktop size",
		"Select to resize your desktop to a certain size. This requires the RandR extension "
		"and a RandR-capable window manager, otherwise you end up with inaccessible menus "
		"or windows.");

       gui.dm_menuitem[0] = gtk_menu_item_new_with_label(gl.leaveunchanged_text);
       gtk_menu_shell_append(GTK_MENU_SHELL(gui.dm_menu), gui.dm_menuitem[0]);
       gtk_widget_show(gui.dm_menuitem[0]);
       gui.dm_menuitem[1] = gtk_menu_item_new_with_label("Resize to display mode");
       gtk_menu_shell_append(GTK_MENU_SHELL(gui.dm_menu), gui.dm_menuitem[1]);
       gtk_widget_show(gui.dm_menuitem[1]);

       for(i = 0; i < gl.numrandrsizes; i++) {
          gui.dm_menuitem[i+2] = gtk_menu_item_new_with_label(gl.sizetextptr[i]);
          gtk_menu_shell_append(GTK_MENU_SHELL(gui.dm_menu), gui.dm_menuitem[i+2]);
          gtk_widget_show(gui.dm_menuitem[i+2]);
       }

       gtk_option_menu_set_menu(GTK_OPTION_MENU(gui.dm_omenu), gui.dm_menu);
       /* gtk_table_attach_defaults(GTK_TABLE(gui.dm_table), gui.dm_omenu, 1, 5, randrrow, randrrow+1); */
       gtk_table_attach(GTK_TABLE(gui.dm_table), gui.dm_omenu, 1, 5, randrrow, randrrow+1, GTK_EXPAND|GTK_FILL, 0, 0, 0);

       gui.dm_size_label = gtk_label_new("Desktop size");
       gtk_table_attach_defaults(GTK_TABLE(gui.dm_table), gui.dm_size_label, 0, 1, randrrow, randrrow+1);
       gtk_widget_show(gui.dm_size_label);

       sisctrl_dm_update_leaveunchanged();

       gtk_widget_show(gui.dm_omenu);
    }
#endif

    gui.dm_liststore = gtk_list_store_new(14, G_TYPE_STRING,   GDK_TYPE_PIXBUF, GDK_TYPE_PIXBUF,
					     GDK_TYPE_PIXBUF, GDK_TYPE_PIXBUF, GDK_TYPE_PIXBUF,
					     GDK_TYPE_PIXBUF, GDK_TYPE_PIXBUF, GDK_TYPE_PIXBUF,
					     GDK_TYPE_PIXBUF, GDK_TYPE_PIXBUF, GDK_TYPE_PIXBUF,
					     GDK_TYPE_PIXBUF, G_TYPE_STRING);

    for(i = 0; i < SISCTRL_MAXMODES; i++) {
       if(gl.modetextptr[i]) {
          gtk_list_store_append(gui.dm_liststore, &gui.dm_iter[i]);
          gtk_list_store_set(gui.dm_liststore, &gui.dm_iter[i], 0, gl.modetextptr[i], -1);
       }
    }

    sisctrl_set_crt2_types_for_modelist();

    gui.dm_list = gtk_tree_view_new_with_model(GTK_TREE_MODEL(gui.dm_liststore));

    gui.dm_renderer = gtk_cell_renderer_text_new();
    gui.dm_column = gtk_tree_view_column_new_with_attributes("Display mode", gui.dm_renderer, "text", 0, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(gui.dm_list), gui.dm_column);

    gui.dm_renderer = gtk_cell_renderer_pixbuf_new();
    gui.dm_column = gtk_tree_view_column_new_with_attributes("LCDA", gui.dm_renderer, "pixbuf", 1, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(gui.dm_list), gui.dm_column);

    gui.dm_renderer = gtk_cell_renderer_pixbuf_new();
    gui.dm_column = gtk_tree_view_column_new_with_attributes("LCD", gui.dm_renderer, "pixbuf", 2, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(gui.dm_list), gui.dm_column);

    gui.dm_renderer = gtk_cell_renderer_pixbuf_new();
    gui.dm_column = gtk_tree_view_column_new_with_attributes("TVPAL", gui.dm_renderer, "pixbuf", 3, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(gui.dm_list), gui.dm_column);
    gui.dm_renderer = gtk_cell_renderer_pixbuf_new();
    gui.dm_column = gtk_tree_view_column_new_with_attributes("TVNTSC", gui.dm_renderer, "pixbuf", 4, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(gui.dm_list), gui.dm_column);
    gui.dm_renderer = gtk_cell_renderer_pixbuf_new();
    gui.dm_column = gtk_tree_view_column_new_with_attributes("TVPALM", gui.dm_renderer, "pixbuf", 5, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(gui.dm_list), gui.dm_column);
    gui.dm_renderer = gtk_cell_renderer_pixbuf_new();
    gui.dm_column = gtk_tree_view_column_new_with_attributes("TVPALN", gui.dm_renderer, "pixbuf", 6, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(gui.dm_list), gui.dm_column);
    gui.dm_renderer = gtk_cell_renderer_pixbuf_new();
    gui.dm_column = gtk_tree_view_column_new_with_attributes("TVHIVISION", gui.dm_renderer, "pixbuf", 7, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(gui.dm_list), gui.dm_column);
    gui.dm_renderer = gtk_cell_renderer_pixbuf_new();
    gui.dm_column = gtk_tree_view_column_new_with_attributes("TVY525I", gui.dm_renderer, "pixbuf", 8, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(gui.dm_list), gui.dm_column);
    gui.dm_renderer = gtk_cell_renderer_pixbuf_new();
    gui.dm_column = gtk_tree_view_column_new_with_attributes("TVY525P", gui.dm_renderer, "pixbuf", 9, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(gui.dm_list), gui.dm_column);
    gui.dm_renderer = gtk_cell_renderer_pixbuf_new();
    gui.dm_column = gtk_tree_view_column_new_with_attributes("TVY750P", gui.dm_renderer, "pixbuf", 10, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(gui.dm_list), gui.dm_column);
    gui.dm_renderer = gtk_cell_renderer_pixbuf_new();
    gui.dm_column = gtk_tree_view_column_new_with_attributes("TVY1080I", gui.dm_renderer, "pixbuf", 11, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(gui.dm_list), gui.dm_column);

    gui.dm_renderer = gtk_cell_renderer_pixbuf_new();
    gui.dm_column = gtk_tree_view_column_new_with_attributes("VGA2", gui.dm_renderer, "pixbuf", 12, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(gui.dm_list), gui.dm_column);

    /* This last column is only for keeping gtk to make wrong alignment assumptions */
    gui.dm_renderer = gtk_cell_renderer_text_new();
    gui.dm_column = gtk_tree_view_column_new_with_attributes("pseudo", gui.dm_renderer, "text", 13, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(gui.dm_list), gui.dm_column);

    gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(gui.dm_list), FALSE);

    gui.dm_select = gtk_tree_view_get_selection(GTK_TREE_VIEW(gui.dm_list));
    gtk_tree_selection_set_mode(gui.dm_select, GTK_SELECTION_SINGLE);

    if(gl.do_mergedfb_details) {
       gui.dm_mmdt_label = gtk_label_new("Metamode details:");
       gtk_table_attach_defaults(GTK_TABLE(gui.dm_table), gui.dm_mmdt_label, 0, 1, 5, 6);
       gtk_widget_show(gui.dm_mmdt_label);
       gui.dm_mmdl_label = gtk_label_new("n/a");
       gtk_table_attach_defaults(GTK_TABLE(gui.dm_table), gui.dm_mmdl_label, 1, 3, 5, 6);
       gtk_widget_show(gui.dm_mmdl_label);
       gui.dm_mmdr_label = gtk_label_new("n/a");
       gtk_table_attach_defaults(GTK_TABLE(gui.dm_table), gui.dm_mmdr_label, 3, 5, 5, 6);
       gtk_widget_show(gui.dm_mmdr_label);
       g_signal_connect(G_OBJECT(gui.dm_select), "changed", G_CALLBACK(dm_selection_changed), NULL);
    }

    gui_dm_set_current_entry(gl.startmodeindex);

    gui.dm_list_window = gtk_scrolled_window_new(NULL, NULL);
    if(gl.do_randr || gl.do_mergedfb_details) {
       gtk_container_set_border_width(GTK_CONTAINER(gui.dm_list_window), 0);
    } else {
       gtk_container_set_border_width(GTK_CONTAINER(gui.dm_list_window), 10);
    }
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(gui.dm_list_window), GTK_POLICY_AUTOMATIC,
									   GTK_POLICY_AUTOMATIC);

    gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(gui.dm_list_window), gui.dm_list);

    if(gl.do_randr || gl.do_mergedfb_details) {
       gtk_table_attach_defaults(GTK_TABLE(gui.dm_table), gui.dm_list_window, 0, 5, 0, 5);
    } else {
       gtk_container_add(GTK_CONTAINER(gui.dm_frame), gui.dm_list_window);
    }

    gtk_widget_show(gui.dm_list);
    gtk_widget_show(gui.dm_list_window);

#if 0    /* gtk bug: this tooltip shows up anywhere but where it should */
    gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.dm_list,
       		"Select display mode (display resolution)", "");
#endif

    gui.dm_label = xpm_label_box(icon_mode, "Mode", NULL, 1,
        (gl.do_randr) ? "Select display mode and desktop size" :
			 "Select display mode", "");

    gtk_notebook_append_page(GTK_NOTEBOOK(gui.notebook), gui.dm_frame, gui.dm_label);
}

/**************************************************
 *                   CRT1 page                    *
 **************************************************/

/* CRT1 type menu and graphics */

void
gui_crt1_show_variant(int variant)
{
   if(!gui.c1_label_onoff) return;

   if(variant) {
      gtk_label_set_text(GTK_LABEL(gui.c1_label_onoff), "Device type");
      gtk_label_set_text(GTK_LABEL(gui.c1_menuitemon_label), "VGA (analog output, primary plug)");
      gtk_widget_show(gui.c1_menuitemlcda);
   } else {
      gtk_label_set_text(GTK_LABEL(gui.c1_label_onoff), "Switch output");
      gtk_label_set_text(GTK_LABEL(gui.c1_menuitemon_label), "ON");
      gtk_widget_hide(gui.c1_menuitemlcda);
   }
}

void
gui_crt1_set_type_menu_index(int index)
{
   /* C1_MI_* are identical with our real menu indices */
   gtk_option_menu_set_history(GTK_OPTION_MENU(gui.c1_omenu), (gint)index);
}

int
gui_crt1_get_type_menu_index(void)
{
   /* C1_MI_* are identical with our real menu indices */
   return(int)(gtk_option_menu_get_history(GTK_OPTION_MENU(gui.c1_omenu)));
}

void
gui_crt1_select_gfx(int gfxnum)
{
   gtk_widget_hide(gui.c1_imageoff);
   gtk_widget_hide(gui.c1_imageon);
   gtk_widget_hide(gui.c1_imagelcda);

   switch(gfxnum) {
   case C1_GFX_OFF:
      gtk_widget_show(gui.c1_imageoff);
      break;
   case C1_GFX_VGA:
      gtk_widget_show(gui.c1_imageon);
      break;
   case C1_GFX_LCDA:
      gtk_widget_show(gui.c1_imagelcda);
   }
}

void
gui_crt1_update_gfx(void)
{
   int c1_index = gui_crt1_get_type_menu_index();
   int c1_gfx;

   switch(c1_index) {
   case C1_MI_OFF:
       c1_gfx = C1_GFX_OFF;
       break;
   case C1_MI_VGA:
       c1_gfx = C1_GFX_VGA;
       break;
   case C1_MI_LCDA:
       if(gl.supportlcda) {
          c1_gfx = C1_GFX_LCDA;
       } else {
          c1_gfx = C1_GFX_VGA;
       }
       break;
   default:
       c1_gfx = C1_GFX_OFF;
   }

   gui_crt1_select_gfx(c1_gfx);
}

/* LCD scaling menu */

void
gui_crt1_set_scaling_menu_index(int index)
{
   /* CO_SC_* are identical with our real menu indices */
   if(gui.c1_omenuls) {
      gtk_option_menu_set_history(GTK_OPTION_MENU(gui.c1_omenuls), (gint)index);
   }
}

int
gui_crt1_get_scaling_menu_index(int retdefault)
{
   /* CO_SC_* are identical with our real menu indices */
   if(!gui.c1_omenuls) return retdefault;
   return(int)(gtk_option_menu_get_history(GTK_OPTION_MENU(gui.c1_omenuls)));
}

void
gui_crt1_show_scaling_menu_and_label(void)
{
   gtk_widget_show(gui.c1_omenuls);
   gtk_widget_show(gui.c1_label_lcds);
}

void
gui_crt1_hide_scaling_menu_and_label(void)
{
   gtk_widget_hide(gui.c1_omenuls);
   gtk_widget_hide(gui.c1_label_lcds);
}

/* LCD centering menu */

void
gui_crt1_set_centering_menu_index(int index)
{
   /* CO_CE_* are identical with our real menu indices */
   if(gui.c1_omenulcc) {
      gtk_option_menu_set_history(GTK_OPTION_MENU(gui.c1_omenulcc), (gint)index);
   }
}

int
gui_crt1_get_centering_menu_index(int retdefault)
{
   /* CO_CE_* are identical with our real menu indices */
   if(!gui.c1_omenulcc) return retdefault;
   return(int)(gtk_option_menu_get_history(GTK_OPTION_MENU(gui.c1_omenulcc)));
}

void
gui_crt1_show_center_menu(void)
{
   gtk_widget_show(gui.c1_omenulcc);
}

void
gui_crt1_hide_center_menu(void)
{
   gtk_widget_hide(gui.c1_omenulcc);
}

/* Signal handlers */

static void
c1_menu_changed(GtkWidget *widget, gpointer data)
{
    gui_crt1_update_gfx();
    gui_show_lcd_menus(1);
}

static void
c1_omenuls_changed(GtkWidget *widget, gpointer data)
{
    int index = gui_crt1_get_scaling_menu_index(CO_MI_SC_DEFAULT);

    if(!gl.supportcenter) return;

    if(index == CO_MI_SC_ON) {
       gui_crt1_hide_center_menu();
    } else {
       gui_crt1_show_center_menu();
    }
}

/* Build page */

static void
build_crt1_page(void)
{
    int i, j, k;

    gui.c1_frame = gtk_frame_new("CRT1 settings");
    gtk_container_set_border_width(GTK_CONTAINER(gui.c1_frame), 8);
    gtk_widget_show(gui.c1_frame);

    i = 4; j = 6; k = 15;
    if(gl.crt1_page_type == CRT1_PAGE_TYPE_NORMAL) {
       if(gl.supportlcda && gl.supportscale) {
          i = 5; j = 4; k = 8;
       }
    }
    gui.c1_table = gtk_table_new(i, 5, TRUE);
    gtk_table_set_row_spacings(GTK_TABLE(gui.c1_table), j);
    gtk_table_set_col_spacings(GTK_TABLE(gui.c1_table), 10);

    gtk_container_set_border_width(GTK_CONTAINER(gui.c1_table), k);

    gtk_container_add(GTK_CONTAINER(gui.c1_frame), gui.c1_table);
    gtk_widget_show(gui.c1_table);

    switch(gl.crt1_page_type) {

    case CRT1_PAGE_TYPE_NORMAL:
       gui.c1_iconon = gdk_pixmap_create_from_xpm_d(gui.window->window, &gui.c1_maskon, NULL, crt1on_xpm);
       gui.c1_imageon = gtk_image_new_from_pixmap(gui.c1_iconon, gui.c1_maskon);
       gui.c1_iconoff = gdk_pixmap_create_from_xpm_d(gui.window->window, &gui.c1_maskoff, NULL, crt1off_xpm);
       gui.c1_imageoff = gtk_image_new_from_pixmap(gui.c1_iconoff, gui.c1_maskoff);
       gui.c1_iconlcda = gdk_pixmap_create_from_xpm_d(gui.window->window, &gui.c1_masklcda, NULL, crt1lcda_xpm);
       gui.c1_imagelcda = gtk_image_new_from_pixmap(gui.c1_iconlcda, gui.c1_masklcda);
       gtk_table_attach_defaults(GTK_TABLE(gui.c1_table), gui.c1_imageon, 0, 5, 0, 3);
       gtk_table_attach_defaults(GTK_TABLE(gui.c1_table), gui.c1_imageoff, 0, 5, 0, 3);
       gtk_table_attach_defaults(GTK_TABLE(gui.c1_table), gui.c1_imagelcda, 0, 5, 0, 3);
       gtk_widget_hide(gui.c1_imageon);
       gtk_widget_hide(gui.c1_imageoff);
       gtk_widget_hide(gui.c1_imagelcda);

       gui.c1_label_onoff = gtk_label_new("Switch output");

       gtk_table_attach_defaults(GTK_TABLE(gui.c1_table), gui.c1_label_onoff, 0, 1, 3, 4);
       gtk_widget_show(gui.c1_label_onoff);

       gui.c1_omenu = gtk_option_menu_new();

       gui.c1_menu = gtk_menu_new();

       gui.c1_menuitemoff  = gtk_menu_item_new_with_label("OFF");
       gui.c1_menuitemon   = gtk_menu_item_new_with_label("ON");
       gui.c1_menuitemlcda = gtk_menu_item_new_with_label(gl.supportdvi ? "LCD/DVI-D (digital output)" :
	  								 "LCD");
       gui.c1_menuitemon_label = gtk_bin_get_child(GTK_BIN(gui.c1_menuitemon));

       gtk_menu_shell_append(GTK_MENU_SHELL(gui.c1_menu), gui.c1_menuitemon);
       gtk_menu_shell_append(GTK_MENU_SHELL(gui.c1_menu), gui.c1_menuitemlcda);
       gtk_menu_shell_append(GTK_MENU_SHELL(gui.c1_menu), gui.c1_menuitemoff);

       gtk_widget_show(gui.c1_menuitemon);
       gtk_widget_show(gui.c1_menuitemoff);

       gtk_option_menu_set_menu(GTK_OPTION_MENU(gui.c1_omenu), gui.c1_menu);

       gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.c1_omenu,
		gl.supportlcda ? "Select first output (CRT1) device type" :
				 "Enable/disable first output (CRT1)", "");

       if(gl.supportlcda && gl.supportscale) {

	  gui.c1_label_lcds = gtk_label_new("LCD scaling");
          gtk_table_attach_defaults(GTK_TABLE(gui.c1_table), gui.c1_label_lcds, 0, 1, 4, 5);

          gui.c1_omenuls = gtk_option_menu_new();
          gui.c1_menuls = gtk_menu_new();
          gui.c1_menuitemlsd = gtk_menu_item_new_with_label("Default");
          gtk_menu_shell_append(GTK_MENU_SHELL(gui.c1_menuls), gui.c1_menuitemlsd);
          gtk_widget_show(gui.c1_menuitemlsd);
	  gui.c1_menuitemlso = gtk_menu_item_new_with_label("On");
          gtk_menu_shell_append(GTK_MENU_SHELL(gui.c1_menuls), gui.c1_menuitemlso);
          gtk_widget_show(gui.c1_menuitemlso);
	  gui.c1_menuitemlsf = gtk_menu_item_new_with_label("Off");
          gtk_menu_shell_append(GTK_MENU_SHELL(gui.c1_menuls), gui.c1_menuitemlsf);
          gtk_widget_show(gui.c1_menuitemlsf);
	  gtk_option_menu_set_menu(GTK_OPTION_MENU(gui.c1_omenuls), gui.c1_menuls);

	  gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.c1_omenuls,
		"Scale low resolution modes to panel's native resolution", "");

	  if(gl.supportcenter) {

	     /* gtk_table_attach_defaults(GTK_TABLE(gui.c1_table), gui.c1_omenuls, 1, 3, 4, 5); */
	     gtk_table_attach(GTK_TABLE(gui.c1_table), gui.c1_omenuls, 1, 3, 4, 5, GTK_EXPAND|GTK_FILL, 0, 0, 0);
	     gui.c1_omenulcc = gtk_option_menu_new();
             gui.c1_menulcc = gtk_menu_new();
             gui.c1_menuitemlccd = gtk_menu_item_new_with_label("Default");
             gtk_menu_shell_append(GTK_MENU_SHELL(gui.c1_menulcc), gui.c1_menuitemlccd);
             gtk_widget_show(gui.c1_menuitemlccd);
	     gui.c1_menuitemlccc = gtk_menu_item_new_with_label("Don't center");
             gtk_menu_shell_append(GTK_MENU_SHELL(gui.c1_menulcc), gui.c1_menuitemlccc);
             gtk_widget_show(gui.c1_menuitemlccc);
	     gui.c1_menuitemlccp = gtk_menu_item_new_with_label("Center");
             gtk_menu_shell_append(GTK_MENU_SHELL(gui.c1_menulcc), gui.c1_menuitemlccp);
             gtk_widget_show(gui.c1_menuitemlccp);
	     gtk_option_menu_set_menu(GTK_OPTION_MENU(gui.c1_omenulcc), gui.c1_menulcc);
	     /* gtk_table_attach_defaults(GTK_TABLE(gui.c1_table), gui.c1_omenulcc, 3, 5, 4, 5); */
	     gtk_table_attach(GTK_TABLE(gui.c1_table), gui.c1_omenulcc, 3, 5, 4, 5, GTK_EXPAND|GTK_FILL, 0, 0, 0);

	     gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.c1_omenulcc,
		   "Center unscaled output or pass native mode timing to the panel", "");
	  } else {

	     /* gtk_table_attach_defaults(GTK_TABLE(gui.c1_table), gui.c1_omenuls, 1, 5, 4, 5); */
	     gtk_table_attach(GTK_TABLE(gui.c1_table), gui.c1_omenuls, 1, 5, 4, 5, GTK_EXPAND|GTK_FILL, 0, 0, 0);

	  }
       }

       gui_crt1_show_variant(gl.supportlcda ? 1 : 0);

       gui_crt1_set_type_menu_index(sisctrl_get_current_crt1type_index());

       gui_crt1_update_gfx();

       /* gtk_table_attach_defaults(GTK_TABLE(gui.c1_table), gui.c1_omenu, 1, 5, 3, 4); */
       gtk_table_attach(GTK_TABLE(gui.c1_table), gui.c1_omenu, 1, 5, 3, 4, GTK_EXPAND|GTK_FILL, 0, 0, 0);

       if(gl.supportlcda && gl.supportscale) {

	  gui_crt1_set_scaling_menu_index(sisctrl_get_lcd_scaling_history());

	  if(gl.supportcenter) {
	     gui_crt1_set_centering_menu_index(sisctrl_get_lcd_center_history());
	  }

	  gui_show_lcd_menus(1);

       }

       g_signal_connect(G_OBJECT(gui.c1_menuitemon), "activate", G_CALLBACK(c1_menu_changed), NULL);
       g_signal_connect(G_OBJECT(gui.c1_menuitemoff), "activate", G_CALLBACK(c1_menu_changed), NULL);
       g_signal_connect(G_OBJECT(gui.c1_menuitemlcda), "activate", G_CALLBACK(c1_menu_changed), NULL);

       if(gl.supportlcda && gl.supportscale && gl.supportcenter)  {
	  g_signal_connect(G_OBJECT(gui.c1_menuitemlsd), "activate", G_CALLBACK(c1_omenuls_changed), NULL);
	  g_signal_connect(G_OBJECT(gui.c1_menuitemlso), "activate", G_CALLBACK(c1_omenuls_changed), NULL);
	  g_signal_connect(G_OBJECT(gui.c1_menuitemlsf), "activate", G_CALLBACK(c1_omenuls_changed), NULL);
       }

       gtk_widget_show(gui.c1_omenu);
       break;

    case CRT1_PAGE_TYPE_NA_NOVB:
       gui.c1_label_nodh = gtk_label_new("No videobridge present");
       gtk_widget_show(gui.c1_label_nodh);
       gtk_table_attach_defaults(GTK_TABLE(gui.c1_table), gui.c1_label_nodh, 0, 5, 0, 4);
       break;

    case CRT1_PAGE_TYPE_NA_DHM:
       gui.c1_label_nodh = gtk_label_new("Not supported in Dual Head mode");
       gtk_widget_show(gui.c1_label_nodh);
       gtk_table_attach_defaults(GTK_TABLE(gui.c1_table), gui.c1_label_nodh, 0, 5, 0, 4);
       break;
    }

    gui.c1_label = xpm_label_box(icon_crt1, "CRT1", NULL, 1, "Configure first output (CRT1)", "");
    gtk_notebook_append_page(GTK_NOTEBOOK(gui.notebook), gui.c1_frame, gui.c1_label);
}


/**************************************************
 *                   CRT2 page                    *
 **************************************************/

/* CRT2 type menu and graphics */

void
gui_crt2_set_type_menu_index(int index)
{
   /* C2_MI_* are identical with our real menu indices */
   gtk_option_menu_set_history(GTK_OPTION_MENU(gui.om_omenu), (gint)index);
}

int
gui_crt2_get_type_menu_index(void)
{
   /* C2_MI_* are identical with our real menu indices */
   return(int)(gtk_option_menu_get_history(GTK_OPTION_MENU(gui.om_omenu)));
}

void
gui_crt2_hide_crt2_menu_items(void)
{
   gtk_widget_hide(gui.om_menuitemlcd);
   gtk_widget_hide(gui.om_menuitemsvhs);
   gtk_widget_hide(gui.om_menuitemcvbs);
   gtk_widget_hide(gui.om_menuitemcvsv);
   gtk_widget_hide(gui.om_menuitemtv);
   gtk_widget_hide(gui.om_menuitemhivi);
   gtk_widget_hide(gui.om_menuitemy525i);
   gtk_widget_hide(gui.om_menuitemy525p);
   gtk_widget_hide(gui.om_menuitemy750p);
   gtk_widget_hide(gui.om_menuitemy1080i);
   gtk_widget_hide(gui.om_menuitemvga2);
   gtk_widget_hide(gui.om_menuitemscart);
   gtk_widget_hide(gui.om_menuitemoff);
}

void
gui_crt2_show_crt2_menu_item(int index)
{
   switch(index) {
   case C2_MI_TV_LCD:
      gtk_widget_show(gui.om_menuitemlcd);
      break;
   case C2_MI_TV_SVHS:
      gtk_widget_show(gui.om_menuitemsvhs);
      break;
   case C2_MI_TV_CVBS:
      gtk_widget_show(gui.om_menuitemcvbs);
      break;
   case C2_MI_TV_CVSV:
      gtk_widget_show(gui.om_menuitemcvsv);
      break;
   case C2_MI_TV_TV:
      gtk_widget_show(gui.om_menuitemtv);
      break;
   case C2_MI_TV_HIVI:
      gtk_widget_show(gui.om_menuitemhivi);
      break;
   case C2_MI_TV_525I:
      gtk_widget_show(gui.om_menuitemy525i);
      break;
   case C2_MI_TV_525P:
      gtk_widget_show(gui.om_menuitemy525p);
      break;
   case C2_MI_TV_750P:
      gtk_widget_show(gui.om_menuitemy750p);
      break;
   case C2_MI_TV_1080I:
      gtk_widget_show(gui.om_menuitemy1080i);
      break;
   case C2_MI_TV_VGA:
      gtk_widget_show(gui.om_menuitemvga2);
      break;
   case C2_MI_TV_SCART:
      gtk_widget_show(gui.om_menuitemscart);
      break;
   case C2_MI_TV_OFF:
      gtk_widget_show(gui.om_menuitemoff);
   }
}

void
gui_crt2_select_gfx(int gfxnum)
{
   gtk_widget_hide(gui.om_lcd);
   gtk_widget_hide(gui.om_tv);
   gtk_widget_hide(gui.om_tvs);
   gtk_widget_hide(gui.om_tvc);
   gtk_widget_hide(gui.om_tvalign);
   gtk_widget_hide(gui.om_ypbpr);
   gtk_widget_hide(gui.om_vga);

   switch(gfxnum) {
   case C2_GFX_TV_LCD:
      gtk_widget_show(gui.om_lcd);
      break;
   case C2_GFX_TV_TV:
      gtk_widget_show(gui.om_tv);
      break;
   case C2_GFX_TV_SVHS:
      gtk_widget_show(gui.om_tvs);
      break;
   case C2_GFX_TV_CVBS:
      gtk_widget_show(gui.om_tvc);
      break;
   case C2_GFX_TV_CVSV:
      gtk_widget_show(gui.om_tvalign);
      break;
   case C2_GFX_TV_HDTV:
      gtk_widget_show(gui.om_ypbpr);
      break;
   case C2_GFX_TV_VGA:
      gtk_widget_show(gui.om_vga);
      break;
   }
}

/* Update CRT2 gfx according to current selection in menu */
void
gui_crt2_update_gfx(void)
{
   int c2_index = gui_crt2_get_type_menu_index();
   int c2_gfx = C2_GFX_NONE;

   switch(c2_index) {
   case C2_MI_TV_LCD:
      c2_gfx = C2_GFX_TV_LCD;
      break;
   case C2_MI_TV_TV:
   case C2_MI_TV_SCART:
      c2_gfx = C2_GFX_TV_TV;
      break;
   case C2_MI_TV_SVHS:
      c2_gfx = C2_GFX_TV_SVHS;
      break;
   case C2_MI_TV_CVBS:
      c2_gfx = C2_GFX_TV_CVBS;
      break;
   case C2_MI_TV_CVSV:
      c2_gfx = C2_GFX_TV_CVSV;
      break;
   case C2_MI_TV_HIVI:
   case C2_MI_TV_525I:
   case C2_MI_TV_525P:
   case C2_MI_TV_750P:
   case C2_MI_TV_1080I:
      c2_gfx = C2_GFX_TV_HDTV;
      break;
   case C2_MI_TV_VGA:
      c2_gfx = C2_GFX_TV_VGA;
      break;
   }

   gui_crt2_select_gfx(c2_gfx);
}

/* LCD scaling menu */

void
gui_crt2_set_scaling_menu_index(int index)
{
   /* CO_SC_* are identical with our real menu indices */
   if(gui.om_omenuls) {
      gtk_option_menu_set_history(GTK_OPTION_MENU(gui.om_omenuls), (gint)index);
   }
}

int
gui_crt2_get_scaling_menu_index(int retdefault)
{
   /* CO_SC_* are identical with our real menu indices */
   if(!gui.om_menuls) return retdefault;
   return(int)(gtk_option_menu_get_history(GTK_OPTION_MENU(gui.om_omenuls)));
}

void
gui_crt2_show_scaling_menu_and_label(void)
{
   gtk_widget_show(gui.om_omenuls);
   gtk_widget_show(gui.om_label_lcds);
}

void
gui_crt2_hide_scaling_menu_and_label(void)
{
   gtk_widget_hide(gui.om_omenuls);
   gtk_widget_hide(gui.om_label_lcds);
}

/* LCD centering menu */

void
gui_crt2_set_centering_menu_index(int index)
{
   /* CO_CE_* are identical with our real menu indices */
   if(gui.om_omenulcc) {
      gtk_option_menu_set_history(GTK_OPTION_MENU(gui.om_omenulcc), (gint)index);
   }
}

int
gui_crt2_get_centering_menu_index(int retdefault)
{
   /* CO_CE_* are identical with our real menu indices */
   if(!gui.om_omenulcc) return retdefault;
   return(int)(gtk_option_menu_get_history(GTK_OPTION_MENU(gui.om_omenulcc)));
}

void
gui_crt2_show_center_menu(void)
{
   gtk_widget_show(gui.om_omenulcc);
}

void
gui_crt2_hide_center_menu(void)
{
   gtk_widget_hide(gui.om_omenulcc);
}

/* TV standard menu */

void
gui_crt2_set_tvstd_menu_index(int index)
{
   /* C2_MI_TV_* are identical with our real menu indices */
   gtk_option_menu_set_history(GTK_OPTION_MENU(gui.om_omenut), index);
}

int
gui_crt2_get_tvstd_menu_index(void)
{
   /* C2_MI_TV_* are identical with our real menu indices */
   return(int)(gtk_option_menu_get_history(GTK_OPTION_MENU(gui.om_omenut)));
}

void
gui_crt2_show_tvstd_menu_and_label(void)
{
   gtk_widget_show(gui.om_omenut);
   gtk_widget_show(gui.om_label_tvtype);
}

void
gui_crt2_hide_tvstd_menu_and_label(void)
{
   gtk_widget_hide(gui.om_omenut);
   gtk_widget_hide(gui.om_label_tvtype);
}

/* Aspect ratio menu */

void
gui_crt2_set_aspect_menu_index(int index)
{
   /* C2_MI_AS_* are identical with our real menu indices */
   gtk_option_menu_set_history(GTK_OPTION_MENU(gui.om_omenuas), index);
}

int
gui_crt2_get_aspect_menu_index(void)
{
   /* C2_MI_AS_* are identical with our real menu indices */
   return(int)(gtk_option_menu_get_history(GTK_OPTION_MENU(gui.om_omenuas)));
}

void
gui_crt2_show_aspect_menu_and_label(void)
{
   gtk_widget_show(gui.om_omenuas);
   gtk_widget_show(gui.om_label_tvas);
}

void
gui_crt2_hide_aspect_menu_and_label(void)
{
   gtk_widget_hide(gui.om_omenuas);
   gtk_widget_hide(gui.om_label_tvas);
}

/* Chrontel overscan menu */

void
gui_crt2_set_overscan_menu_index(int index)
{
   /* C2_MI_OV_* are identical with our real menu indices */
   gtk_option_menu_set_history(GTK_OPTION_MENU(gui.om_omenuo), (gint)index);
}

int
gui_crt2_get_overscan_menu_index(void)
{
   /* C2_MI_OV_* are identical with our real menu indices */
   return(int)(gtk_option_menu_get_history(GTK_OPTION_MENU(gui.om_omenuo)));
}

void
gui_crt2_show_overscan_superover_menu_item(void)
{
   gtk_widget_show(gui.om_menuitemsover);
}

void
gui_crt2_hide_overscan_superover_menu_item(void)
{
   gtk_widget_hide(gui.om_menuitemsover);
}

/* "Show detected devices toggle button */

int
gui_crt2_show_detected_devices_checked(void)
{
   return (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(gui.om_check))) ? 1 : 0;
}

void
gui_crt2_show_hide_overscan_menu(void)
{
   gint index1, index2;

   if(gl.crt2_page_type != CRT2_PAGE_TYPE_NORMAL) return;
   if(!gl.supportoverscan) return;

   index1 = gui_crt2_get_type_menu_index();
   index2 = gui_crt2_get_tvstd_menu_index();

   if(index1 >= C2_MI_TV_SVHS && index1 <= C2_MI_TV_TV) {
      gtk_widget_show(gui.om_omenuo);
      if((index2 == C2_MI_TV_PAL) && gl.supportsoverscan) {
         gui_crt2_show_overscan_superover_menu_item();
      } else {
	 if(gui_crt2_get_overscan_menu_index() == C2_MI_OV_SUPEROVERSCAN) {
	    gui_crt2_set_overscan_menu_index(C2_MI_OV_OVERSCAN);
	 }
	 gui_crt2_hide_overscan_superover_menu_item();
      }
   } else {
      gtk_widget_hide(gui.om_omenuo);
   }
}

void
gui_crt2_show_tv_std_and_overscan_menus(void)
{
   int index;

   if(gl.crt2_page_type != CRT2_PAGE_TYPE_NORMAL) return;

   index = gui_crt2_get_type_menu_index();

   if(index >= C2_MI_TV_SVHS && index <= C2_MI_TV_TV) {
      gui_crt2_show_tvstd_menu_and_label();
      gui_crt2_show_hide_overscan_menu();
   } else {
      gui_crt2_hide_tvstd_menu_and_label();
      gui_crt2_show_hide_overscan_menu();
   }
}

void
gui_crt2_show_hide_aspect_menu(void)
{
   int index;

   if(gl.crt2_page_type != CRT2_PAGE_TYPE_NORMAL) return;
   if(!gl.supportypbprar) return;

   index = gui_crt2_get_type_menu_index();

   if(index >= C2_MI_TV_525I && index <= C2_MI_TV_1080I) {
      gui_crt2_show_aspect_menu_and_label();
   } else {
      gui_crt2_hide_aspect_menu_and_label();
   }
}

/* signal handlers */

static void
om_check_toggled(GtkWidget *widget, gpointer data)
{
    sisctrl_crt2_show_detected_only_changed();
}


static void
om_redetect(GtkWidget *widget, gpointer data)
{
    sisctrl_crt2_redetect_devices();
}

static void
om_menu_changed(GtkWidget *widget, gpointer data)
{
   gui_crt2_show_tv_std_and_overscan_menus();
   gui_crt2_show_hide_aspect_menu();
   gui_show_lcd_menus(2);
   gui_crt2_update_gfx();
}

static void
om_menut_changed(GtkWidget *widget, gpointer data)
{
   gui_crt2_show_hide_overscan_menu();
}

static void
om_omenuls_changed(GtkWidget *widget, gpointer data)
{
    int index = gui_crt2_get_scaling_menu_index(CO_MI_SC_DEFAULT);

    if(!gl.supportcenter) return;

    if(index == CO_MI_SC_ON) {
       gui_crt2_hide_center_menu();
    } else {
       gui_crt2_show_center_menu();
    }
}


/* Build page */

static void
build_crt2_page(void)
{
    gui.om_frame = gtk_frame_new("CRT2 settings");
    gtk_container_set_border_width(GTK_CONTAINER(gui.om_frame), 8);
    gtk_widget_show(gui.om_frame);

    gui.om_table = gtk_table_new(6, 5, TRUE);
    gtk_table_set_row_spacings(GTK_TABLE(gui.om_table), 2);
    gtk_table_set_col_spacings(GTK_TABLE(gui.om_table), 10);

    gtk_container_set_border_width(GTK_CONTAINER(gui.om_table), 8);

    gtk_container_add(GTK_CONTAINER(gui.om_frame), gui.om_table);
    gtk_widget_show(gui.om_table);

    switch(gl.crt2_page_type) {

    case CRT2_PAGE_TYPE_NORMAL:
       /* Build device type menu and graphics */
       gui.om_label_type = gtk_label_new("Device type");
       gtk_table_attach_defaults(GTK_TABLE(gui.om_table), gui.om_label_type, 0, 1, 3, 4);
       gtk_widget_show(gui.om_label_type);

       gui.om_ilcd = gdk_pixmap_create_from_xpm_d(gui.window->window, &gui.om_mlcd, NULL, crt2lcd_xpm);
       gui.om_lcd = gtk_image_new_from_pixmap(gui.om_ilcd, gui.om_mlcd);
       gui.om_itv = gdk_pixmap_create_from_xpm_d(gui.window->window, &gui.om_mtv, NULL, crt2tv_xpm);
       gui.om_tv = gtk_image_new_from_pixmap(gui.om_itv, gui.om_mtv);
       gui.om_iypbpr = gdk_pixmap_create_from_xpm_d(gui.window->window, &gui.om_mypbpr, NULL, crt2ypbpr_xpm);
       gui.om_ypbpr = gtk_image_new_from_pixmap(gui.om_iypbpr, gui.om_mypbpr);
       gui.om_itvs = gdk_pixmap_create_from_xpm_d(gui.window->window, &gui.om_mtvs, NULL, crt2svideo_xpm);
       gui.om_tvs = gtk_image_new_from_pixmap(gui.om_itvs, gui.om_mtvs);
       gui.om_itvc = gdk_pixmap_create_from_xpm_d(gui.window->window, &gui.om_mtvc, NULL, crt2cvbs_xpm);
       gui.om_tvc = gtk_image_new_from_pixmap(gui.om_itvc, gui.om_mtvc);
       gui.om_ivga = gdk_pixmap_create_from_xpm_d(gui.window->window, &gui.om_mvga, NULL, crt2vga_xpm);
       gui.om_vga = gtk_image_new_from_pixmap(gui.om_ivga, gui.om_mvga);

       gui.om_tvbox = gtk_hbox_new(FALSE, 0);
       gtk_container_set_border_width(GTK_CONTAINER(gui.om_tvbox), 1);
       gui.om_tvc2 = gtk_image_new_from_pixmap(gui.om_itvc, gui.om_mtvc);
       gui.om_tvs2 = gtk_image_new_from_pixmap(gui.om_itvs, gui.om_mtvs);
       gtk_box_pack_start(GTK_BOX(gui.om_tvbox), gui.om_tvc2, FALSE, FALSE, 5);
       gtk_box_pack_start(GTK_BOX(gui.om_tvbox), gui.om_tvs2, FALSE, FALSE, 0);
       gtk_widget_show(gui.om_tvc2);
       gtk_widget_show(gui.om_tvs2);
       gtk_widget_show(gui.om_tvbox);
       gui.om_tvalign = gtk_alignment_new(0.5, 0.5, 0, 0);
       gtk_container_add(GTK_CONTAINER(gui.om_tvalign), gui.om_tvbox);

       gtk_table_attach_defaults(GTK_TABLE(gui.om_table), gui.om_lcd, 0, 5, 0, 3);
       gtk_table_attach_defaults(GTK_TABLE(gui.om_table), gui.om_tv, 0, 5, 0, 3);
       gtk_table_attach_defaults(GTK_TABLE(gui.om_table), gui.om_tvs, 0, 5, 0, 3);
       gtk_table_attach_defaults(GTK_TABLE(gui.om_table), gui.om_tvc, 0, 5, 0, 3);
       gtk_table_attach_defaults(GTK_TABLE(gui.om_table), gui.om_tvalign, 0, 5, 0, 3);
       gtk_table_attach_defaults(GTK_TABLE(gui.om_table), gui.om_ypbpr, 0, 5, 0, 3);
       gtk_table_attach_defaults(GTK_TABLE(gui.om_table), gui.om_vga, 0, 5, 0, 3);
       gtk_widget_hide(gui.om_lcd);
       gtk_widget_hide(gui.om_tv);
       gtk_widget_hide(gui.om_tvs);
       gtk_widget_hide(gui.om_tvc);
       gtk_widget_hide(gui.om_tvalign);
       gtk_widget_hide(gui.om_ypbpr);
       gtk_widget_hide(gui.om_vga);

       gui.om_omenu = gtk_option_menu_new();

       gui.om_menu = gtk_menu_new();

       if(gl.supportdvi) {
          gui.om_menuitemlcd= gtk_menu_item_new_with_label("LCD/DVI-D (digital output)");
       } else {
	  gui.om_menuitemlcd= gtk_menu_item_new_with_label("LCD");
       }
       gui.om_menuitemsvhs  = gtk_menu_item_new_with_label("TV (S-Video)");
       gui.om_menuitemcvbs  = gtk_menu_item_new_with_label("TV (Composite)");
       gui.om_menuitemcvsv  = gtk_menu_item_new_with_label("TV (Composite + S-Video)");
       gui.om_menuitemtv    = gtk_menu_item_new_with_label("TV");
       gui.om_menuitemhivi  = gtk_menu_item_new_with_label("HDTV (HiVision 1080i)");
       gui.om_menuitemy525i = gtk_menu_item_new_with_label("HDTV (YPbPr 480i/525i)");
       gui.om_menuitemy525p = gtk_menu_item_new_with_label("HDTV (YPbPr 480p/525p)");
       gui.om_menuitemy750p = gtk_menu_item_new_with_label("HDTV (YPbPr 720p/750p)");
       gui.om_menuitemy1080i= gtk_menu_item_new_with_label("HDTV (YPbPr 1080i)");
       gui.om_menuitemvga2  = gtk_menu_item_new_with_label("VGA/DVI-A (analog; secondary plug)");
       gui.om_menuitemscart = gtk_menu_item_new_with_label("TV (SCART PAL)");
       gui.om_menuitemoff   = gtk_menu_item_new_with_label("OFF");

       gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menu), gui.om_menuitemlcd);
       gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menu), gui.om_menuitemsvhs);
       gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menu), gui.om_menuitemcvbs);
       gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menu), gui.om_menuitemcvsv);
       gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menu), gui.om_menuitemtv);
       gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menu), gui.om_menuitemhivi);
       gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menu), gui.om_menuitemy525i);
       gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menu), gui.om_menuitemy525p);
       gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menu), gui.om_menuitemy750p);
       gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menu), gui.om_menuitemy1080i);
       gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menu), gui.om_menuitemvga2);
       gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menu), gui.om_menuitemscart);
       gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menu), gui.om_menuitemoff);

       sisctrl_crt2_show_initial_crt2_type_menu();

       gtk_option_menu_set_menu(GTK_OPTION_MENU(gui.om_omenu), gui.om_menu);

       sisctrl_crt2_set_initial_crt2_type_history();

       /* gtk_table_attach_defaults(GTK_TABLE(gui.om_table), gui.om_omenu, 1, 5, 3, 4); */
       gtk_table_attach(GTK_TABLE(gui.om_table), gui.om_omenu, 1, 5, 3, 4, GTK_FILL|GTK_EXPAND, 0, 0, 0);

       gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.om_omenu,
		"Select second output (CRT2) device type", "");

       /* Build "TV standard" menu */
       gui.om_label_tvtype = gtk_label_new("TV standard");
       gtk_table_attach_defaults(GTK_TABLE(gui.om_table), gui.om_label_tvtype, 0, 1, 5, 6);

       gui.om_omenut = gtk_option_menu_new();

       gui.om_menut = gtk_menu_new();

       gui.om_menuitempal = gtk_menu_item_new_with_label("PAL");
       gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menut), gui.om_menuitempal);
       gtk_widget_show(gui.om_menuitempal);

       gui.om_menuitemntsc = gtk_menu_item_new_with_label("NTSC");
       gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menut), gui.om_menuitemntsc);
       gtk_widget_show(gui.om_menuitemntsc);

       if(gl.supportpalmn) {
          gui.om_menuitempalm = gtk_menu_item_new_with_label("PAL-M");
          gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menut), gui.om_menuitempalm);
          gtk_widget_show(gui.om_menuitempalm);
          gui.om_menuitempaln = gtk_menu_item_new_with_label("PAL-N");
          gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menut), gui.om_menuitempaln);
          gtk_widget_show(gui.om_menuitempaln);
       }
       if(gl.supportntscj) {
          gui.om_menuitemntscj = gtk_menu_item_new_with_label("NTSC-J");
          gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menut), gui.om_menuitemntscj);
          gtk_widget_show(gui.om_menuitemntscj);
       }

       gtk_option_menu_set_menu(GTK_OPTION_MENU(gui.om_omenut), gui.om_menut);

       /* gtk_table_attach_defaults(GTK_TABLE(gui.om_table), gui.om_omenut, 1,
	       (vbsupportsoverscan()) ? 3 : 5, 5, 6); */
       gtk_table_attach(GTK_TABLE(gui.om_table), gui.om_omenut, 1,
		(gl.supportoverscan) ? 3 : 5, 5, 6, GTK_FILL|GTK_EXPAND, 0, 0, 0);

       gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.om_omenut,
		   "Select TV standard", "");

       /* Build "aspect ratio" menu */
       if(gl.supportypbprar) {
	  gui.om_label_tvas = gtk_label_new("Aspect ratio");
          gtk_table_attach_defaults(GTK_TABLE(gui.om_table), gui.om_label_tvas, 0, 1, 5, 6);

          gui.om_omenuas = gtk_option_menu_new();

	  gui.om_menuas = gtk_menu_new();

          gui.om_menuitem43 = gtk_menu_item_new_with_label("4:3");
          gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menuas), gui.om_menuitem43);
          gtk_widget_show(gui.om_menuitem43);
	  gui.om_menuitem43lb = gtk_menu_item_new_with_label("4:3 Letterbox");
          gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menuas), gui.om_menuitem43lb);
          gtk_widget_show(gui.om_menuitem43lb);
	  gui.om_menuitem169 = gtk_menu_item_new_with_label("16:9");
          gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menuas), gui.om_menuitem169);
          gtk_widget_show(gui.om_menuitem169);

	  gtk_option_menu_set_menu(GTK_OPTION_MENU(gui.om_omenuas), gui.om_menuas);
	  /* gtk_table_attach_defaults(GTK_TABLE(gui.om_table), gui.om_omenuas, 1, 5, 5, 6); */
	  gtk_table_attach(GTK_TABLE(gui.om_table), gui.om_omenuas, 1, 5, 5, 6, GTK_FILL|GTK_EXPAND, 0, 0, 0);

	  gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.om_omenuas,
		   "Select TV aspect ratio", "");
       }

       /* Build "overscan" menu */
       if(gl.supportoverscan) {
	  gui.om_omenuo = gtk_option_menu_new();
          gui.om_menuo = gtk_menu_new();

          gui.om_menuitembios = gtk_menu_item_new_with_label("BIOS");
          gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menuo), gui.om_menuitembios);
          gtk_widget_show(gui.om_menuitembios);

          gui.om_menuitemunder = gtk_menu_item_new_with_label("Underscan");
          gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menuo), gui.om_menuitemunder);
          gtk_widget_show(gui.om_menuitemunder);

	  gui.om_menuitemover = gtk_menu_item_new_with_label("Overscan");
          gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menuo), gui.om_menuitemover);
          gtk_widget_show(gui.om_menuitemover);

	  gui.om_menuitemsover = gtk_menu_item_new_with_label("S-Overscan");
          gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menuo), gui.om_menuitemsover);
          gtk_widget_show(gui.om_menuitemsover);

	  gtk_option_menu_set_menu(GTK_OPTION_MENU(gui.om_omenuo), gui.om_menuo);

	  /* gtk_table_attach_defaults(GTK_TABLE(gui.om_table), gui.om_omenuo, 3, 5, 5, 6); */
	  gtk_table_attach(GTK_TABLE(gui.om_table), gui.om_omenuo, 3, 5, 5, 6, GTK_FILL|GTK_EXPAND, 0, 0, 0);

	  gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.om_omenuo,
		   "Enable overscan output", "");
       }

       /* Build "LCD scaling" and "LCD center" menus */
       if(gl.supportscale) {
	  gui.om_label_lcds = gtk_label_new("LCD scaling");
          gtk_table_attach_defaults(GTK_TABLE(gui.om_table), gui.om_label_lcds, 0, 1, 5, 6);

          gui.om_omenuls = gtk_option_menu_new();
          gui.om_menuls = gtk_menu_new();
          gui.om_menuitemlsd = gtk_menu_item_new_with_label("Default");
          gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menuls), gui.om_menuitemlsd);
          gtk_widget_show(gui.om_menuitemlsd);
	  gui.om_menuitemlso = gtk_menu_item_new_with_label("On");
          gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menuls), gui.om_menuitemlso);
          gtk_widget_show(gui.om_menuitemlso);
	  gui.om_menuitemlsf = gtk_menu_item_new_with_label("Off");
          gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menuls), gui.om_menuitemlsf);
          gtk_widget_show(gui.om_menuitemlsf);
	  gtk_option_menu_set_menu(GTK_OPTION_MENU(gui.om_omenuls), gui.om_menuls);

	  gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.om_omenuls,
			"Scale low resolution modes to panel's native resolution", "");

	  if(gl.supportcenter) {
	     /* gtk_table_attach_defaults(GTK_TABLE(gui.om_table), gui.om_omenuls, 1, 3, 5, 6); */
	     gtk_table_attach(GTK_TABLE(gui.om_table), gui.om_omenuls, 1, 3, 5, 6, GTK_FILL|GTK_EXPAND, 0, 0, 0);
	     gui.om_omenulcc = gtk_option_menu_new();
             gui.om_menulcc = gtk_menu_new();
             gui.om_menuitemlccd = gtk_menu_item_new_with_label("Default");
             gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menulcc), gui.om_menuitemlccd);
             gtk_widget_show(gui.om_menuitemlccd);
	     gui.om_menuitemlccc = gtk_menu_item_new_with_label("Don't center");
             gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menulcc), gui.om_menuitemlccc);
             gtk_widget_show(gui.om_menuitemlccc);
	     gui.om_menuitemlccp = gtk_menu_item_new_with_label("Center");
             gtk_menu_shell_append(GTK_MENU_SHELL(gui.om_menulcc), gui.om_menuitemlccp);
             gtk_widget_show(gui.om_menuitemlccp);
	     gtk_option_menu_set_menu(GTK_OPTION_MENU(gui.om_omenulcc), gui.om_menulcc);
	     /* gtk_table_attach_defaults(GTK_TABLE(gui.om_table), gui.om_omenulcc, 3, 5, 5, 6); */
	     gtk_table_attach(GTK_TABLE(gui.om_table), gui.om_omenulcc, 3, 5, 5, 6, GTK_FILL|GTK_EXPAND, 0, 0, 0);
	     gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.om_omenulcc,
		 	"Center unscaled output or pass mode timing to the panel", "");
	  } else {
	     /* gtk_table_attach_defaults(GTK_TABLE(gui.om_table), gui.om_omenuls, 1, 5, 5, 6); */
	     gtk_table_attach(GTK_TABLE(gui.om_table), gui.om_omenuls, 1, 5, 5, 6, GTK_FILL|GTK_EXPAND, 0, 0, 0);
	  }
       }

       /* Show/hide menus and set menu histories (=currently selected items) */
       gui_crt2_show_tv_std_and_overscan_menus();

       sisctrl_crt2_set_initial_tv_std();

       if(gl.supportoverscan) {
          gui_crt2_set_overscan_menu_index(sisctrl_get_overscan_history());
       }

       if(gl.supportypbprar) {
	  gui_crt2_show_hide_aspect_menu();
	  sisctrl_crt2_set_initial_aspect_ratio();
       }

       if(gl.supportscale) {
          gui_crt2_set_scaling_menu_index(sisctrl_get_lcd_scaling_history());
	  if(gl.supportcenter) {
	     gui_crt2_set_centering_menu_index(sisctrl_get_lcd_center_history());
	  }
	  gui_show_lcd_menus(2);
       }

       /* Build "show detected only" check box and "redetect" button */
       gui.om_check = gtk_check_button_new_with_label("Show detected devices only");
       if(gl.supportcrt2redetect) {
	  gui.om_table2 = gtk_table_new(1, 2, FALSE);
          gtk_table_set_row_spacings(GTK_TABLE(gui.om_table2), 0);
          gtk_table_set_col_spacings(GTK_TABLE(gui.om_table2), 0);
          gtk_container_set_border_width(GTK_CONTAINER(gui.om_table2), 0);
	  gtk_table_attach_defaults(GTK_TABLE(gui.om_table2), gui.om_check, 0, 1, 0, 1);
       } else {
	  gtk_table_attach_defaults(GTK_TABLE(gui.om_table), gui.om_check, 1, 5, 4, 5);
       }
       gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.om_check,
		 	"Show all device types or detected ones only", "");
       gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(gui.om_check), TRUE);
       gtk_widget_show(gui.om_check);
       if(gl.supportcrt2redetect) {
	  gui.om_redetect = gtk_button_new();
	  gui.om_redetectbox = xpm_label_box(icon_redetect, "Re_detect", gui.om_redetect, 3, "", "");
	  gtk_container_add(GTK_CONTAINER(gui.om_redetect), gui.om_redetectbox);
          /* gtk_table_attach_defaults(GTK_TABLE(gui.om_table2), gui.om_redetect, 1, 2, 0, 1); */
	  gtk_table_attach(GTK_TABLE(gui.om_table2), gui.om_redetect, 1, 2, 0, 1, GTK_FILL|GTK_EXPAND, 0, 0, 0);
	  gtk_widget_show(gui.om_redetectbox);
	  gtk_widget_show(gui.om_redetect);
          gtk_widget_show(gui.om_table2);
	  gtk_table_attach_defaults(GTK_TABLE(gui.om_table), gui.om_table2, 1, 5, 4, 5);
	  gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.om_redetect,
		 	"Re-detect connected output devices", "");
       }

       /* Attach signal handler for "device type" menu */
       g_signal_connect(G_OBJECT(gui.om_menuitemlcd),  "activate", G_CALLBACK(om_menu_changed), NULL);
       g_signal_connect(G_OBJECT(gui.om_menuitemsvhs), "activate", G_CALLBACK(om_menu_changed), NULL);
       g_signal_connect(G_OBJECT(gui.om_menuitemcvbs), "activate", G_CALLBACK(om_menu_changed), NULL);
       g_signal_connect(G_OBJECT(gui.om_menuitemcvsv), "activate", G_CALLBACK(om_menu_changed), NULL);
       g_signal_connect(G_OBJECT(gui.om_menuitemtv),   "activate", G_CALLBACK(om_menu_changed), NULL);
       if(gl.supporthivision) {
	  g_signal_connect(G_OBJECT(gui.om_menuitemhivi), "activate", G_CALLBACK(om_menu_changed), NULL);
       }
       if(gl.supportypbpr) {
	  g_signal_connect(G_OBJECT(gui.om_menuitemy525i), "activate", G_CALLBACK(om_menu_changed), NULL);
	  g_signal_connect(G_OBJECT(gui.om_menuitemy525p), "activate", G_CALLBACK(om_menu_changed), NULL);
	  g_signal_connect(G_OBJECT(gui.om_menuitemy750p), "activate", G_CALLBACK(om_menu_changed), NULL);
	  g_signal_connect(G_OBJECT(gui.om_menuitemy1080i), "activate", G_CALLBACK(om_menu_changed), NULL);
       }
       g_signal_connect(G_OBJECT(gui.om_menuitemvga2),  "activate", G_CALLBACK(om_menu_changed), NULL);
       g_signal_connect(G_OBJECT(gui.om_menuitemscart), "activate", G_CALLBACK(om_menu_changed), NULL);
       g_signal_connect(G_OBJECT(gui.om_menuitemoff),   "activate", G_CALLBACK(om_menu_changed), NULL);

       /* Attach signal handler for "lcd scaling" menu */
       if(gl.supportscale && gl.supportcenter)  {
	  g_signal_connect(G_OBJECT(gui.om_menuitemlsd), "activate", G_CALLBACK(om_omenuls_changed), NULL);
	  g_signal_connect(G_OBJECT(gui.om_menuitemlso), "activate", G_CALLBACK(om_omenuls_changed), NULL);
	  g_signal_connect(G_OBJECT(gui.om_menuitemlsf), "activate", G_CALLBACK(om_omenuls_changed), NULL);
       }

       /* Attach signal handler for "tv standard" menu */
       g_signal_connect(G_OBJECT(gui.om_menuitempal),   "activate", G_CALLBACK(om_menut_changed), NULL);
       g_signal_connect(G_OBJECT(gui.om_menuitemntsc),  "activate", G_CALLBACK(om_menut_changed), NULL);
       if(gl.supportpalmn) {
          g_signal_connect(G_OBJECT(gui.om_menuitempalm), "activate", G_CALLBACK(om_menut_changed), NULL);
          g_signal_connect(G_OBJECT(gui.om_menuitempaln), "activate", G_CALLBACK(om_menut_changed), NULL);
       }
       if(gl.supportntscj) {
	  g_signal_connect(G_OBJECT(gui.om_menuitemntscj), "activate", G_CALLBACK(om_menut_changed), NULL);
       }

       /* Attach signal handler for "show detected only" check box */
       g_signal_connect(G_OBJECT(gui.om_check), "toggled", G_CALLBACK(om_check_toggled), NULL);

       /* Attach signal handler for "redetect" button */
       if(gl.supportcrt2redetect) {
          g_signal_connect(G_OBJECT(gui.om_redetect), "clicked", G_CALLBACK(om_redetect), NULL);
       }

       gtk_widget_show(gui.om_omenu);

       break;

    case CRT2_PAGE_TYPE_NA_NOVB:
       gui.om_label_nodh = gtk_label_new("No videobridge present");
       gtk_widget_show(gui.om_label_nodh);
       gtk_table_attach_defaults(GTK_TABLE(gui.om_table), gui.om_label_nodh, 0, 5, 2, 4);
       break;

    case CRT2_PAGE_TYPE_NA_DHM:
       gui.om_label_nodh = gtk_label_new("Not supported in Dual Head mode");
       gtk_widget_show(gui.om_label_nodh);
       gtk_table_attach_defaults(GTK_TABLE(gui.om_table), gui.om_label_nodh, 0, 5, 2, 4);
       break;

    }

    gui.om_label = xpm_label_box(icon_crt2, "CRT2", NULL, 1, "Configure second output (CRT2)", "");
    gtk_notebook_append_page(GTK_NOTEBOOK(gui.notebook), gui.om_frame, gui.om_label);

}

/**************************************************
 *         Common for CRT1 and CRT2 pages         *
 **************************************************/

/* Show/hide LCD scaling/centering menu(s) for CRT1 (1) or CRT2 (2) pages */
void
gui_show_lcd_menus(int crtnum)
{
   int index1, index2, index3;

   if(crtnum == 1) {
      if(gl.crt1_page_type != CRT1_PAGE_TYPE_NORMAL) return;
   } else {
      if(gl.crt2_page_type != CRT2_PAGE_TYPE_NORMAL) return;
   }

   if(!gl.supportscale) return;

   if(crtnum == 2) {
      index1 = gui_crt2_get_type_menu_index();
      index2 = C2_MI_TV_LCD;
      index3 = gui_crt2_get_scaling_menu_index(CO_MI_SC_DEFAULT);
   } else {
      if(!gl.supportlcda) return;
      index1 = gui_crt1_get_type_menu_index();
      index2 = C1_MI_LCDA;
      index3 = gui_crt1_get_scaling_menu_index(CO_MI_SC_DEFAULT);
   }

   if(index1 == index2) {
      if(crtnum == 2) {
	 gui_crt2_show_scaling_menu_and_label();
      } else {
         gui_crt1_show_scaling_menu_and_label();
      }
      if(gl.supportcenter) {
         if(index3 == CO_MI_SC_ON) {
	    if(crtnum == 2) gui_crt2_hide_center_menu();
	    else	    gui_crt1_hide_center_menu();
	 } else {
            if(crtnum == 2) gui_crt2_show_center_menu();
	    else	    gui_crt1_show_center_menu();
	 }
      }
   } else {
      if(crtnum == 2) {
         gui_crt2_hide_scaling_menu_and_label();
      } else {
         gui_crt1_hide_scaling_menu_and_label();
      }
      if(gl.supportcenter) {
         if(crtnum == 2) gui_crt2_hide_center_menu();
	 else	         gui_crt1_hide_center_menu();
      }
   }
}


/**************************************************
 *                   Gamma page                   *
 **************************************************/

void
gui_gamma_set_all_spin_values(double s1, double s2, double s3,
			      double s4, double s5, double s6)
{
   gtk_spin_button_set_value(GTK_SPIN_BUTTON(gui.ga_spin1), s1);
   gtk_spin_button_set_value(GTK_SPIN_BUTTON(gui.ga_spin2), s2);
   gtk_spin_button_set_value(GTK_SPIN_BUTTON(gui.ga_spin3), s3);
   gtk_spin_button_set_value(GTK_SPIN_BUTTON(gui.ga_spin4), s4);
   gtk_spin_button_set_value(GTK_SPIN_BUTTON(gui.ga_spin5), s5);
   gtk_spin_button_set_value(GTK_SPIN_BUTTON(gui.ga_spin6), s6);
}

void
gui_gamma_set_gamma_enabled_check_box_stati(int b1, int b2)
{
   gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(gui.ga_check1), b1 ? TRUE : FALSE);
   gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(gui.ga_check2), b2 ? TRUE : FALSE);
}

void
gui_gamma_show_separate_gamma_radio_buttons(void)
{
   gtk_widget_show(gui.ga_radio_crt1);
   gtk_widget_show(gui.ga_radio_crt2);
}

void
gui_gamma_hide_separate_gamma_radio_buttons(void)
{
   gtk_widget_hide(gui.ga_radio_crt1);
   gtk_widget_hide(gui.ga_radio_crt2);
}

void
gui_gamma_enable_and_activate_separate_gamma_check_box(int active)
{
   gtk_widget_set_sensitive(gui.ga_check3, TRUE);
   gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(gui.ga_check3), active ? TRUE : FALSE);
}

void
gui_gamma_disable_separate_gamma_check_box(void)
{
   gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(gui.ga_check3), FALSE);
   gtk_widget_set_sensitive(gui.ga_check3, FALSE);
}

int
gui_gamma_get_active_separate_gamma_radio(void)
{
   /* Since there are only two radio buttons,
    * checking whether the first is active gives
    * us all we need to know.
    */
   return (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(gui.ga_radio_crt1))) ? 1 : 0;
}

void
gui_gamma_drawggfx(void)
{
   int i, j, k;

   for(i = 0; i < GACW*3; i++) gui.rgbbuf[i] = 0x44;
   for(i = 1; i < GACH-1; i++) {
      gui.rgbbuf[i * GACW * 3] = 0x44;
      gui.rgbbuf[(i * GACW * 3) + 1] = 0x44;
      gui.rgbbuf[(i * GACW * 3) + 2] = 0x44;
      for(j = 3; j < (GACW-1) * 3; j++) gui.rgbbuf[(i * GACW * 3) + j] = 0;
      gui.rgbbuf[(i * GACW * 3) + ((GACW - 1) * 3)] = 0xff;
      gui.rgbbuf[(i * GACW * 3) + ((GACW - 1) * 3) + 1] = 0xff;
      gui.rgbbuf[(i * GACW * 3) + ((GACW - 1) * 3) + 2] = 0xff;
   }
   for(i = 0; i < GACW*3; i++) gui.rgbbuf[((GACH-1) * GACW * 3) + i] = 0xff;

   /* curve */
   for(i = 0; i < 3; i++) {
      for(k = 1; k < GACW-1; k++) {
         j = (k - 1) * (gl.nramp / (GACW-2));
         gui.rgbbuf[(((GACH-3) - (gl.ramp[i][j] / (65535 / (GACH-3)))) * (GACW * 3)) + (GACW * 3) + (k * 3) + i] = 0xff;
      }
   }
}

/* Update the gamma curve gfx */
void
gui_gamma_update_gamma_gfx(void)
{
   if((gui.ga_drawable) && (gui.notebook_page == 3)) {
      gdk_draw_rgb_image(gui.ga_drawable->window, gui.ga_drawable->style->fg_gc[GTK_STATE_NORMAL],
   	                 0, 0, GACW, GACH, GDK_RGB_DITHER_MAX, gui.rgbbuf, GACW * 3);
   }
}

/* Signal handlers */

/* CRT1-gamma toggled */
static void
ga_check1_toggled(GtkWidget *widget, gpointer data)
{
    sisctrl_gamma_enable_1(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(gui.ga_check1)) ? 1 : 0);
}

/* CRT2-gamma toggled */
static void
ga_check2_toggled(GtkWidget *widget, gpointer data)
{
    sisctrl_gamma_enable_2(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(gui.ga_check2)) ? 1 : 0);
}

/* Gamma: "Adjust separately" toggled */
static void
ga_check3_toggled(GtkWidget *widget, gpointer data)
{
    sisctrl_gamma_separate_changed(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(gui.ga_check3)) ? 1 : 0);
}

/* Gamma: Radio "separate" CRT1 <> CRT2 toggled */
static void
ga_radio_changed(GtkWidget *widget, gpointer data)
{
    /* We connected out handler to first one only, so this suffices to check */
    sisctrl_gamma_separate_toggled(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget)) ? 1 : 0);
}

static void
ga_radio_rgb_changed(GtkWidget *widget, gpointer data)
{
   gui.ga_current_adj = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget)) ? 0 : 1;

   switch(gui.ga_current_adj) {
   case 0:	/* R/G/B */
      gtk_widget_show(gui.ga_spin2);
      gtk_widget_show(gui.ga_spin3);
      gtk_widget_show(gui.ga_spin5);
      gtk_widget_show(gui.ga_spin6);
      gtk_widget_show(gui.ga_label_red);
      gtk_widget_show(gui.ga_label_green);
      gtk_widget_show(gui.ga_label_blue);
      gtk_widget_show(gui.ga_label_red1);
      gtk_widget_show(gui.ga_label_green1);
      gtk_widget_show(gui.ga_label_blue1);
      sisctrl_gamma_spin_red_changed((double)gtk_spin_button_get_value(GTK_SPIN_BUTTON(gui.ga_spin1)));
      sisctrl_gamma_spin_green_changed((double)gtk_spin_button_get_value(GTK_SPIN_BUTTON(gui.ga_spin2)));
      sisctrl_gamma_spin_blue_changed((double)gtk_spin_button_get_value(GTK_SPIN_BUTTON(gui.ga_spin3)));
      sisctrl_gamma_spin_bri_red_changed((double)gtk_spin_button_get_value(GTK_SPIN_BUTTON(gui.ga_spin4)));
      sisctrl_gamma_spin_bri_green_changed((double)gtk_spin_button_get_value(GTK_SPIN_BUTTON(gui.ga_spin5)));
      sisctrl_gamma_spin_bri_blue_changed((double)gtk_spin_button_get_value(GTK_SPIN_BUTTON(gui.ga_spin6)));
      gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.ga_spin1,
		 	"Adjust red channel gamma", "");
      gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.ga_spin4,
		 	"Adjust red channel gamma brightness", "");
      break;

   default:	/* All */
      gtk_widget_hide(gui.ga_spin2);
      gtk_widget_hide(gui.ga_spin3);
      gtk_widget_hide(gui.ga_spin5);
      gtk_widget_hide(gui.ga_spin6);
      gtk_widget_hide(gui.ga_label_red);
      gtk_widget_hide(gui.ga_label_green);
      gtk_widget_hide(gui.ga_label_blue);
      gtk_widget_hide(gui.ga_label_red1);
      gtk_widget_hide(gui.ga_label_green1);
      gtk_widget_hide(gui.ga_label_blue1);
      sisctrl_gamma_spin_red_changed((double)gtk_spin_button_get_value(GTK_SPIN_BUTTON(gui.ga_spin1)));
      sisctrl_gamma_spin_green_changed((double)gtk_spin_button_get_value(GTK_SPIN_BUTTON(gui.ga_spin1)));
      sisctrl_gamma_spin_blue_changed((double)gtk_spin_button_get_value(GTK_SPIN_BUTTON(gui.ga_spin1)));
      sisctrl_gamma_spin_bri_red_changed((double)gtk_spin_button_get_value(GTK_SPIN_BUTTON(gui.ga_spin4)));
      sisctrl_gamma_spin_bri_green_changed((double)gtk_spin_button_get_value(GTK_SPIN_BUTTON(gui.ga_spin4)));
      sisctrl_gamma_spin_bri_blue_changed((double)gtk_spin_button_get_value(GTK_SPIN_BUTTON(gui.ga_spin4)));
      gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.ga_spin1,
		 	"Adjust gamma", "");
      gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.ga_spin4,
		 	"Adjust gamma brightness", "");
      break;
   }

}

static void
ga_spin1_changed(GtkWidget *widget, gpointer data)
{
   sisctrl_gamma_spin_red_changed((double)gtk_spin_button_get_value(GTK_SPIN_BUTTON(gui.ga_spin1)));
   if(gui.ga_current_adj) {
      sisctrl_gamma_spin_green_changed((double)gtk_spin_button_get_value(GTK_SPIN_BUTTON(gui.ga_spin1)));
      sisctrl_gamma_spin_blue_changed((double)gtk_spin_button_get_value(GTK_SPIN_BUTTON(gui.ga_spin1)));
   }
}

static void
ga_spin2_changed(GtkWidget *widget, gpointer data)
{
   if(gui.ga_current_adj) return;
   sisctrl_gamma_spin_green_changed((double)gtk_spin_button_get_value(GTK_SPIN_BUTTON(gui.ga_spin2)));
}

static void
ga_spin3_changed(GtkWidget *widget, gpointer data)
{
   if(gui.ga_current_adj) return;
   sisctrl_gamma_spin_blue_changed((double)gtk_spin_button_get_value(GTK_SPIN_BUTTON(gui.ga_spin3)));
}

static void
ga_spin4_changed(GtkWidget *widget, gpointer data)
{
   sisctrl_gamma_spin_bri_red_changed((double)gtk_spin_button_get_value(GTK_SPIN_BUTTON(gui.ga_spin4)));
   if(gui.ga_current_adj) {
      sisctrl_gamma_spin_bri_green_changed((double)gtk_spin_button_get_value(GTK_SPIN_BUTTON(gui.ga_spin4)));
      sisctrl_gamma_spin_bri_blue_changed((double)gtk_spin_button_get_value(GTK_SPIN_BUTTON(gui.ga_spin4)));
   }
}

static void
ga_spin5_changed(GtkWidget *widget, gpointer data)
{
   if(gui.ga_current_adj) return;
   sisctrl_gamma_spin_bri_green_changed((double)gtk_spin_button_get_value(GTK_SPIN_BUTTON(gui.ga_spin5)));
}

static void
ga_spin6_changed(GtkWidget *widget, gpointer data)
{
   if(gui.ga_current_adj) return;
   sisctrl_gamma_spin_bri_blue_changed((double)gtk_spin_button_get_value(GTK_SPIN_BUTTON(gui.ga_spin6)));
}

static gboolean
ga_picture_exposed(GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
   gdk_draw_rgb_image(widget->window, widget->style->fg_gc[GTK_STATE_NORMAL],
   	0, 0, GACW, GACH, GDK_RGB_DITHER_MAX, gui.rgbbuf, GACW * 3);
   return TRUE;
}

static void
ga_draw_big_boxes(int stepw, int cello, int cellh, int cellvo)
{
	 gdk_gc_set_rgb_bg_color(gui.ga_context, &ga_colorblack);
	 gdk_gc_set_stipple(gui.ga_context, gui.ga_stripple2);
	 gdk_gc_set_ts_origin(gui.ga_context, 0, 0);
	 gdk_gc_set_fill(gui.ga_context, GDK_OPAQUE_STIPPLED);

	 gdk_gc_set_rgb_fg_color(gui.ga_context, &ga_colorred);
	 gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context, TRUE,
		      (stepw * 0) + cello, cellvo,
		      stepw * 2,           cellh);

	 gdk_gc_set_rgb_fg_color(gui.ga_context, &ga_colorgreen);
	 gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context, TRUE,
		      (stepw * 3) + cello, cellvo,
		      stepw  * 2,          cellh);

	 gdk_gc_set_rgb_fg_color(gui.ga_context, &ga_colorblue);
	 gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context, TRUE,
		      (stepw * 6) + cello, cellvo,
		      stepw * 2,           cellh);

	 gdk_gc_set_rgb_fg_color(gui.ga_context, &ga_colorwhite);
	 gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context, TRUE,
		      (stepw * 9) + cello, cellvo,
		      stepw * 2,           cellh);
}

static void
ga_draw_small_boxes(int stepw, int cello, int cellh, int cellvo)
{
	 gdk_gc_set_fill(gui.ga_context, GDK_SOLID);

	 gdk_gc_set_rgb_fg_color(gui.ga_context, &ga_colorredh);
	 gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context, TRUE,
		      (stepw * 0) + cello + (stepw / 2), cellvo + (cellh / 4),
		      stepw,                             cellh / 2 - 1);

	 gdk_gc_set_rgb_fg_color(gui.ga_context, &ga_colorgreenh);
	 gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context, TRUE,
		      (stepw * 3) + cello + (stepw / 2), cellvo + (cellh / 4),
		      stepw,                             cellh / 2 - 1);

	 gdk_gc_set_rgb_fg_color(gui.ga_context, &ga_colorblueh);
	 gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context, TRUE,
		      (stepw * 6) + cello + (stepw / 2), cellvo + (cellh / 4),
		      stepw,                             cellh / 2 - 1);


	 gdk_gc_set_rgb_fg_color(gui.ga_context, &ga_colorwhiteh);
	 gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context, TRUE,
		      (stepw * 9) + cello + (stepw / 2), cellvo + (cellh / 4),
		      stepw,                             cellh / 2 - 1);
}

static void
ga_update_test_pattern(void)
{
   gint lwidth;
   gint lheight;
   int  i, cellw, stepw, cello, cellh, steph, cellvo;
   float ga;

   if(!gui.ga_layout) return;

   if(!GTK_WIDGET_REALIZED(gui.ga_layout)) return;

   if(!(GTK_LAYOUT(gui.ga_layout)->bin_window)) return;

   lwidth = gui.ga_layout->allocation.width;
   lheight = gui.ga_layout->allocation.height;
   if(!lwidth || !lheight) return;

   if(!gui.ga_context) {
      gui.ga_context = gdk_gc_new(GTK_LAYOUT(gui.ga_layout)->bin_window);
   }

   gdk_gc_set_rgb_fg_color(gui.ga_context, &ga_bgcolors[gui.ga_current_gfx]);
   gdk_gc_set_fill(gui.ga_context, GDK_SOLID);
   gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context,
		      TRUE,
		      0, 0,
		      lwidth, lheight);

   switch(gui.ga_current_gfx) {
      case 0:
      case 1:
         stepw = (((lwidth * 95) / 100) / 11);
	 cello = (lwidth - (stepw * 11)) / 2;
         cellh = ((lheight * 95) / 100);
	 cellvo = (lheight - cellh) / 2;

	 ga_draw_big_boxes(stepw, cello, cellh, cellvo);

	 ga = (gui.ga_current_gfx == 0) ? gl.crt1mongamma : gl.crt2mongamma;
	 ga_colorredh.red =
	    ga_colorgreenh.green =
	    ga_colorblueh.blue =
	    ga_colorwhiteh.red =
	    ga_colorwhiteh.green =
	    ga_colorwhiteh.blue = (guint16)((float)(pow(0.5, 1.0 / ga) * 65536.0));

	 ga_draw_small_boxes(stepw, cello, cellh, cellvo);

	 break;

      case 2:
      case 3:
      case 4:
         stepw = (((lwidth * 95) / 100) / 11);
	 cello = (lwidth - (stepw * 11)) / 2;
         cellh = ((lheight * 95) / 100);
	 cellvo = (lheight - cellh) / 2;

	 ga_draw_big_boxes(stepw, cello, cellh, cellvo);

	 switch(gui.ga_current_gfx) {
	 case 2:  ga = 2.2; break;
	 case 3:  ga = 1.8; break;
	 default: ga = (float)gtk_spin_button_get_value(GTK_SPIN_BUTTON(gui.ga_sping));
	 }
	 ga_colorredh.red =
	    ga_colorgreenh.green =
	    ga_colorblueh.blue =
	    ga_colorwhiteh.red =
	    ga_colorwhiteh.green =
	    ga_colorwhiteh.blue = (guint16)((float)(pow(0.5, 1.0 / ga) * 65536.0));

	 ga_draw_small_boxes(stepw, cello, cellh, cellvo);

	 break;

      case 5:
      case 6:
      case 7:
         stepw = (((lwidth * 95) / 100) / 11);
	 cello = (lwidth - (stepw * 11)) / 2;
         cellh = ((lheight * 95) / 100);
	 cellvo = (lheight - cellh) / 2;

	 gdk_gc_set_rgb_bg_color(gui.ga_context, &ga_colorblack);
	 gdk_gc_set_stipple(gui.ga_context, gui.ga_stripple2);
	 gdk_gc_set_ts_origin(gui.ga_context, 0, 0);
	 gdk_gc_set_fill(gui.ga_context, GDK_OPAQUE_STIPPLED);

	 gdk_gc_set_rgb_fg_color(gui.ga_context, &ga_colorred);
	 gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context, TRUE,
		      (stepw * 1) + cello, cellvo,
		      stepw,               cellh);

	 gdk_gc_set_rgb_fg_color(gui.ga_context, &ga_colorgreen);
	 gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context, TRUE,
		      (stepw * 4) + cello, cellvo,
		      stepw,               cellh);

	 gdk_gc_set_rgb_fg_color(gui.ga_context, &ga_colorblue);
	 gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context, TRUE,
		      (stepw * 7) + cello, cellvo,
		      stepw,               cellh);

	 gdk_gc_set_rgb_fg_color(gui.ga_context, &ga_colorwhite);
	 gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context, TRUE,
		      (stepw * 10) + cello, cellvo,
		      stepw,               cellh);

	 switch(gui.ga_current_gfx) {
	 case 5:  ga = 2.2; break;
	 case 6:  ga = 1.8; break;
	 default: ga = (float)gtk_spin_button_get_value(GTK_SPIN_BUTTON(gui.ga_sping));
	 }

	 ga_colorredh.red =
	    ga_colorgreenh.green =
	    ga_colorblueh.blue =
	    ga_colorwhiteh.red =
	    ga_colorwhiteh.green =
	    ga_colorwhiteh.blue = (guint16)((float)(pow(0.5, 1.0 / ga) * 65536.0));

	 gdk_gc_set_fill(gui.ga_context, GDK_SOLID);

	 gdk_gc_set_rgb_fg_color(gui.ga_context, &ga_colorredh);
	 gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context, TRUE,
		      (stepw * 0) + cello, cellvo,
		      stepw,               cellh);

	 gdk_gc_set_rgb_fg_color(gui.ga_context, &ga_colorgreenh);
	 gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context, TRUE,
		      (stepw * 3) + cello, cellvo,
		      stepw,               cellh);

	 gdk_gc_set_rgb_fg_color(gui.ga_context, &ga_colorblueh);
	 gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context, TRUE,
		      (stepw * 6) + cello, cellvo,
		      stepw,               cellh);


	 gdk_gc_set_rgb_fg_color(gui.ga_context, &ga_colorwhiteh);
	 gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context, TRUE,
		      (stepw * 9) + cello, cellvo,
		      stepw,               cellh);

	 break;

      case 8: /* Brightness */
         cellw = (lwidth * 66) / 100;
	 cello = (lwidth - cellw) / 2;
	 cellh = (lheight * 85) / 100;
	 cellvo = (lheight - cellh) / 2;

	 gdk_gc_set_fill(gui.ga_context, GDK_SOLID);
	 gdk_gc_set_rgb_fg_color(gui.ga_context, &ga_colorblack);
	 gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context, TRUE,
		      cello, cellvo,
		      cellw, cellh);

	 cellw = (lwidth * 33) / 100;
	 cello = (lwidth - cellw) / 2;
	 cellh = (lheight * 65) / 100;
	 cellvo = (lheight - cellh) / 2;
	 gdk_gc_set_rgb_fg_color(gui.ga_context, &ga_colordgrey);
	 gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context, TRUE,
		      cello, cellvo,
		      cellw, cellh);
         break;

      case 9: /* RGB */
         stepw = (((lwidth * 90) / 100) / 16);
	 cello = (lwidth - (stepw * 16)) / 2;
         cellh = ((lheight * 70) / 100) / 3;
	 cellvo = (lheight - (cellh * 3)) / 4;
	 steph = cellh + cellvo;
	 gdk_gc_set_fill(gui.ga_context, GDK_SOLID);
         for(i = 0; i < 16; i++) {
	    gdk_gc_set_rgb_fg_color(gui.ga_context, &gui.ga_colorrgb[i]);
            gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context, TRUE,
		      (stepw * i) + cello, (steph * 0) + cellvo,
		      stepw,               cellh);
	    gdk_gc_set_rgb_fg_color(gui.ga_context, &gui.ga_colorrgb[i+16]);
            gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context, TRUE,
		      (stepw * i) + cello, (steph * 1) + cellvo,
		      stepw,               cellh);
	    gdk_gc_set_rgb_fg_color(gui.ga_context, &gui.ga_colorrgb[i+(16*2)]);
            gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context, TRUE,
		      (stepw * i) + cello, (steph * 2) + cellvo,
		      stepw,               cellh);
         }
         break;

      case 10: /* Grey */
         stepw = (((lwidth * 90) / 100) / 10);
	 cello = (lwidth - (stepw * 10)) / 2;
         cellh = ((lheight * 80) / 100) / 2;
	 cellvo = (lheight - (cellh * 2)) / 3;
	 steph = cellh + cellvo;
	 gdk_gc_set_fill(gui.ga_context, GDK_SOLID);
         for(i = 0; i < 10; i++) {
	    gdk_gc_set_rgb_fg_color(gui.ga_context, &gui.ga_colorgrey[i]);
            gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context, TRUE,
		      (stepw * i) + cello, (steph * 0) + cellvo,
		      stepw,               cellh);
	    gdk_gc_set_rgb_fg_color(gui.ga_context, &gui.ga_colorgrey[i+10]);
            gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context, TRUE,
		      (stepw * i) + cello, (steph * 1) + cellvo,
		      stepw,               cellh);
         }
         break;

      case 11: /* CMY */
         stepw = (((lwidth * 90) / 100) / 11);
	 cello = (lwidth - (stepw * 11)) / 2;
         cellh = ((lheight * 80) / 100) / 3;
	 cellvo = (lheight - (cellh * 3)) / 4;
	 steph = cellh + cellvo;
	 gdk_gc_set_fill(gui.ga_context, GDK_SOLID);
         for(i = 0; i < 11; i++) {
	    gdk_gc_set_rgb_fg_color(gui.ga_context, &gui.ga_colorcmy[i]);
            gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context, TRUE,
		      (stepw * i) + cello, (steph * 0) + cellvo,
		      stepw,               cellh);
	    gdk_gc_set_rgb_fg_color(gui.ga_context, &gui.ga_colorcmy[i+11]);
            gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context, TRUE,
		      (stepw * i) + cello, (steph * 1) + cellvo,
		      stepw,               cellh);
	    gdk_gc_set_rgb_fg_color(gui.ga_context, &gui.ga_colorcmy[i+(11*2)]);
            gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context, TRUE,
		      (stepw * i) + cello, (steph * 2) + cellvo,
		      stepw,               cellh);
         }
         break;

      case 12: /* Dark */
         cellw = (((lwidth * 95) / 100) / 19);
         stepw = cellw * 2;
	 cello = (lwidth - (cellw * 19)) / 2;
         cellh = ((lheight * 80) / 100);
	 cellvo = (lheight - cellh) / 2;
	 gdk_gc_set_fill(gui.ga_context, GDK_SOLID);
         for(i = 0; i < 10; i++) {
	    gdk_gc_set_rgb_fg_color(gui.ga_context, &gui.ga_colordgrey[i]);
            gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context, TRUE,
		      (stepw * i) + cello, cellvo,
		      cellw,               cellh);
         }
         break;

      case 13: /* Med */
         cellw = (((lwidth * 95) / 100) / 21);
         stepw = cellw * 2;
	 cello = (lwidth - (cellw * 21)) / 2;
         cellh = ((lheight * 80) / 100);
	 cellvo = (lheight - cellh) / 2;
	 gdk_gc_set_fill(gui.ga_context, GDK_SOLID);
         for(i = 0; i < 11; i++) {
	    gdk_gc_set_rgb_fg_color(gui.ga_context, &gui.ga_colormgrey[i]);
            gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context, TRUE,
		      (stepw * i) + cello, cellvo,
		      cellw,               cellh);
         }
         break;

      case 14: /* Light */
         cellw = (((lwidth * 95) / 100) / 19);
         stepw = cellw * 2;
	 cello = (lwidth - (cellw * 19)) / 2;
         cellh = ((lheight * 80) / 100);
	 cellvo = (lheight - cellh) / 2;
	 gdk_gc_set_fill(gui.ga_context, GDK_SOLID);
         for(i = 0; i < 10; i++) {
	    gdk_gc_set_rgb_fg_color(gui.ga_context, &gui.ga_colorlgrey[i]);
            gdk_draw_rectangle(GTK_LAYOUT(gui.ga_layout)->bin_window,
		      gui.ga_context, TRUE,
		      (stepw * i) + cello, cellvo,
		      cellw,               cellh);
         }

         break;
   }
}

static gboolean
ga_scale_exposed(GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
   ga_update_test_pattern();
   return FALSE;
}

static void
ga_menu_changed(GtkWidget *widget, gpointer data)
{
   gui.ga_current_gfx = gtk_option_menu_get_history(GTK_OPTION_MENU(gui.ga_omenu));

   ga_update_test_pattern();

   switch(gui.ga_current_gfx) {

   case 0:
      gtk_widget_hide(gui.ga_sping);
      gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.ga_eventbox,
		 	"For gamma correction: Currently, the reference gamma is set "
			"to your current CRT1 device's level as read from DDC. Now adjust the "
			"gamma correction so that when you defocus your eyes, "
			"the boxes in the middle vanish.", "");
      break;
   case 1:
      gtk_widget_hide(gui.ga_sping);
      gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.ga_eventbox,
		 	"For gamma correction: Currently, the reference gamma is set "
			"to your current CRT2 device's level as read from DDC. Now adjust the "
			"gamma correction so that when you defocus your eyes, "
			"the boxes in the middle vanish.", "");
      break;
   case 2:
   case 5:
      gtk_widget_hide(gui.ga_sping);
      gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.ga_eventbox,
			"For gamma correction: Currently, the reference gamma is set "
			"to 2.2. This is the standard for most devices, and the basis "
			"of the Internet sRGB colorspace. Now adjust the "
			"gamma correction so that when you defocus your eyes, "
			"the boxes in the middle vanish, or the bars have the same "
			"color.", "");
      break;
   case 4:
   case 7:
      gtk_widget_show(gui.ga_sping);
      gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.ga_eventbox,
			"For gamma correction: Currently, the reference gamma is set "
			"to the value to the left. Now adjust the "
			"gamma correction so that when you defocus your eyes, "
			"the boxes in the middle vanish, or the bars have the same "
			"color.", "");
      break;
   case 3:
   case 6:
      gtk_widget_hide(gui.ga_sping);
      gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.ga_eventbox,
		 	"For gamma correction: Currently, the reference gamma is set "
			"to 1.8. Now adjust the "
			"gamma correction so that when you defocus your eyes, "
			"the boxes in the middle vanish, or the bars have the same "
			"color.", "");
      break;
   case 8:
      gtk_widget_hide(gui.ga_sping);
      gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.ga_eventbox,
		 	"For monitor brightness: Set your monitor's contrast control "
			"to its maximum, and then adjust the monitor's brightness "
			"so that you still see the grey box in the middle while the "
			"white is still bright white.", "");
      break;
   case 9:
   case 10:
   case 11:
      gtk_widget_hide(gui.ga_sping);
      gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.ga_eventbox,
      			"Color gradients for gamma and/or brightness correction", "");
      break;
   case 12:
   case 13:
   case 14:
      gtk_widget_hide(gui.ga_sping);
      gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.ga_eventbox,
      			"For monitor brightness adjustment", "");
      break;
   }
}

static void
ga_sping_changed(GtkWidget *widget, gpointer data)
{
   ga_update_test_pattern();
}

void
gui_gamma_update_gamma_pattern_menu(void)
{
   char buffer[16];
   GtkWidget *label;
   gint backup;

   backup = gtk_option_menu_get_history(GTK_OPTION_MENU(gui.ga_omenu));

   gtk_widget_hide(gui.ga_menuitem1);
   gtk_widget_hide(gui.ga_menuitem2);

   gtk_option_menu_set_history(GTK_OPTION_MENU(gui.ga_omenu), 2);

   sisctrl_gamma_get_monitor_gamma(0);
   sisctrl_gamma_get_monitor_gamma(1);

   if(gl.crt1mongamma != 0.0) {
      sprintf(buffer, "CRT1 %.2f", gl.crt1mongamma);
      label = gtk_bin_get_child(GTK_BIN(gui.ga_menuitem1));
      gtk_label_set_text(GTK_LABEL(label), (gchar *)buffer);
      gtk_widget_show(gui.ga_menuitem1);
   }

   if(gl.crt2mongamma != 0.0) {
      sprintf(buffer, "CRT2 %.2f", gl.crt2mongamma);
      label = gtk_bin_get_child(GTK_BIN(gui.ga_menuitem2));
      gtk_label_set_text(GTK_LABEL(label), (gchar *)buffer);
      gtk_widget_show(gui.ga_menuitem2);
   }

   if( ((backup == 0) && (gl.crt1mongamma != 0.0)) ||
       ((backup == 1) && (gl.crt2mongamma != 0.0)) ||
       (backup >= 2) ) {
      gtk_option_menu_set_history(GTK_OPTION_MENU(gui.ga_omenu), backup);
   } else if((backup == 0) && (gl.crt2mongamma != 0.0)) {
      gtk_option_menu_set_history(GTK_OPTION_MENU(gui.ga_omenu), 1);
   } else if((backup == 1) && (gl.crt1mongamma != 0.0)) {
      gtk_option_menu_set_history(GTK_OPTION_MENU(gui.ga_omenu), 0);
   }
}

/* Build page */

static void
build_gamma_page(void)
{
   int temp, switchpos, i, j;

   gui.ga_frame = gtk_frame_new("Gamma correction");
   gtk_container_set_border_width(GTK_CONTAINER(gui.ga_frame), 8);
   gtk_widget_show(gui.ga_frame);

   gui.ga_table = gtk_table_new(6, 8, FALSE);
   gtk_table_set_row_spacings(GTK_TABLE(gui.ga_table), 2);
   gtk_table_set_col_spacings(GTK_TABLE(gui.ga_table), 2);

   gtk_container_set_border_width(GTK_CONTAINER(gui.ga_table), 10);

   gtk_container_add(GTK_CONTAINER(gui.ga_frame), gui.ga_table);
   gtk_widget_show(gui.ga_table);

   temp = 2; switchpos = 2;

   switch(gl.gamma_page_type) {

   case GAMMA_PAGE_TYPE_XINERAMA:
      {
         char buffer[64];

	 gui.ga_subtable = gtk_table_new(temp, 4, FALSE);
         gtk_table_set_row_spacings(GTK_TABLE(gui.ga_subtable), 0);
         gtk_table_set_col_spacings(GTK_TABLE(gui.ga_subtable), 5);
         gtk_container_set_border_width(GTK_CONTAINER(gui.ga_subtable), 0);

	 sprintf(buffer, "Setting for Screen %d", gl.screennum);
         gui.ga_label_enable = gtk_label_new(buffer);

         gtk_table_attach_defaults(GTK_TABLE(gui.ga_subtable), gui.ga_label_enable, 0, 4, 0, 1);
         gtk_widget_show(gui.ga_label_enable);
      }
      break;

   default:
      if(!gl.havesepgamma2) {
         /* gui.ga_plabel2 = gtk_label_new(" ");
         gtk_table_attach_defaults(GTK_TABLE(gui.ga_table), gui.ga_plabel2, 1, 4, 1, 2);
         gtk_widget_show(gui.ga_plabel2); */
      } else {
         temp = 3; switchpos = 3;
      }

      gui.ga_subtable = gtk_table_new(temp, 4, FALSE);
      gtk_table_set_row_spacings(GTK_TABLE(gui.ga_subtable), 0);
      gtk_table_set_col_spacings(GTK_TABLE(gui.ga_subtable), 5);
      gtk_container_set_border_width(GTK_CONTAINER(gui.ga_subtable), 0);

      gui.ga_label_enable = gtk_label_new("Enable for");
      gtk_table_attach_defaults(GTK_TABLE(gui.ga_subtable), gui.ga_label_enable, 0, 2, 0, 1);
      gtk_widget_show(gui.ga_label_enable);

      gui.ga_check1 = gtk_check_button_new_with_label("CRT1");
      gtk_table_attach_defaults(GTK_TABLE(gui.ga_subtable), gui.ga_check1, 2, 3, 0, 1);
      gtk_widget_show(gui.ga_check1);

      gui.ga_check2 = gtk_check_button_new_with_label("CRT2");
      gtk_table_attach_defaults(GTK_TABLE(gui.ga_subtable), gui.ga_check2, 3, 4, 0, 1);
      gtk_widget_show(gui.ga_check2);

      gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.ga_check1,
       		 	"Enable gamma correction for first output (CRT1)", "");
      gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.ga_check2,
       		 	"Enable gamma correction for second output (CRT2)", "");

      if(gl.havesepgamma2) {
         gui.ga_check3 = gtk_check_button_new_with_label("Separate");
         gtk_table_attach_defaults(GTK_TABLE(gui.ga_subtable), gui.ga_check3, 1, 2, 1, 2);
         gtk_widget_show(gui.ga_check3);

	 gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.ga_check3,
		 	"Adjust gamma correction separately for first and second output", "");

	 gui.ga_radio_crt1 = gtk_radio_button_new_with_label(NULL, "CRT1");
	 gui.ga_group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(gui.ga_radio_crt1));
	 gui.ga_radio_crt2 = gtk_radio_button_new_with_label(gui.ga_group, "CRT2");

         gtk_table_attach_defaults(GTK_TABLE(gui.ga_subtable), gui.ga_radio_crt1, 2, 3, 1, 2);
         gtk_table_attach_defaults(GTK_TABLE(gui.ga_subtable), gui.ga_radio_crt2, 3, 4, 1, 2);

	 gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(gui.ga_radio_crt1), TRUE);

	 gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.ga_radio_crt1,
		 	"Adjust gamma correction for first output (CRT1)", "");
	 gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.ga_radio_crt2,
		 	"Adjust gamma correction for second output (CRT2)", "");
      }

      sisctrl_gamma_set_status();

      if(gl.gammaenablecrt1) {
         gtk_widget_set_sensitive(gui.ga_check1, TRUE);
	 g_signal_connect(G_OBJECT(gui.ga_check1), "toggled", G_CALLBACK (ga_check1_toggled), NULL);
      } else {
         gtk_widget_set_sensitive(gui.ga_check1, FALSE);
      }

      if(gl.gammaenablecrt2) {
         gtk_widget_set_sensitive(gui.ga_check2, TRUE);
	 g_signal_connect(G_OBJECT(gui.ga_check2), "toggled", G_CALLBACK (ga_check2_toggled), NULL);
      } else {
         gtk_widget_set_sensitive(gui.ga_check2, FALSE);
      }

      if(gl.havesepgamma2) {
	 g_signal_connect(G_OBJECT(gui.ga_check3), "toggled", G_CALLBACK (ga_check3_toggled), NULL);
	 g_signal_connect(G_OBJECT(gui.ga_radio_crt1), "toggled", G_CALLBACK(ga_radio_changed), NULL);
      }

      break;
   }

   gui.ga_label_switch = gtk_label_new("Adjust");
   gtk_table_attach_defaults(GTK_TABLE(gui.ga_subtable), gui.ga_label_switch, 0, 2, switchpos, switchpos+1);
   gtk_widget_show(gui.ga_label_switch);

   gui.ga_radio_rgb = gtk_radio_button_new_with_label(NULL, "R/G/B");
   gui.ga_group_rgb = gtk_radio_button_get_group(GTK_RADIO_BUTTON(gui.ga_radio_rgb));
   gui.ga_radio_all = gtk_radio_button_new_with_label(gui.ga_group_rgb, "All");

   gtk_table_attach_defaults(GTK_TABLE(gui.ga_subtable), gui.ga_radio_rgb, 2, 3, switchpos, switchpos+1);
   gtk_table_attach_defaults(GTK_TABLE(gui.ga_subtable), gui.ga_radio_all, 3, 4, switchpos, switchpos+1);

   gtk_widget_show(gui.ga_radio_rgb);
   gtk_widget_show(gui.ga_radio_all);

   gui.ga_current_adj = 0;
   gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(gui.ga_radio_rgb), TRUE);

   gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.ga_radio_rgb,
	 	"Adjust gamma correction for red, green and blue channels", "");
   gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.ga_radio_all,
	 	"Adjust gamma correction for all color channels simulantiously", "");

   g_signal_connect(G_OBJECT(gui.ga_radio_rgb), "toggled", G_CALLBACK(ga_radio_rgb_changed), NULL);

   gtk_table_attach_defaults(GTK_TABLE(gui.ga_table), gui.ga_subtable, 1, 6, 0 /*1*/, 2);

   gtk_widget_show(gui.ga_subtable);

   gui.ga_icon = gdk_pixmap_create_from_xpm_d(gui.window->window, &gui.ga_mask, NULL, gammaicon_xpm);
   gui.ga_image = gtk_image_new_from_pixmap(gui.ga_icon, gui.ga_mask);
   gtk_table_attach_defaults(GTK_TABLE(gui.ga_table), gui.ga_image, 0, 1, 0, 4);
   gtk_widget_show(gui.ga_image);

   gui.ga_label_gamma = gtk_label_new("Gamma");
   gtk_table_attach_defaults(GTK_TABLE(gui.ga_table), gui.ga_label_gamma, 1, 2, 2, 3);
   gtk_widget_show(gui.ga_label_gamma);
   gui.ga_label_bri = gtk_label_new("Brightness");
   gtk_table_attach_defaults(GTK_TABLE(gui.ga_table), gui.ga_label_bri, 1, 2, 3, 4);
   gtk_widget_show(gui.ga_label_bri);

   gui.ga_label_red = gtk_label_new("R");
   gtk_table_attach_defaults(GTK_TABLE(gui.ga_table), gui.ga_label_red, 2, 3, 2, 3);
   gtk_widget_show(gui.ga_label_red);
   gui.ga_label_green = gtk_label_new("G");
   gtk_table_attach_defaults(GTK_TABLE(gui.ga_table), gui.ga_label_green, 4, 5, 2, 3);
   gtk_widget_show(gui.ga_label_green);
   gui.ga_label_blue = gtk_label_new("B");
   gtk_table_attach_defaults(GTK_TABLE(gui.ga_table), gui.ga_label_blue, 6, 7, 2, 3);
   gtk_widget_show(gui.ga_label_blue);
   gui.ga_label_red1 = gtk_label_new("R");
   gtk_table_attach_defaults(GTK_TABLE(gui.ga_table), gui.ga_label_red1, 2, 3, 3, 4);
   gtk_widget_show(gui.ga_label_red1);
   gui.ga_label_green1 = gtk_label_new("G");
   gtk_table_attach_defaults(GTK_TABLE(gui.ga_table), gui.ga_label_green1, 4, 5, 3, 4);
   gtk_widget_show(gui.ga_label_green1);
   gui.ga_label_blue1 = gtk_label_new("B");
   gtk_table_attach_defaults(GTK_TABLE(gui.ga_table), gui.ga_label_blue1, 6, 7, 3, 4);
   gtk_widget_show(gui.ga_label_blue1);

   gui.ga_spin1_adj = (GtkAdjustment *)gtk_adjustment_new(1.0, 0.1, 10.0, 0.1, 1.0, 0.0);
   gui.ga_spin2_adj = (GtkAdjustment *)gtk_adjustment_new(1.0, 0.1, 10.0, 0.1, 1.0, 0.0);
   gui.ga_spin3_adj = (GtkAdjustment *)gtk_adjustment_new(1.0, 0.1, 10.0, 0.1, 1.0, 0.0);
   gui.ga_spin4_adj = (GtkAdjustment *)gtk_adjustment_new(1.0, 0.1, 10.0, 0.01, 0.1, 0.0);
   gui.ga_spin5_adj = (GtkAdjustment *)gtk_adjustment_new(1.0, 0.1, 10.0, 0.01, 0.1, 0.0);
   gui.ga_spin6_adj = (GtkAdjustment *)gtk_adjustment_new(1.0, 0.1, 10.0, 0.01, 0.1, 0.0);

   gui.ga_spin1 = gtk_spin_button_new(gui.ga_spin1_adj, 0.1, 3);
   gui.ga_spin2 = gtk_spin_button_new(gui.ga_spin2_adj, 0.1, 3);
   gui.ga_spin3 = gtk_spin_button_new(gui.ga_spin3_adj, 0.1, 3);
   gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.ga_spin1,
		 	"Adjust red channel gamma", "");
   gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.ga_spin2,
		 	"Adjust green channel gamma", "");
   gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.ga_spin3,
		 	"Adjust blue channel gamma", "");

   gui.ga_spin4 = gtk_spin_button_new(gui.ga_spin4_adj, 0.1, 3);
   gui.ga_spin5 = gtk_spin_button_new(gui.ga_spin5_adj, 0.1, 3);
   gui.ga_spin6 = gtk_spin_button_new(gui.ga_spin6_adj, 0.1, 3);
   gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.ga_spin4,
		 	"Adjust red channel gamma brightness", "");
   gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.ga_spin5,
		 	"Adjust green channel gamma brightness", "");
   gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.ga_spin6,
		 	"Adjust blue channel gamma brightness", "");

   gtk_table_attach_defaults(GTK_TABLE(gui.ga_table), gui.ga_spin1, 3, 4, 2, 3);
   gtk_table_attach_defaults(GTK_TABLE(gui.ga_table), gui.ga_spin2, 5, 6, 2, 3);
   gtk_table_attach_defaults(GTK_TABLE(gui.ga_table), gui.ga_spin3, 7, 8, 2, 3);
   gtk_table_attach_defaults(GTK_TABLE(gui.ga_table), gui.ga_spin4, 3, 4, 3, 4);
   gtk_table_attach_defaults(GTK_TABLE(gui.ga_table), gui.ga_spin5, 5, 6, 3, 4);
   gtk_table_attach_defaults(GTK_TABLE(gui.ga_table), gui.ga_spin6, 7, 8, 3, 4);

   sisctrl_gamma_set_all_spinvalues();

   gtk_widget_show(gui.ga_spin1);
   gtk_widget_show(gui.ga_spin2);
   gtk_widget_show(gui.ga_spin3);
   gtk_widget_show(gui.ga_spin4);
   gtk_widget_show(gui.ga_spin5);
   gtk_widget_show(gui.ga_spin6);

   g_signal_connect(G_OBJECT(gui.ga_spin1_adj), "value_changed", G_CALLBACK(ga_spin1_changed), NULL);
   g_signal_connect(G_OBJECT(gui.ga_spin2_adj), "value_changed", G_CALLBACK(ga_spin2_changed), NULL);
   g_signal_connect(G_OBJECT(gui.ga_spin3_adj), "value_changed", G_CALLBACK(ga_spin3_changed), NULL);
   g_signal_connect(G_OBJECT(gui.ga_spin4_adj), "value_changed", G_CALLBACK(ga_spin4_changed), NULL);
   g_signal_connect(G_OBJECT(gui.ga_spin5_adj), "value_changed", G_CALLBACK(ga_spin5_changed), NULL);
   g_signal_connect(G_OBJECT(gui.ga_spin6_adj), "value_changed", G_CALLBACK(ga_spin6_changed), NULL);

   gui.ga_drawable = gtk_drawing_area_new();
   gtk_widget_set_size_request(gui.ga_drawable, GACW, GACH);
   gui.ga_vbox = gtk_vbox_new(FALSE, 0);
   gui.ga_hbox = gtk_hbox_new(FALSE, 0);
   gtk_box_pack_start(GTK_BOX(gui.ga_vbox), gui.ga_hbox, FALSE, FALSE, 0);
   gtk_box_pack_end(GTK_BOX(gui.ga_hbox), gui.ga_drawable, FALSE, FALSE, 0);
   gtk_table_attach_defaults(GTK_TABLE(gui.ga_table), gui.ga_vbox, 6, 8, 0, 2);
   gtk_widget_show(gui.ga_drawable);
   gtk_widget_show(gui.ga_hbox);
   gtk_widget_show(gui.ga_vbox);
   g_signal_connect(G_OBJECT(gui.ga_drawable), "expose-event", GTK_SIGNAL_FUNC(ga_picture_exposed), NULL);

   /* RGB */
   for(i = 0, j = 0; i < 16; i++, j+= (65536 / 16)) {
      gui.ga_colorrgb[0*16 + i].red = j;
      gui.ga_colorrgb[0*16 + i].green = gui.ga_colorrgb[0*16 + i].blue = 0;
      gui.ga_colorrgb[1*16 + i].green = j;
      gui.ga_colorrgb[1*16 + i].red = gui.ga_colorrgb[1*16 + i].blue = 0;
      gui.ga_colorrgb[2*16 + i].blue = j;
      gui.ga_colorrgb[2*16 + i].red = gui.ga_colorrgb[2*16 + i].green = 0;
   }

   /* CMY */
   for(i = 0; i < 11; i++) {
      gui.ga_colorcmy[i].red   = cmyc[(i*3) + 0] << 8;
      gui.ga_colorcmy[i].green = cmyc[(i*3) + 1] << 8;
      gui.ga_colorcmy[i].blue  = cmyc[(i*3) + 2] << 8;
      gui.ga_colorcmy[i + 11].red   = cmym[(i*3) + 0] << 8;
      gui.ga_colorcmy[i + 11].green = cmym[(i*3) + 1] << 8;
      gui.ga_colorcmy[i + 11].blue  = cmym[(i*3) + 2] << 8;
      gui.ga_colorcmy[i + (2*11)].red   = cmyy[(i*3) + 0] << 8;
      gui.ga_colorcmy[i + (2*11)].green = cmyy[(i*3) + 1] << 8;
      gui.ga_colorcmy[i + (2*11)].blue  = cmyy[(i*3) + 2] << 8;
   }

   /* Grey */
   for(i = 0; i < 20; i++) {
      gui.ga_colorgrey[i].red =
        gui.ga_colorgrey[i].green =
	  gui.ga_colorgrey[i].blue = (int)((float)((float)(256.0 / 20.0) * (float)(i))) << 8;
   }

   /* DGrey, LGrey */
   for(i = 0; i < 10; i++) {
      gui.ga_colordgrey[i].red =
        gui.ga_colordgrey[i].green =
	  gui.ga_colordgrey[i].blue = dgrey[i] << 8;
      gui.ga_colorlgrey[i].red =
        gui.ga_colorlgrey[i].green =
	  gui.ga_colorlgrey[i].blue = lgrey[i] << 8;
   }

   /* MGrey */
   for(i = 0; i < 11; i++) {
      gui.ga_colormgrey[i].red =
        gui.ga_colormgrey[i].green =
	  gui.ga_colormgrey[i].blue = mgrey[i] << 8;
   }

   {
      GdkPixmap *pseudo;
      pseudo = gdk_pixmap_create_from_xpm_d(gui.window->window, &gui.ga_stripple, NULL, gammastipple);
      pseudo = gdk_pixmap_create_from_xpm_d(gui.window->window, &gui.ga_stripple2, NULL, gammastipple2);
   }

   gui.ga_current_gfx = 0;

   gui.ga_layout = gtk_layout_new(NULL, NULL);
   gtk_widget_set_size_request(gui.ga_layout, 0, 48);
   gui.ga_eventbox = gtk_event_box_new();
   gtk_widget_set_size_request(gui.ga_eventbox, 0, 48);
   gtk_container_set_border_width(GTK_CONTAINER(gui.ga_eventbox), 0);
   gtk_container_add(GTK_CONTAINER(gui.ga_eventbox), gui.ga_layout);
   gtk_table_attach_defaults(GTK_TABLE(gui.ga_table), gui.ga_eventbox, 1, 8, 4, 6);
   gtk_widget_show(gui.ga_eventbox);
   gtk_widget_show(gui.ga_layout);

   gui.ga_omenu = gtk_option_menu_new();

   gui.ga_menu = gtk_menu_new();

   gui.ga_menuitem1  = gtk_menu_item_new_with_label("CRT1");
   gui.ga_menuitem2  = gtk_menu_item_new_with_label("CRT2");
   gui.ga_menuitem3  = gtk_menu_item_new_with_label("Box 2.2");
   gui.ga_menuitem4  = gtk_menu_item_new_with_label("Box 1.8");
   gui.ga_menuitem5  = gtk_menu_item_new_with_label("Box custom");
   gui.ga_menuitem6  = gtk_menu_item_new_with_label("Bar 2.2");
   gui.ga_menuitem7  = gtk_menu_item_new_with_label("Bar 1.8");
   gui.ga_menuitem8  = gtk_menu_item_new_with_label("Bar custom");
   gui.ga_menuitem9  = gtk_menu_item_new_with_label("Brightness");
   gui.ga_menuitem10 = gtk_menu_item_new_with_label("RGB");
   gui.ga_menuitem11 = gtk_menu_item_new_with_label("Grey");
   gui.ga_menuitem12 = gtk_menu_item_new_with_label("CMY");
   gui.ga_menuitem13 = gtk_menu_item_new_with_label("Dark");
   gui.ga_menuitem14 = gtk_menu_item_new_with_label("Medium");
   gui.ga_menuitem15 = gtk_menu_item_new_with_label("Light");

   gtk_menu_shell_append(GTK_MENU_SHELL(gui.ga_menu), gui.ga_menuitem1);
   gtk_menu_shell_append(GTK_MENU_SHELL(gui.ga_menu), gui.ga_menuitem2);
   gtk_menu_shell_append(GTK_MENU_SHELL(gui.ga_menu), gui.ga_menuitem3);
   gtk_menu_shell_append(GTK_MENU_SHELL(gui.ga_menu), gui.ga_menuitem4);
   gtk_menu_shell_append(GTK_MENU_SHELL(gui.ga_menu), gui.ga_menuitem5);
   gtk_menu_shell_append(GTK_MENU_SHELL(gui.ga_menu), gui.ga_menuitem6);
   gtk_menu_shell_append(GTK_MENU_SHELL(gui.ga_menu), gui.ga_menuitem7);
   gtk_menu_shell_append(GTK_MENU_SHELL(gui.ga_menu), gui.ga_menuitem8);
   gtk_menu_shell_append(GTK_MENU_SHELL(gui.ga_menu), gui.ga_menuitem9);
   gtk_menu_shell_append(GTK_MENU_SHELL(gui.ga_menu), gui.ga_menuitem10);
   gtk_menu_shell_append(GTK_MENU_SHELL(gui.ga_menu), gui.ga_menuitem11);
   gtk_menu_shell_append(GTK_MENU_SHELL(gui.ga_menu), gui.ga_menuitem12);
   gtk_menu_shell_append(GTK_MENU_SHELL(gui.ga_menu), gui.ga_menuitem13);
   gtk_menu_shell_append(GTK_MENU_SHELL(gui.ga_menu), gui.ga_menuitem14);
   gtk_menu_shell_append(GTK_MENU_SHELL(gui.ga_menu), gui.ga_menuitem15);

   gtk_option_menu_set_menu(GTK_OPTION_MENU(gui.ga_omenu), gui.ga_menu);

   gtk_widget_show(gui.ga_menuitem1);
   gtk_widget_show(gui.ga_menuitem2);
   gtk_widget_show(gui.ga_menuitem3);
   gtk_widget_show(gui.ga_menuitem4);
   gtk_widget_show(gui.ga_menuitem5);
   gtk_widget_show(gui.ga_menuitem6);
   gtk_widget_show(gui.ga_menuitem7);
   gtk_widget_show(gui.ga_menuitem8);
   gtk_widget_show(gui.ga_menuitem9);
   gtk_widget_show(gui.ga_menuitem10);
   gtk_widget_show(gui.ga_menuitem11);
   gtk_widget_show(gui.ga_menuitem12);
   gtk_widget_show(gui.ga_menuitem13);
   gtk_widget_show(gui.ga_menuitem14);
   gtk_widget_show(gui.ga_menuitem15);

   gtk_widget_show(gui.ga_omenu);

   gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.ga_omenu,
		 	"Select test pattern", "");

   gtk_table_attach(GTK_TABLE(gui.ga_table), gui.ga_omenu, 0, 1, 4, 5, GTK_FILL|GTK_EXPAND, 0, 0, 0);

   g_signal_connect(G_OBJECT(gui.ga_omenu), "changed", G_CALLBACK(ga_menu_changed), NULL);

   gui.ga_sping_adj = (GtkAdjustment *)gtk_adjustment_new(2.2, 1.0, 3.55, 0.01, 0.1, 0.0);
   gui.ga_sping = gtk_spin_button_new(gui.ga_sping_adj, 0.1, 2);
   gtk_table_attach_defaults(GTK_TABLE(gui.ga_table), gui.ga_sping, 0, 1, 5, 6);

   g_signal_connect(G_OBJECT(gui.ga_sping_adj), "value_changed", G_CALLBACK(ga_sping_changed), NULL);

   gui_gamma_update_gamma_pattern_menu();

   g_signal_connect(G_OBJECT(gui.ga_layout), "expose-event", GTK_SIGNAL_FUNC(ga_scale_exposed), NULL);

   gui.ga_label = xpm_label_box(icon_gamma, "Gamma", NULL, 1, "Gamma correction", "");
   gtk_notebook_append_page(GTK_NOTEBOOK(gui.notebook), gui.ga_frame, gui.ga_label);
}


/**************************************************
 *                     TV page                    *
 **************************************************/

void
gui_tv_antiflicker_enable_and_set_menu_index(int enable, int index)
{
   if(gl.tv_page_subtype != TV_PAGE_SUBTYPE_SIS) return;

   gtk_widget_set_sensitive(gui.tp_omenu, enable ? TRUE : FALSE);
   gtk_option_menu_set_history(GTK_OPTION_MENU(gui.tp_omenu), (gint)index);
}

void
gui_tv_colorcalib_c_enable_and_set_value(int enable, double value)
{
   GtkAdjustment *myadj;

   if(gl.tv_page_subtype != TV_PAGE_SUBTYPE_SIS) return;

   myadj = gtk_range_get_adjustment(GTK_RANGE(gui.tp_range_coc));

   if(!enable) {
      gtk_widget_set_sensitive(gui.tp_range_coc, FALSE);
   } else {
      gtk_widget_set_sensitive(gui.tp_range_coc, TRUE);
      gtk_adjustment_set_value(myadj, value);
      g_signal_emit_by_name(G_OBJECT(myadj), "changed");
   }

}

void
gui_tv_colorcalib_f_enable_and_set_value(int enable, double value)
{
   GtkAdjustment *myadj;

   if(gl.tv_page_subtype != TV_PAGE_SUBTYPE_SIS) return;

   myadj = gtk_range_get_adjustment(GTK_RANGE(gui.tp_range_cof));

   if(!enable) {
      gtk_widget_set_sensitive(gui.tp_range_cof, FALSE);
   } else {
      gtk_widget_set_sensitive(gui.tp_range_cof, TRUE);
      gtk_adjustment_set_value(myadj, value);
      g_signal_emit_by_name(G_OBJECT(myadj), "changed");
   }
}

void
gui_tv_saturation_enable_and_set_value(int enable, double value)
{
   GtkAdjustment *myadj;

   if(gl.tv_page_subtype != TV_PAGE_SUBTYPE_SIS) return;
   if(!gl.supporttvsaturation) return;

   myadj = gtk_range_get_adjustment(GTK_RANGE(gui.tp_range_sa));

   if(!enable) {
      gtk_widget_set_sensitive(gui.tp_range_sa, FALSE);
   } else {
      gtk_widget_set_sensitive(gui.tp_range_sa, TRUE);
      gtk_adjustment_set_value(myadj, value);
      g_signal_emit_by_name(G_OBJECT(myadj), "changed");
   }
}

void
gui_tv_edgeenhance_enable_and_set_value(int enable, double value)
{
   GtkAdjustment *myadj;

   if(gl.tv_page_subtype != TV_PAGE_SUBTYPE_SIS) return;
   if(!gl.supporttvsaturation) return;

   myadj = gtk_range_get_adjustment(GTK_RANGE(gui.tp_range_ee));

   if(!enable) {
      gtk_widget_set_sensitive(gui.tp_range_ee, FALSE);
   } else {
      gtk_widget_set_sensitive(gui.tp_range_ee, TRUE);
      gtk_adjustment_set_value(myadj, value);
      g_signal_emit_by_name(G_OBJECT(myadj), "changed");
   }
}

void
gui_tv_filter_c_enable_and_set_value(int enable, int value)
{
   if(gl.tv_page_subtype != TV_PAGE_SUBTYPE_SIS) return;

   if(!enable) {
      gtk_widget_set_sensitive(gui.tp_check_cfi, FALSE);
   } else {
      gtk_widget_set_sensitive(gui.tp_check_cfi, TRUE);
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(gui.tp_check_cfi), value ? TRUE : FALSE);
   }
}

void
gui_tv_filter_y_enable_and_set_value(int enable, int index)
{
    if(gl.tv_page_subtype != TV_PAGE_SUBTYPE_SIS) return;

    if(!enable) {
       gtk_widget_set_sensitive(gui.tp_omenuy, FALSE);
       gtk_option_menu_set_history(GTK_OPTION_MENU(gui.tp_omenuy), 1);
    } else {
       gtk_widget_set_sensitive(gui.tp_omenuy, TRUE);
       gtk_option_menu_set_history(GTK_OPTION_MENU(gui.tp_omenuy), index);
    }
}

void
gui_tv_chrontel_slider_set_value(int slider, int enable, double value)
{
    GtkWidget *widget;
    GtkAdjustment *myadj;

    if(gl.tv_page_subtype != TV_PAGE_SUBTYPE_CHRONTEL) return;

    switch(slider) {
    case TV_CH_SLI_CONTRAST:
       widget = gui.tp_range_co;
       break;
    case TV_CH_SLI_TEXTENHANCE:
       widget = gui.tp_range_te;
       break;
    case TV_CH_SLI_CHROMAFF:
       widget = gui.tp_range_cf;
       break;
    case TV_CH_SLI_LUMAFF:
       widget = gui.tp_range_lf;
       break;
    default:
       return;
    }

    myadj = gtk_range_get_adjustment(GTK_RANGE(widget));

    if(!enable) {
       gtk_widget_set_sensitive(widget, FALSE);
    } else {
       gtk_widget_set_sensitive(widget, TRUE);
       gtk_adjustment_set_value(myadj, value);
       g_signal_emit_by_name(G_OBJECT(myadj), "changed");
    }
}

void
gui_tv_chrontel_cvbscolor_set_value(int enable, int val)
{
    if(gl.tv_page_subtype != TV_PAGE_SUBTYPE_CHRONTEL) return;

    if(!enable) {
      gtk_widget_set_sensitive(gui.tp_check_col, FALSE);
   } else {
      gtk_widget_set_sensitive(gui.tp_check_col, TRUE);
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(gui.tp_check_col), val ? TRUE : FALSE);
   }
}

/* Signal handlers */

static void
tv_de_clicked(GtkWidget *widget, gpointer data)
{
   sisctrl_tv_de_clicked();
}

static void
tv_up_clicked(GtkWidget *widget, gpointer data)
{
   sisctrl_tv_up_clicked();
}

static void
tv_dn_clicked(GtkWidget *widget, gpointer data)
{
   sisctrl_tv_dn_clicked();
}

static void
tv_le_clicked(GtkWidget *widget, gpointer data)
{
   sisctrl_tv_le_clicked();
}

static void
tv_ri_clicked(GtkWidget *widget, gpointer data)
{
   sisctrl_tv_ri_clicked();
}

static void
tv_xm_clicked(GtkWidget *widget, gpointer data)
{
   sisctrl_tv_xm_clicked();
}

static void
tv_xp_clicked(GtkWidget *widget, gpointer data)
{
   sisctrl_tv_xp_clicked();
}

static void
tv_ym_clicked(GtkWidget *widget, gpointer data)
{
   sisctrl_tv_ym_clicked();
}

static void
tv_yp_clicked(GtkWidget *widget, gpointer data)
{
   sisctrl_tv_yp_clicked();
}

static void
tp_af_changed(GtkWidget *widget, gpointer data)
{
   int index = gtk_option_menu_get_history(GTK_OPTION_MENU(gui.tp_omenu));

   sisctrl_tp_af_changed(index);
}

static void
tp_sa_changed(GtkWidget *widget, gpointer data)
{
   double myval = gtk_adjustment_get_value(GTK_ADJUSTMENT(widget));

   sisctrl_tp_sa_changed(myval);
}

static void
tp_ee_changed(GtkWidget *widget, gpointer data)
{
   double myval = gtk_adjustment_get_value(GTK_ADJUSTMENT(widget));

   sisctrl_tp_ee_changed(myval);
}

static void
tp_check_cfi_toggled(GtkWidget *widget, gpointer data)
{
    int myval = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget)) ? 1 : 0;

    sisctrl_tp_check_cfi_toggled(myval);
}

static void
tp_yfi_changed(GtkWidget *widget, gpointer data)
{
    int index = gtk_option_menu_get_history(GTK_OPTION_MENU(gui.tp_omenuy));

    sisctrl_tp_yfi_changed(index);
}

static void
tp_color_calib_changed(void)
{
    GtkAdjustment *myadj1 = gtk_range_get_adjustment(GTK_RANGE(gui.tp_range_coc));
    double myval1 = gtk_adjustment_get_value(myadj1);
    GtkAdjustment *myadj2 = gtk_range_get_adjustment(GTK_RANGE(gui.tp_range_cof));
    double myval2 = gtk_adjustment_get_value(myadj2);

    sisctrl_tp_col_calib_changed(myval1, myval2);
}

static void
tp_coc_changed(GtkWidget *widget, gpointer data)
{
   tp_color_calib_changed();
}

static void
tp_cof_changed(GtkWidget *widget, gpointer data)
{
   tp_color_calib_changed();
}

static void
tp_co_changed(GtkWidget *widget, gpointer data)
{
   double myval = gtk_adjustment_get_value(GTK_ADJUSTMENT(widget));

   sisctrl_tp_co_changed(myval);
}

static void
tp_te_changed(GtkWidget *widget, gpointer data)
{
   double myval = gtk_adjustment_get_value(GTK_ADJUSTMENT(widget));

   sisctrl_tp_te_changed(myval);
}

static void
tp_cf_changed(GtkWidget *widget, gpointer data)
{
   double myval = gtk_adjustment_get_value(GTK_ADJUSTMENT(widget));

   sisctrl_tp_cf_changed(myval);
}

static void
tp_lf_changed(GtkWidget *widget, gpointer data)
{
   double myval = gtk_adjustment_get_value(GTK_ADJUSTMENT(widget));

   sisctrl_tp_lf_changed(myval);
}

static void
tp_check_col_toggled(GtkWidget *widget, gpointer data)
{
   int myval = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget)) ? 1 : 0;

   sisctrl_tp_check_col_toggled(myval);
}

/* Build page */

static void
build_tv_page(void)
{
   int rows, eepos, colcpos, colfpos, filtpos;

   switch(gl.tv_page_type) {

   case TV_PAGE_TYPE_NORMAL:

      /* gui.tp_box = gtk_hbox_new(FALSE, 3); */
      gui.tp_box = gtk_table_new(1, 2, FALSE);
      gtk_table_set_row_spacings(GTK_TABLE(gui.tp_box), 0);
      gtk_table_set_col_spacings(GTK_TABLE(gui.tp_box), 0);
      gtk_container_set_border_width(GTK_CONTAINER(gui.tp_box), 0);
      gtk_widget_show(gui.tp_box);

      gui.tp_frame = gtk_frame_new("TV settings");
      gtk_container_set_border_width(GTK_CONTAINER(gui.tp_frame), 8);
      gtk_widget_show(gui.tp_frame);

      /* gtk_box_pack_start(GTK_BOX(gui.tp_box), gui.tp_frame, FALSE, FALSE, 0); */
      gtk_table_attach_defaults(GTK_TABLE(gui.tp_box), gui.tp_frame, 0, 1, 0, 1);

      switch(gl.tv_page_subtype) {

      case TV_PAGE_SUBTYPE_SIS:
         rows = 4; eepos = 1; colcpos = 1; colfpos = 2; filtpos = 3;

	 if(gl.supporttvsaturation) {
	    rows++; eepos++; colcpos++; colfpos++; filtpos++;
	 }

	 if(gl.supporttvedgeenhance) {
	    rows++; colcpos++; colfpos++; filtpos++;
	 }

         gui.tp_table = gtk_table_new(rows, 3, TRUE);
         gtk_table_set_row_spacings(GTK_TABLE(gui.tp_table), 0);
         gtk_table_set_col_spacings(GTK_TABLE(gui.tp_table), 13);

         gtk_container_set_border_width(GTK_CONTAINER(gui.tp_table), 4);

	 gtk_container_add(GTK_CONTAINER(gui.tp_frame), gui.tp_table);
         gtk_widget_show(gui.tp_table);

         gui.tp_label_af = gtk_label_new("Antiflicker");
         gtk_widget_show(gui.tp_label_af);
         gtk_table_attach_defaults(GTK_TABLE(gui.tp_table), gui.tp_label_af, 0, 1, 0, 1);

         gui.tp_omenu = gtk_option_menu_new();

         gui.tp_menu = gtk_menu_new();

         gui.tp_menuitem1 = gtk_menu_item_new_with_label("Off");
         gui.tp_menuitem2 = gtk_menu_item_new_with_label("Low");
         gui.tp_menuitem3 = gtk_menu_item_new_with_label("Medium");
         gui.tp_menuitem4 = gtk_menu_item_new_with_label("High");
         gui.tp_menuitem5 = gtk_menu_item_new_with_label("Adaptive");

         gtk_menu_shell_append(GTK_MENU_SHELL(gui.tp_menu), gui.tp_menuitem1);
         gtk_menu_shell_append(GTK_MENU_SHELL(gui.tp_menu), gui.tp_menuitem2);
         gtk_menu_shell_append(GTK_MENU_SHELL(gui.tp_menu), gui.tp_menuitem3);
         gtk_menu_shell_append(GTK_MENU_SHELL(gui.tp_menu), gui.tp_menuitem4);
         gtk_menu_shell_append(GTK_MENU_SHELL(gui.tp_menu), gui.tp_menuitem5);

         gtk_widget_show(gui.tp_menuitem1);
         gtk_widget_show(gui.tp_menuitem2);
         gtk_widget_show(gui.tp_menuitem3);
         gtk_widget_show(gui.tp_menuitem4);
         gtk_widget_show(gui.tp_menuitem5);

         gtk_option_menu_set_menu(GTK_OPTION_MENU(gui.tp_omenu), gui.tp_menu);

         sisctrl_tv_antiflicker_set_history();

         gtk_widget_show(gui.tp_omenu);

         /* gtk_table_attach_defaults(GTK_TABLE(gui.tp_table), gui.tp_omenu, 1, 3, 0, 1); */
	 gtk_table_attach(GTK_TABLE(gui.tp_table), gui.tp_omenu, 1, 3, 0, 1, GTK_FILL|GTK_EXPAND, 0, 0, 0);

         g_signal_connect(G_OBJECT(gui.tp_omenu), "changed", G_CALLBACK(tp_af_changed), NULL);

	 if(gl.supporttvsaturation) {
            gui.tp_label_sa = gtk_label_new("Saturation");
            gtk_widget_show(gui.tp_label_sa);
            gtk_table_attach_defaults(GTK_TABLE(gui.tp_table), gui.tp_label_sa, 0, 1, 1, 2);

            gui.tp_range_sa = gtk_hscale_new_with_range(0, 15, 1);
	    gtk_scale_set_value_pos(GTK_SCALE(gui.tp_range_sa), GTK_POS_RIGHT);
            gtk_widget_show(gui.tp_range_sa);
            sisctrl_tv_saturation_set_value();
            gtk_table_attach_defaults(GTK_TABLE(gui.tp_table), gui.tp_range_sa, 1, 3, 1, 2);

	    {
	    GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(gui.tp_range_sa));
            g_signal_connect(G_OBJECT(myadj), "value_changed", G_CALLBACK(tp_sa_changed), NULL);
	    }
         }

	 if(gl.supporttvedgeenhance) {
            gui.tp_label_ee = gtk_label_new("Edge enhance");
            gtk_widget_show(gui.tp_label_ee);
            gtk_table_attach_defaults(GTK_TABLE(gui.tp_table), gui.tp_label_ee, 0, 1, eepos, eepos+1);

            gui.tp_range_ee = gtk_hscale_new_with_range(0, 15, 1);
	    gtk_scale_set_value_pos(GTK_SCALE(gui.tp_range_ee), GTK_POS_RIGHT);
            gtk_widget_show(gui.tp_range_ee);
            sisctrl_tv_edgeenhance_set_value();
            gtk_table_attach_defaults(GTK_TABLE(gui.tp_table), gui.tp_range_ee, 1, 3, eepos, eepos+1);

	    {
            GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(gui.tp_range_ee));
            g_signal_connect(G_OBJECT(myadj), "value_changed", G_CALLBACK(tp_ee_changed), NULL);
	    }
         }

	 gui.tp_label_coc = gtk_label_new("Color (coarse)");
         gtk_widget_show(gui.tp_label_coc);
         gtk_table_attach_defaults(GTK_TABLE(gui.tp_table), gui.tp_label_coc, 0, 1, colcpos, colcpos+1);

         gui.tp_range_coc = gtk_hscale_new_with_range(-120, 120, 1);
	 gtk_scale_set_value_pos(GTK_SCALE(gui.tp_range_coc), GTK_POS_RIGHT);
         gtk_widget_show(gui.tp_range_coc);

         sisctrl_tv_col_calib_set_c_value();

         gtk_table_attach_defaults(GTK_TABLE(gui.tp_table), gui.tp_range_coc, 1, 3, colcpos, colcpos+1);

	 gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.tp_range_coc,
		 	"Adjust color carrier frequency (coarse)", "");

         {
         GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(gui.tp_range_coc));
         g_signal_connect(G_OBJECT(myadj), "value_changed", G_CALLBACK(tp_coc_changed), NULL);
         }

	 gui.tp_label_cof = gtk_label_new("Color (fine)");
         gtk_widget_show(gui.tp_label_cof);
         gtk_table_attach_defaults(GTK_TABLE(gui.tp_table), gui.tp_label_cof, 0, 1, colfpos, colfpos+1);

         gui.tp_range_cof = gtk_hscale_new_with_range(-128, 127, 1);
	 gtk_scale_set_value_pos(GTK_SCALE(gui.tp_range_cof), GTK_POS_RIGHT);
         gtk_widget_show(gui.tp_range_cof);

         sisctrl_tv_col_calib_set_f_value();

         gtk_table_attach_defaults(GTK_TABLE(gui.tp_table), gui.tp_range_cof, 1, 3, colfpos, colfpos+1);

	 gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.tp_range_cof,
		 	"Adjust color carrier frequency (fine)", "");

         {
         GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(gui.tp_range_cof));
         g_signal_connect(G_OBJECT(myadj), "value_changed", G_CALLBACK(tp_cof_changed), NULL);
         }

	 gui.tp_check_cfi = gtk_check_button_new_with_label("C-Filter");
         gtk_table_attach_defaults(GTK_TABLE(gui.tp_table), gui.tp_check_cfi, 0, 1, filtpos, filtpos+1);
         gtk_widget_show(gui.tp_check_cfi);

	 gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.tp_check_cfi,
		 	"Enable chroma (C) filter", "");

         sisctrl_tv_c_filter_set_value();

         gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(gui.tp_check_cfi), TRUE);

	 g_signal_connect(G_OBJECT(gui.tp_check_cfi), "toggled", G_CALLBACK(tp_check_cfi_toggled), NULL);

	 gui.tp_label_yfi = gtk_label_new("      Y-Filter");
	 gtk_label_set_justify(GTK_LABEL(gui.tp_label_yfi), GTK_JUSTIFY_RIGHT);
         gtk_widget_show(gui.tp_label_yfi);
         gtk_table_attach_defaults(GTK_TABLE(gui.tp_table), gui.tp_label_yfi, 1, 2, filtpos, filtpos+1);

         gui.tp_omenuy = gtk_option_menu_new();

         gui.tp_menuy = gtk_menu_new();

         gui.tp_menuitemy1 = gtk_menu_item_new_with_label("Off");
         gui.tp_menuitemy2 = gtk_menu_item_new_with_label("Default");
         gui.tp_menuitemy3 = gtk_menu_item_new_with_label("Filter 1");
         gui.tp_menuitemy4 = gtk_menu_item_new_with_label("Filter 2");
         gui.tp_menuitemy5 = gtk_menu_item_new_with_label("Filter 3");
	 gui.tp_menuitemy6 = gtk_menu_item_new_with_label("Filter 4");
	 gui.tp_menuitemy7 = gtk_menu_item_new_with_label("Filter 5");
	 gui.tp_menuitemy8 = gtk_menu_item_new_with_label("Filter 6");
	 gui.tp_menuitemy9 = gtk_menu_item_new_with_label("Filter 7");

         gtk_menu_shell_append(GTK_MENU_SHELL(gui.tp_menuy), gui.tp_menuitemy1);
         gtk_menu_shell_append(GTK_MENU_SHELL(gui.tp_menuy), gui.tp_menuitemy2);
         gtk_menu_shell_append(GTK_MENU_SHELL(gui.tp_menuy), gui.tp_menuitemy3);
         gtk_menu_shell_append(GTK_MENU_SHELL(gui.tp_menuy), gui.tp_menuitemy4);
         gtk_menu_shell_append(GTK_MENU_SHELL(gui.tp_menuy), gui.tp_menuitemy5);
	 gtk_menu_shell_append(GTK_MENU_SHELL(gui.tp_menuy), gui.tp_menuitemy6);
	 gtk_menu_shell_append(GTK_MENU_SHELL(gui.tp_menuy), gui.tp_menuitemy7);
	 gtk_menu_shell_append(GTK_MENU_SHELL(gui.tp_menuy), gui.tp_menuitemy8);
	 gtk_menu_shell_append(GTK_MENU_SHELL(gui.tp_menuy), gui.tp_menuitemy9);

         gtk_widget_show(gui.tp_menuitemy1);
         gtk_widget_show(gui.tp_menuitemy2);
         gtk_widget_show(gui.tp_menuitemy3);
         gtk_widget_show(gui.tp_menuitemy4);
         gtk_widget_show(gui.tp_menuitemy5);
	 gtk_widget_show(gui.tp_menuitemy6);
	 gtk_widget_show(gui.tp_menuitemy7);
	 gtk_widget_show(gui.tp_menuitemy8);
	 gtk_widget_show(gui.tp_menuitemy9);

         gtk_option_menu_set_menu(GTK_OPTION_MENU(gui.tp_omenuy), gui.tp_menuy);

	 sisctrl_tv_y_filter_set_value();

	 gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.tp_omenuy,
		 	"Select luma (Y) filter", "");

         gtk_widget_show(gui.tp_omenuy);

         /* gtk_table_attach_defaults(GTK_TABLE(gui.tp_table), gui.tp_omenuy, 2, 3, filtpos, filtpos+1); */
	 gtk_table_attach(GTK_TABLE(gui.tp_table), gui.tp_omenuy, 2, 3, filtpos, filtpos+1, GTK_FILL|GTK_EXPAND, 0, 0, 0);

         g_signal_connect(G_OBJECT(gui.tp_omenuy), "changed", G_CALLBACK(tp_yfi_changed), NULL);
         break;

      case TV_PAGE_SUBTYPE_CHRONTEL:
         gui.tp_table = gtk_table_new(5, 3, TRUE);
         gtk_table_set_row_spacings(GTK_TABLE(gui.tp_table), 2);
         gtk_table_set_col_spacings(GTK_TABLE(gui.tp_table), 1);

         gtk_container_set_border_width(GTK_CONTAINER(gui.tp_table), 4);

         gtk_container_add(GTK_CONTAINER(gui.tp_frame), gui.tp_table);
         gtk_widget_show(gui.tp_table);

         gui.tp_label_co = gtk_label_new("Contrast");
         gtk_widget_show(gui.tp_label_co);
         gtk_table_attach_defaults(GTK_TABLE(gui.tp_table), gui.tp_label_co, 0, 1, 0, 1);

         gui.tp_range_co = gtk_hscale_new_with_range(0, 15, 1);
         gtk_scale_set_value_pos(GTK_SCALE(gui.tp_range_co), GTK_POS_RIGHT);
         gtk_widget_show(gui.tp_range_co);
         sisctrl_tv_ch_contrast_set_value();
         gtk_table_attach_defaults(GTK_TABLE(gui.tp_table), gui.tp_range_co, 1, 3, 0, 1);

         {
         GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(gui.tp_range_co));
         g_signal_connect(G_OBJECT(myadj), "value_changed", G_CALLBACK(tp_co_changed), NULL);
         }

         gui.tp_label_te = gtk_label_new("Text enhance");
         gtk_widget_show(gui.tp_label_te);
         gtk_table_attach_defaults(GTK_TABLE(gui.tp_table), gui.tp_label_te, 0, 1, 1, 2);

         gui.tp_range_te = gtk_hscale_new_with_range(0, 15, 1);
         gtk_scale_set_value_pos(GTK_SCALE(gui.tp_range_te), GTK_POS_RIGHT);
         gtk_widget_show(gui.tp_range_te);
         sisctrl_tv_ch_textenhance_set_value();
         gtk_table_attach_defaults(GTK_TABLE(gui.tp_table), gui.tp_range_te, 1, 3, 1, 2);

         {
         GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(gui.tp_range_te));
         g_signal_connect(G_OBJECT(myadj), "value_changed", G_CALLBACK(tp_te_changed), NULL);
         }

         gui.tp_label_cf = gtk_label_new("C-flickerfilter");
         gtk_widget_show(gui.tp_label_cf);
         gtk_table_attach_defaults(GTK_TABLE(gui.tp_table), gui.tp_label_cf, 0, 1, 2, 3);

         gui.tp_range_cf = gtk_hscale_new_with_range(0, 15, 1);
         gtk_scale_set_value_pos(GTK_SCALE(gui.tp_range_cf), GTK_POS_RIGHT);
         gtk_widget_show(gui.tp_range_cf);
         sisctrl_tv_ch_chromaff_set_value();
         gtk_table_attach_defaults(GTK_TABLE(gui.tp_table), gui.tp_range_cf, 1, 3, 2, 3);

         {
         GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(gui.tp_range_cf));
         g_signal_connect(G_OBJECT(myadj), "value_changed", G_CALLBACK(tp_cf_changed), NULL);
         }

         gui.tp_label_lf = gtk_label_new("L-flickerfilter");
         gtk_widget_show(gui.tp_label_lf);
         gtk_table_attach_defaults(GTK_TABLE(gui.tp_table), gui.tp_label_lf, 0, 1, 3, 4);

	 gui.tp_range_lf = gtk_hscale_new_with_range(0, 15, 1);
         gtk_scale_set_value_pos(GTK_SCALE(gui.tp_range_lf), GTK_POS_RIGHT);
         gtk_widget_show(gui.tp_range_lf);
         sisctrl_tv_ch_lumaff_set_value();
         gtk_table_attach_defaults(GTK_TABLE(gui.tp_table), gui.tp_range_lf, 1, 3, 3, 4);

         {
         GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(gui.tp_range_lf));
         g_signal_connect(G_OBJECT(myadj), "value_changed", G_CALLBACK(tp_lf_changed), NULL);
         }

         gui.tp_check_col = gtk_check_button_new_with_label("Color (CVBS)");
         gtk_table_attach_defaults(GTK_TABLE(gui.tp_table), gui.tp_check_col, 0, 3, 4, 5);
         gtk_widget_show(gui.tp_check_col);
         sisctrl_tv_ch_cvbscol_set_value();
         g_signal_connect(G_OBJECT(gui.tp_check_col), "toggled", G_CALLBACK (tp_check_col_toggled), NULL);

         break;
      }

      if(gl.supporttvsize) {
         gui.tv_frame = gtk_frame_new("TV position & size");
      } else {
         gui.tv_frame = gtk_frame_new("TV position");
      }
      gtk_container_set_border_width(GTK_CONTAINER(gui.tv_frame), 8);
      gtk_widget_show(gui.tv_frame);

      /* gtk_box_pack_start(GTK_BOX(gui.tp_box), gui.tv_frame, FALSE, FALSE, 0); */
      gtk_table_attach_defaults(GTK_TABLE(gui.tp_box), gui.tv_frame, 1, 2, 0, 1);

      gui.tv_table = gtk_table_new(3, 3, TRUE);
      gtk_table_set_row_spacings(GTK_TABLE(gui.tv_table), 8);

      gtk_table_set_col_spacings(GTK_TABLE(gui.tv_table), 4);

      gtk_container_set_border_width(GTK_CONTAINER(gui.tv_table), 4);

      gtk_container_add(GTK_CONTAINER(gui.tv_frame), gui.tv_table);
      gtk_widget_show(gui.tv_table);

      gui.tv_arrow_le = xpm_label_box(icon_tvleft, NULL, NULL, 1, "", "");
      gui.tv_arrow_ri = xpm_label_box(icon_tvright, NULL, NULL, 1, "", "");
      gui.tv_arrow_up = xpm_label_box(icon_tvup, NULL, NULL, 1, "", "");
      gui.tv_arrow_dn = xpm_label_box(icon_tvdown, NULL, NULL, 1, "", "");

      gui.tv_button_up = gtk_button_new();
      gui.tv_button_dn = gtk_button_new();
      gui.tv_button_ri = gtk_button_new();
      gui.tv_button_le = gtk_button_new();

      gtk_container_add(GTK_CONTAINER(gui.tv_button_up), gui.tv_arrow_up);
      gtk_container_add(GTK_CONTAINER(gui.tv_button_dn), gui.tv_arrow_dn);
      gtk_container_add(GTK_CONTAINER(gui.tv_button_le), gui.tv_arrow_le);
      gtk_container_add(GTK_CONTAINER(gui.tv_button_ri), gui.tv_arrow_ri);

      gui.tv_button_de = gtk_button_new_with_mnemonic("_Default");

      gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.tv_button_up,
		 	"Move TV output upwards", "");
      gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.tv_button_dn,
		 	"Move TV output downwards", "");
      gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.tv_button_ri,
		 	"Move TV output right", "");
      gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.tv_button_le,
		 	"Move TV output left", "");

      gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.tv_button_de,
			gl.supporttvsize ? "Reset TV position and size" :
					   "Reset TV position", "");

      gtk_widget_show(gui.tv_button_up);
      gtk_widget_show(gui.tv_button_ri);
      gtk_widget_show(gui.tv_button_le);
      gtk_widget_show(gui.tv_button_dn);
      gtk_widget_show(gui.tv_button_de);

      gtk_widget_show(gui.tv_arrow_le);
      gtk_widget_show(gui.tv_arrow_ri);
      gtk_widget_show(gui.tv_arrow_dn);
      gtk_widget_show(gui.tv_arrow_up);

      gtk_table_attach_defaults(GTK_TABLE(gui.tv_table), gui.tv_button_up, 1, 2, 0, 1);
      gtk_table_attach_defaults(GTK_TABLE(gui.tv_table), gui.tv_button_le, 0, 1, 1, 2);
      gtk_table_attach_defaults(GTK_TABLE(gui.tv_table), gui.tv_button_de, 1, 2, 1, 2);
      gtk_table_attach_defaults(GTK_TABLE(gui.tv_table), gui.tv_button_ri, 2, 3, 1, 2);
      gtk_table_attach_defaults(GTK_TABLE(gui.tv_table), gui.tv_button_dn, 1, 2, 2, 3);

      if(!gl.supporttvpos) {
         gtk_widget_set_sensitive(gui.tv_button_up, FALSE);
         gtk_widget_set_sensitive(gui.tv_button_dn, FALSE);
         gtk_widget_set_sensitive(gui.tv_button_le, FALSE);
         gtk_widget_set_sensitive(gui.tv_button_ri, FALSE);
         gtk_widget_set_sensitive(gui.tv_button_de, FALSE);
      }

      g_signal_connect(G_OBJECT (gui.tv_button_up), "clicked", G_CALLBACK (tv_up_clicked), NULL);
      g_signal_connect(G_OBJECT (gui.tv_button_dn), "clicked", G_CALLBACK (tv_dn_clicked), NULL);
      g_signal_connect(G_OBJECT (gui.tv_button_le), "clicked", G_CALLBACK (tv_le_clicked), NULL);
      g_signal_connect(G_OBJECT (gui.tv_button_ri), "clicked", G_CALLBACK (tv_ri_clicked), NULL);
      g_signal_connect(G_OBJECT (gui.tv_button_de), "clicked", G_CALLBACK (tv_de_clicked), NULL);

      if(gl.supporttvsize) {
         gui.tv_arrow_xp = xpm_label_box(icon_tvxp, NULL, NULL, 1, "", "");
	 gui.tv_arrow_xm = xpm_label_box(icon_tvxm, NULL, NULL, 1, "", "");
	 gui.tv_arrow_yp = xpm_label_box(icon_tvyp, NULL, NULL, 1, "", "");
	 gui.tv_arrow_ym = xpm_label_box(icon_tvym, NULL, NULL, 1, "", "");

	 gui.tv_button_xp = gtk_button_new();
         gui.tv_button_xm = gtk_button_new();
	 gui.tv_button_yp = gtk_button_new();
         gui.tv_button_ym = gtk_button_new();

	 gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.tv_button_xp,
       		 	"Enlarge TV output horizontally", "");
         gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.tv_button_xm,
       		 	"Shrink TV output horizontally", "");
         gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.tv_button_yp,
       		 	"Enlarge TV output vertically", "");
         gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.tv_button_ym,
       		 	"Shrink TV output vertically", "");

	 gtk_container_set_border_width(GTK_CONTAINER(gui.tv_button_xp), 1);
	 gtk_container_set_border_width(GTK_CONTAINER(gui.tv_button_xm), 1);
	 gtk_container_set_border_width(GTK_CONTAINER(gui.tv_button_yp), 1);
	 gtk_container_set_border_width(GTK_CONTAINER(gui.tv_button_ym), 1);

	 gtk_container_add(GTK_CONTAINER(gui.tv_button_xp), gui.tv_arrow_xp);
         gtk_container_add(GTK_CONTAINER(gui.tv_button_xm), gui.tv_arrow_xm);
	 gtk_container_add(GTK_CONTAINER(gui.tv_button_yp), gui.tv_arrow_yp);
         gtk_container_add(GTK_CONTAINER(gui.tv_button_ym), gui.tv_arrow_ym);

	 gtk_widget_show(gui.tv_button_xp);
         gtk_widget_show(gui.tv_button_xm);
	 gtk_widget_show(gui.tv_button_yp);
         gtk_widget_show(gui.tv_button_ym);

	 gtk_widget_show(gui.tv_arrow_xp);
	 gtk_widget_show(gui.tv_arrow_xm);
         gtk_widget_show(gui.tv_arrow_yp);
	 gtk_widget_show(gui.tv_arrow_ym);

	 gtk_table_attach_defaults(GTK_TABLE(gui.tv_table), gui.tv_button_xm, 0, 1, 2, 3);
         gtk_table_attach_defaults(GTK_TABLE(gui.tv_table), gui.tv_button_xp, 2, 3, 2, 3);
	 gtk_table_attach_defaults(GTK_TABLE(gui.tv_table), gui.tv_button_ym, 0, 1, 0, 1);
         gtk_table_attach_defaults(GTK_TABLE(gui.tv_table), gui.tv_button_yp, 2, 3, 0, 1);

	 g_signal_connect(G_OBJECT (gui.tv_button_xp), "clicked", G_CALLBACK (tv_xp_clicked), NULL);
         g_signal_connect(G_OBJECT (gui.tv_button_xm), "clicked", G_CALLBACK (tv_xm_clicked), NULL);
	 g_signal_connect(G_OBJECT (gui.tv_button_yp), "clicked", G_CALLBACK (tv_yp_clicked), NULL);
         g_signal_connect(G_OBJECT (gui.tv_button_ym), "clicked", G_CALLBACK (tv_ym_clicked), NULL);
      }

      gui.tp_label = xpm_label_box(icon_tvsettings, "TV", NULL, 1, "TV settings", "");
      gtk_notebook_append_page(GTK_NOTEBOOK(gui.notebook), gui.tp_box, gui.tp_label);  /* @@@@ */

      break;

   case TV_PAGE_TYPE_NA_NOTV:
      gui.tp_frame = gtk_frame_new("TV settings");
      gtk_container_set_border_width(GTK_CONTAINER(gui.tp_frame), 8);
      gtk_widget_show(gui.tp_frame);

      gui.tv_table = gtk_table_new(3, 3, TRUE);
      gtk_table_set_row_spacings(GTK_TABLE(gui.tv_table), 10);
      gtk_table_set_col_spacings(GTK_TABLE(gui.tv_table), 4);

      gtk_container_set_border_width(GTK_CONTAINER(gui.tv_table), 8);

      gtk_container_add(GTK_CONTAINER(gui.tp_frame), gui.tv_table);
      gtk_widget_show(gui.tv_table);

      gui.tv_label_notv = gtk_label_new("No TV connector present");
      gtk_widget_show(gui.tv_label_notv);
      gtk_table_attach_defaults(GTK_TABLE(gui.tv_table), gui.tv_label_notv, 0, 3, 1, 2);

      gui.tp_label = xpm_label_box(icon_tvsettings, "TV", NULL, 1, "TV settings", "");
      gtk_notebook_append_page(GTK_NOTEBOOK(gui.notebook), gui.tp_frame, gui.tp_label);
      break;
   }
}

/**************************************************
 *                    Video page                  *
 **************************************************/

void
gui_vi_slider_set_value(int slider, double value)
{
   GtkAdjustment *myadj;

   if(gl.video_page_type != VIDEO_PAGE_TYPE_NORMAL) return;

   switch(slider) {
   case VI_SLI_CONTRAST:
      myadj = gtk_range_get_adjustment(GTK_RANGE(gui.vi_range_co));
      break;
   case VI_SLI_BRIGHTNESS:
      myadj = gtk_range_get_adjustment(GTK_RANGE(gui.vi_range_br));
      break;
   case VI_SLI_HUE:
      myadj = gtk_range_get_adjustment(GTK_RANGE(gui.vi_range_hu));
      break;
   case VI_SLI_SATURATION:
      myadj = gtk_range_get_adjustment(GTK_RANGE(gui.vi_range_sa));
      break;
   default:
      return;
   }

   gtk_adjustment_set_value(myadj, value);
   g_signal_emit_by_name(G_OBJECT(myadj), "changed");
}

void
gui_vi_enable_and_set_xv_gamma_check(int enable, int active)
{
   if(gl.video_page_type != VIDEO_PAGE_TYPE_NORMAL) return;

   gtk_widget_set_sensitive(gui.vi_check_gamma, enable ? TRUE : FALSE);
   gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(gui.vi_check_gamma), active ? TRUE : FALSE);
}

void
gui_vi_enable_xvgamma_spins(int enable)
{
   if(gl.video_page_type != VIDEO_PAGE_TYPE_NORMAL) return;
   if(!gl.supportxvgamma1) return;

   gtk_widget_set_sensitive(gui.vi_spin1, enable ? TRUE : FALSE);
   gtk_widget_set_sensitive(gui.vi_spin2, enable ? TRUE : FALSE);
   gtk_widget_set_sensitive(gui.vi_spin3, enable ? TRUE : FALSE);
}

void
gui_vi_set_xvgamma_spins(double s1, double s2, double s3)
{
   if(gl.video_page_type != VIDEO_PAGE_TYPE_NORMAL) return;

   gtk_spin_button_set_value(GTK_SPIN_BUTTON(gui.vi_spin1), s1);
   gtk_spin_button_set_value(GTK_SPIN_BUTTON(gui.vi_spin2), s2);
   gtk_spin_button_set_value(GTK_SPIN_BUTTON(gui.vi_spin3), s3);
}

void
gui_vi_enable_switchcrt(int enable)
{
   if(gl.video_page_type != VIDEO_PAGE_TYPE_NORMAL) return;

   gtk_widget_set_sensitive(gui.vi_radio_crt1, enable ? TRUE : FALSE);
   gtk_widget_set_sensitive(gui.vi_radio_crt2, enable ? TRUE : FALSE);
}

void
gui_vi_set_switchcrt_active_num(int num)
{
   if(gl.video_page_type != VIDEO_PAGE_TYPE_NORMAL) return;

   switch(num) {
   case 1:
      if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(gui.vi_radio_crt1)) == FALSE) {
         gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(gui.vi_radio_crt1), TRUE);
      }
      break;
   case 2:
   default:
      if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(gui.vi_radio_crt2)) == FALSE) {
         gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(gui.vi_radio_crt2), TRUE);
      }
   }
}

void
gui_vi_set_video_only_value(int active)
{
   if(gl.video_page_type != VIDEO_PAGE_TYPE_NORMAL) return;

   gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(gui.vi_check_video_only), active ? TRUE : FALSE);
}

/* Xv demo stuff */

#ifdef USEXV
static int
myxvwinerrorhandler(Display *dpy, XErrorEvent *error)
{
   gui.xerror = 1;
   return 0;
}

static Bool
testxvwindow(void)
{
   XErrorHandler oldErrHandler;

   if(gl.noxvdemo) return FALSE;

   /* We now create a child window of our GUI window. If that
    * fails, we don't use the xv demo image.
    * We use our (eventually user-provided) dpy here on purpose.
    * This is some sort of hack-ish way to determine if we run
    * on a remote display (where the window creation will fail
    * because one can't make child windows on different displays).
    */

   gui.xerror = 0;
   oldErrHandler = XSetErrorHandler(myxvwinerrorhandler);

   gui.xvwin = XCreateSimpleWindow(gl.dpy,
				   GDK_WINDOW_XWINDOW(gui.window->window),
				   0, 0,
				   1, 1,
				   1, 0, 0);
   XSync(gl.dpy, False);
   XSetErrorHandler(oldErrHandler);

   if(!gui.xvwin) return FALSE;

   if(gui.xerror) {
      /* No, don't call Destroy() here ... BadWindow! */
      gui.xvwin = 0;
      return FALSE;
   }

   XDestroyWindow(gl.dpy, gui.xvwin);
   gui.xvwin = 0;
   return TRUE;
}

static void
destroyxvimage(void)
{
   if(gui.xvshown) {
      XvStopVideo(gl.dpy, gl.xv_port_nr, gui.xvwin);
      gui.xvshown = FALSE;
   }
   if(gui.xvattached) {
      XShmDetach(gl.dpy, &gui.xvshminfo);
      shmdt(gui.xvshminfo.shmaddr);
      shmctl(gui.xvshminfo.shmid, IPC_RMID, 0);
      XFree(gui.myxvimage);
      gui.myxvimage = NULL;
      gui.xvattached = FALSE;
   }
   if(gui.xvcontext) {
      XFreeGC(gl.dpy, gui.xvcontext);
      gui.xvcontext = NULL;
   }
   if(gui.xvwin) {
      XDestroyWindow(gl.dpy, gui.xvwin);
      gui.xvwin = 0;
   }
}

static Bool
makexvwindow(void)
{
   char *ptr;
   int i, j;
   XErrorHandler oldErrHandler;

   if(gl.noxvdemo) return FALSE;

   if(gui.xvwin) return TRUE;

   gui.xerror = 0;
   oldErrHandler = XSetErrorHandler(myxvwinerrorhandler);

   gui.xvwin = XCreateSimpleWindow(gl.dpy,
				   GDK_WINDOW_XWINDOW(gui.window->window),
				   0, 0,
				   1, 1,
				   1, 0, 0);
   XSync(gl.dpy, False);
   XSetErrorHandler(oldErrHandler);

   if(!gui.xvwin) return FALSE;

   if(gui.xerror) {
      gui.xvwin = 0;
      return FALSE;
   }

   XMapWindow(gl.dpy, gui.xvwin);
   XFlush(gl.dpy);
   gui.xvraised = FALSE;

   if(!gui.xvcontext) {
      gui.xvcontext = XCreateGC(gl.dpy, gui.xvwin, 0, 0);
   }
   if(!gui.xvcontext) {
      destroyxvimage();
      return FALSE;
   }

   if(!gui.myxvimage) {
      gui.myxvimage = XvShmCreateImage(gl.dpy, gl.xv_port_nr,
             		0x59565955, 0, XVWIDTH, XVHEIGHT,
			&gui.xvshminfo);
   }
   if(!gui.myxvimage) {
      fprintf(stderr, "XvShmCreateImage failed\n");
      destroyxvimage();
      return FALSE;
   }

   gui.xvshminfo.shmid = shmget(IPC_PRIVATE, gui.myxvimage->data_size, IPC_CREAT|0777);
   gui.xvshminfo.shmaddr = gui.myxvimage->data = shmat(gui.xvshminfo.shmid, 0, 0);
   gui.xvshminfo.readOnly = False;
   gui.xvattached = XShmAttach(gl.dpy, &gui.xvshminfo);
   if(gui.xvattached) {
      ptr = gui.myxvimage->data;
      if(!ptr) {
         fprintf(stderr, "XvImage->data is NULL\n");
         destroyxvimage();
         return FALSE;
      } else {
         for(i = 0; i < XVHEIGHT; i++, ptr += gui.myxvimage->pitches[0]) {
            for(j = 0; j < XVWIDTH; j++) {
	       ptr[(j << 1)] = xviconbuffer[(i * XVWIDTH * 2) + (j*2)];
	       ptr[(j << 1) + 1] = xviconbuffer[(i * XVWIDTH * 2) + (j*2) + 1];
            }
	 }
      }
      return TRUE;
   } else {
      fprintf(stderr, "XShmAttach failed\n");
      destroyxvimage();
      return FALSE;
   }
}

void
gui_destroy_xv_image(void)
{
   if(!gl.noxvdemo) {
      destroyxvimage();
   }
}

static void
resizexvwindow(void)
{
   XWindowAttributes wa;

   if(!gui.xvinit) return;
   if(!gui.xvwin) return;

   if(!GTK_WIDGET_REALIZED(gui.vi_layout)) {
      /* fprintf(stderr, "resetting size\n"); */
      gui.xvx = gui.xvy = 0;
      gui.xvwidth = gui.xvheight = 1;
      return;
   }

   gui.xvx = gui.vi_layout->allocation.x;
   gui.xvy = gui.vi_layout->allocation.y;
   gui.xvwidth = gui.vi_layout->allocation.width;
   gui.xvheight = gui.vi_layout->allocation.height;

   XGetWindowAttributes(gl.dpy, gui.xvwin, &wa);

   if( (gui.xvx != wa.x) ||
       (gui.xvy != wa.y) ||
       (gui.xvwidth != wa.width) ||
       (gui.xvheight != wa.height) ) {
      XMoveResizeWindow(gl.dpy, gui.xvwin,
   		gui.vi_layout->allocation.x,
		gui.vi_layout->allocation.y,
		gui.vi_layout->allocation.width,
   		gui.vi_layout->allocation.height);
   }
}

static void
hidexvimage(void)
{
   if(!gui.xvinit) return;
   if(gui.xvshown) {
      sisctrl_restore_disable_gfx();
      XvStopVideo(gl.dpy, gl.xv_port_nr, gui.xvwin);
      XSync(gl.dpy, False);
      gui.xvshown = FALSE;
   }
   if(gui.xvraised) {
      XUnmapWindow(gl.dpy,gui.xvwin);
      XFlush(gl.dpy);
      gui.xvraised = FALSE;
   }
}

static void
showxvimage(void)
{
   gui.xvshown = FALSE;

   if(!gui.xvinit) return;

   if(!gui.xvwin) return;
   if(!gui.xvattached) return;
   if((gui.xvwidth == 0) || (gui.xvheight == 0)) return;

   if(XvGrabPort(gl.dpy, gl.xv_port_nr, CurrentTime) == Success) {
      XvUngrabPort(gl.dpy, gl.xv_port_nr, CurrentTime);
   } else {
      fprintf(stderr, "Xv busy\n");
      hidexvimage();
      gui.xvwasbusy = TRUE;
      return;
   }

   if((!gui.xvraised) || (gui.xvwasbusy)) {
      /* fprintf(stderr, "raised %d, wasbusy %d\n", gui.xvraised, gui.xvwasbusy); */
      XMapWindow(gl.dpy,gui.xvwin);
      XRaiseWindow(gl.dpy, gui.xvwin);
      XFlush(gl.dpy);
      resizexvwindow();
      XSync(gl.dpy, False);
      gui.xvraised = TRUE;
      gui.xvwasbusy = FALSE;
   }

   sisctrl_set_disable_gfx(0, 0);

   XvShmPutImage(gl.dpy, gl.xv_port_nr,
                 gui.xvwin,
                 gui.xvcontext,
		 gui.myxvimage,
		 0, 0,
		 XVWIDTH, XVHEIGHT,
		 0, 0,
		 gui.xvwidth, gui.xvheight,
		 False);

   XSync(gl.dpy, False);

   gui.xvshown = TRUE;
}

int
gui_vi_get_xv_shown(void)
{
   return gui.xvshown ? 1 : 0;
}
#endif

/* Signal handlers */

static void
vi_co_changed(GtkWidget *widget, gpointer data)
{
   double myval = gtk_adjustment_get_value(GTK_ADJUSTMENT(widget));

   sisctrl_vi_co_changed(myval);

#ifdef USEXV
   if(!gl.noxvdemo) showxvimage();
#endif
}

static void
vi_br_changed(GtkWidget *widget, gpointer data)
{
   double myval = gtk_adjustment_get_value(GTK_ADJUSTMENT(widget));

   sisctrl_vi_br_changed(myval);

#ifdef USEXV
   if(!gl.noxvdemo) showxvimage();
#endif
}

static void
vi_hu_changed(GtkWidget *widget, gpointer data)
{
   double myval = gtk_adjustment_get_value(GTK_ADJUSTMENT(widget));

   sisctrl_vi_hu_changed(myval);

#ifdef USEXV
   if(!gl.noxvdemo) showxvimage();
#endif
}

static void
vi_sa_changed(GtkWidget *widget, gpointer data)
{
   double myval = gtk_adjustment_get_value(GTK_ADJUSTMENT(widget));

   sisctrl_vi_sa_changed(myval);

#ifdef USEXV
   if(!gl.noxvdemo) showxvimage();
#endif
}

static void
vi_radio_changed(GtkWidget *widget, gpointer data)
{
   sisctrl_vi_radio_changed(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget)) ? 0 : 1);
#ifdef USEXV
   if(!gl.noxvdemo) showxvimage();
#endif
}

static void
vi_check_gamma_toggled(GtkWidget *widget, gpointer data)
{
   sisctrl_vi_check_gamma_toggled(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget)) ? 1 : 0);
#ifdef USEXV
   if(!gl.noxvdemo) showxvimage();
#endif
}

static void
vi_spin1_changed(GtkWidget *widget, gpointer data)
{
   double val = gtk_spin_button_get_value(GTK_SPIN_BUTTON(gui.vi_spin1));
   sisctrl_vi_spin1_changed(val);
}

static void
vi_spin2_changed(GtkWidget *widget, gpointer data)
{
   double val = gtk_spin_button_get_value(GTK_SPIN_BUTTON(gui.vi_spin2));
   sisctrl_vi_spin2_changed(val);
}

static void
vi_spin3_changed(GtkWidget *widget, gpointer data)
{
   double val = gtk_spin_button_get_value(GTK_SPIN_BUTTON(gui.vi_spin3));
   sisctrl_vi_spin3_changed(val);
}

static void
vi_check_video_only_toggled(GtkWidget *widget, gpointer data)
{
   int val = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(gui.vi_check_video_only)) ? 1 : 0;
   sisctrl_vi_check_video_only_toggled(val);
   sisctrl_restore_disable_gfx_lr();
}

static void
vi_radio_gl_changed(GtkWidget *widget, gpointer data)
{
   if(!gui.haveglobal) return;

   gui.isglobal = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget)) ? TRUE : FALSE;

   sisctrl_vi_global_changed(gui.isglobal ? 1 : 0);

   sisctrl_vi_set_contrast_value();
   sisctrl_vi_set_brightness_value();
   if(gl.supporthuesat) {
      sisctrl_vi_set_hue_value();
      sisctrl_vi_set_saturation_value();
   }
}

#ifdef USEXV
static gboolean
vi_layout_expose(GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
   /* fprintf(stderr, "layout expose\n"); */
   if(gl.noxvdemo) return TRUE;
   if(gui.notebook_page != 5) return TRUE;
   if(!gui.xvwin) return TRUE;
   if(!gui.xvinit) return TRUE;
   resizexvwindow();
   showxvimage();
   return TRUE;
}

static gboolean
vi_layout_configure(GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
   /* fprintf(stderr, "layout config\n"); */
   if(gl.noxvdemo) return TRUE;
   if(gui.notebook_page != 5) return;
   if(!gui.xvwin) return;
   if(!gui.xvinit) return TRUE;
   resizexvwindow();
   showxvimage();
   return TRUE;
}

static gboolean
window_configure(GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
   /* fprintf(stderr, "window config\n"); */
   if(gl.noxvdemo) return TRUE;
   if(gui.notebook_page != 5) return FALSE;
   if(!gui.xvwin) return FALSE;
   if(!gui.xvinit) return TRUE;
   resizexvwindow();
   showxvimage();
   return FALSE;
}

static gboolean
window_expose(GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
   /* fprintf(stderr, "window expose\n");    */
   if(gl.noxvdemo) return TRUE;
   if(gui.notebook_page != 5) return FALSE;
   if(!gui.xvwin) return FALSE;
   if(!gui.xvinit) return TRUE;
   resizexvwindow();
   showxvimage();
   return FALSE;
}

static gboolean
window_visibility(GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
   /* fprintf(stderr, "window visibility\n"); */
   if(gl.noxvdemo) return TRUE;
   if(gui.notebook_page != 5) return FALSE;
   if(!gui.xvwin) return FALSE;
   if(!gui.xvinit) return TRUE;
   /* hidexvimage(); */ /* causes blinking, but required; no - window expose fixes this */
   resizexvwindow();
   showxvimage();
   return FALSE;
}
#endif

/* Build page */

static void
build_video_page(void)
{
   GtkWidget *vitable;
   int rows, colspacing, border, contrastpos, gammapos, switchpos, rows2;

   gui.vi_frame = gtk_frame_new("Video settings (Xv overlay)");
   gtk_container_set_border_width(GTK_CONTAINER(gui.vi_frame), 8);
   gtk_widget_show(gui.vi_frame);

   switch(gl.video_page_type) {

   case VIDEO_PAGE_TYPE_NORMAL:

      rows = 5; colspacing = 10; border = 15;
      rows2 = 4;
      contrastpos = 0;
      gammapos = 4; switchpos = 4;

      if(gl.video_page_subtype == VIDEO_PAGE_SUBTYPE_WITH_GLOBAL) {
         rows++;
	 rows2++;
	 contrastpos++;
	 gammapos++;
	 switchpos++;
      }

      if(gl.supportxvgamma1) {
         rows++;
	 switchpos++;
      }

#if 0
      if(vbhavevb()) {
         if( (!(gui.sdflags & SiS_SD_ISMERGEDFB)) &&
	     ((!(gui.sdflags & SiS_SD_SUPPORT2OVL)) || (gui.sd2flags & SiS_SD2_SUPPORT760OO)) ) {
	    rows++;
	 }
      }
#endif

#ifdef USEXV
      if(!gl.noxvdemo) {
         colspacing = 5; border = 10;
      }
#endif
      gui.vi_table = gtk_table_new(rows, 3, FALSE);
      gtk_table_set_row_spacings(GTK_TABLE(gui.vi_table), 2);
      gtk_table_set_col_spacings(GTK_TABLE(gui.vi_table), colspacing);
      gtk_container_set_border_width(GTK_CONTAINER(gui.vi_table), border);
      vitable = gui.vi_table;

      gtk_container_add(GTK_CONTAINER(gui.vi_frame), gui.vi_table);
      gtk_widget_show(gui.vi_table);

#ifdef USEXV
      if(!gl.noxvdemo) {
         gui.vi_table2 = gtk_table_new(rows2, 3, FALSE);
         gtk_table_set_row_spacings(GTK_TABLE(gui.vi_table), 2);
         gtk_table_set_col_spacings(GTK_TABLE(gui.vi_table), 2);
         gtk_container_set_border_width(GTK_CONTAINER(gui.vi_table2), 0);
         gtk_table_attach_defaults(GTK_TABLE(gui.vi_table), gui.vi_table2, 1, 3, 0, rows2);
         gtk_widget_show(gui.vi_table2);
         vitable = gui.vi_table2;
      }
#endif

      gui.haveglobal = FALSE;
      gui.isglobal = FALSE;

      if(gl.video_page_subtype == VIDEO_PAGE_SUBTYPE_WITH_GLOBAL) {

         char buffer[16];
	 sprintf(buffer, "for screen %d", gl.screennum);

	 gui.vi_label_gl = gtk_label_new("Adjust");
         gtk_widget_show(gui.vi_label_gl);
         gtk_table_attach_defaults(GTK_TABLE(vitable), gui.vi_label_gl, 0, 1, 0, 1);

	 gui.vi_radio_gl = gtk_radio_button_new_with_label(NULL, "globally");
	 gtk_table_attach_defaults(GTK_TABLE(vitable), gui.vi_radio_gl, 1, 2, 0, 1);
	 gtk_widget_show(gui.vi_radio_gl);

	 gui.vi_radio_th = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(gui.vi_radio_gl), buffer);
	 gtk_table_attach_defaults(GTK_TABLE(vitable), gui.vi_radio_th, 2, 3, 0, 1);
	 gtk_widget_show(gui.vi_radio_th);

	 gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(gui.vi_radio_gl), TRUE);
	 gui.haveglobal = TRUE;
	 gui.isglobal = TRUE;

	 g_signal_connect(G_OBJECT(gui.vi_radio_gl), "toggled", G_CALLBACK(vi_radio_gl_changed), NULL);

	 gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.vi_radio_gl,
		 	"Adjust contrast, brightness, hue and saturation for all screens", "");
	 gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.vi_radio_th,
		 	"Adjust contrast, brightness, hue and saturation for that screen only", "");

	 sisctrl_vi_global_changed(gui.isglobal ? 1 : 0);
      }

      gui.vi_label_co = gtk_label_new("Contrast");
      gtk_widget_show(gui.vi_label_co);
      gtk_table_attach_defaults(GTK_TABLE(vitable), gui.vi_label_co, 0, 1, contrastpos+0, contrastpos+1);

      gui.vi_range_co = gtk_hscale_new_with_range(0, 7, 1);
      gtk_scale_set_value_pos(GTK_SCALE(gui.vi_range_co), GTK_POS_RIGHT);
      gtk_widget_show(gui.vi_range_co);
      sisctrl_vi_set_contrast_value();
      gtk_table_attach_defaults(GTK_TABLE(vitable), gui.vi_range_co, 1, 3, contrastpos+0, contrastpos+1);

      gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.vi_range_co,
		 	"Adjust video contrast", "");

      {
         GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(gui.vi_range_co));
         g_signal_connect(G_OBJECT(myadj), "value_changed", G_CALLBACK(vi_co_changed), NULL);
      }

      gui.vi_label_br = gtk_label_new("Brightness");
      gtk_widget_show(gui.vi_label_br);
      gtk_table_attach_defaults(GTK_TABLE(vitable), gui.vi_label_br, 0, 1, contrastpos+1, contrastpos+2);

      gui.vi_range_br = gtk_hscale_new_with_range(-128, 127, 1);
      gtk_scale_set_value_pos(GTK_SCALE(gui.vi_range_br), GTK_POS_RIGHT);
      gtk_widget_show(gui.vi_range_br);
      sisctrl_vi_set_brightness_value();
      gtk_table_attach_defaults(GTK_TABLE(vitable), gui.vi_range_br, 1, 3, contrastpos+1, contrastpos+2);

      gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.vi_range_br,
		 	"Adjust video brightness", "");

      {
         GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(gui.vi_range_br));
         g_signal_connect(G_OBJECT(myadj), "value_changed", G_CALLBACK(vi_br_changed), NULL);
      }

      if(gl.supporthuesat) {
         gui.vi_label_hu = gtk_label_new("Hue");
         gtk_widget_show(gui.vi_label_hu);
         gtk_table_attach_defaults(GTK_TABLE(vitable), gui.vi_label_hu, 0, 1, contrastpos+2, contrastpos+3);

         gui.vi_range_hu = gtk_hscale_new_with_range(-8, 7, 1);
         gtk_scale_set_value_pos(GTK_SCALE(gui.vi_range_hu), GTK_POS_RIGHT);
         gtk_widget_show(gui.vi_range_hu);
         sisctrl_vi_set_hue_value();
         gtk_table_attach_defaults(GTK_TABLE(vitable), gui.vi_range_hu, 1, 3, contrastpos+2, contrastpos+3);

	 gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.vi_range_hu,
		 	"Adjust video hue", "");

         {
            GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(gui.vi_range_hu));
            g_signal_connect(G_OBJECT(myadj), "value_changed", G_CALLBACK(vi_hu_changed), NULL);
         }

         gui.vi_label_sa = gtk_label_new("Saturation");
         gtk_widget_show(gui.vi_label_sa);
         gtk_table_attach_defaults(GTK_TABLE(vitable), gui.vi_label_sa, 0, 1, contrastpos+3, contrastpos+4);

         gui.vi_range_sa = gtk_hscale_new_with_range(-7, 7, 1);
         gtk_scale_set_value_pos(GTK_SCALE(gui.vi_range_sa), GTK_POS_RIGHT);
         gtk_widget_show(gui.vi_range_sa);
         sisctrl_vi_set_saturation_value();
         gtk_table_attach_defaults(GTK_TABLE(vitable), gui.vi_range_sa, 1, 3, contrastpos+3, contrastpos+4);

	 gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.vi_range_sa,
		 	"Adjust video saturation", "");

         {
            GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(gui.vi_range_sa));
            g_signal_connect(G_OBJECT(myadj), "value_changed", G_CALLBACK(vi_sa_changed), NULL);
         }
      }

#ifdef USEXV
      if(!gl.noxvdemo) {
         gui.vi_layout = gtk_label_new("  ");
         gtk_table_attach_defaults(GTK_TABLE(gui.vi_table), gui.vi_layout, 0, 1, 0, rows2);
         gtk_widget_show(gui.vi_layout);
      }
#endif

      if(gl.supportxvgamma1) {

         gui.vi_check_gamma = gtk_check_button_new_with_label("Gamma (CRT1 only)");
         gtk_table_attach_defaults(GTK_TABLE(gui.vi_table), gui.vi_check_gamma, 0, 1, gammapos, gammapos+1);
         gtk_widget_show(gui.vi_check_gamma);

         gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.vi_check_gamma,
		 	"Enable gamma correction for the video overlay. This works on first output (CRT1) only.", "");

         gui.vi_ga_table = gtk_table_new(1, 6, FALSE);
         gtk_table_set_row_spacings(GTK_TABLE(gui.vi_ga_table), 1);
         gtk_table_set_col_spacings(GTK_TABLE(gui.vi_ga_table), 1);
         gtk_container_set_border_width(GTK_CONTAINER(gui.vi_ga_table), 1);
         gtk_table_attach_defaults(GTK_TABLE(gui.vi_table), gui.vi_ga_table, 1, 3, gammapos, gammapos+1);
         gtk_widget_show(gui.vi_ga_table);

         gui.vi_label_red = gtk_label_new("R");
         gtk_table_attach_defaults(GTK_TABLE(gui.vi_ga_table), gui.vi_label_red, 0, 1, 0, 1);
         gtk_widget_show(gui.vi_label_red);
         gui.vi_label_green = gtk_label_new("G");
         gtk_table_attach_defaults(GTK_TABLE(gui.vi_ga_table), gui.vi_label_green, 2, 3, 0, 1);
         gtk_widget_show(gui.vi_label_green);
         gui.vi_label_blue = gtk_label_new("B");
         gtk_table_attach_defaults(GTK_TABLE(gui.vi_ga_table), gui.vi_label_blue, 4, 5, 0, 1);
         gtk_widget_show(gui.vi_label_blue);

         gui.vi_spin1_adj = (GtkAdjustment *)gtk_adjustment_new(1.0, 0.1, 10.0, 0.1, 1.0, 0.0);
         gui.vi_spin2_adj = (GtkAdjustment *)gtk_adjustment_new(1.0, 0.1, 10.0, 0.1, 1.0, 0.0);
         gui.vi_spin3_adj = (GtkAdjustment *)gtk_adjustment_new(1.0, 0.1, 10.0, 0.1, 1.0, 0.0);
         gui.vi_spin1 = gtk_spin_button_new(gui.vi_spin1_adj, 0.1, 3);
         gui.vi_spin2 = gtk_spin_button_new(gui.vi_spin2_adj, 0.1, 3);
         gui.vi_spin3 = gtk_spin_button_new(gui.vi_spin3_adj, 0.1, 3);
         gtk_table_attach_defaults(GTK_TABLE(gui.vi_ga_table), gui.vi_spin1, 1, 2, 0, 1);
         gtk_table_attach_defaults(GTK_TABLE(gui.vi_ga_table), gui.vi_spin2, 3, 4, 0, 1);
         gtk_table_attach_defaults(GTK_TABLE(gui.vi_ga_table), gui.vi_spin3, 5, 6, 0, 1);
         sisctrl_vi_set_xvgamma_spin_values();
         gtk_widget_show(gui.vi_spin1);
         gtk_widget_show(gui.vi_spin2);
         gtk_widget_show(gui.vi_spin3);

         gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.vi_spin1,
		 	"Adjust red channel gamma for the video overlay", "");
         gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.vi_spin2,
		 	"Adjust green channel gamma for the video overlay", "");
         gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.vi_spin3,
		 	"Adjust blue channel gamma for the video overlay", "");

         sisctrl_vi_set_xv_gamma_status();

         g_signal_connect(G_OBJECT(gui.vi_check_gamma), "toggled", G_CALLBACK(vi_check_gamma_toggled), NULL);
         g_signal_connect(G_OBJECT(gui.vi_spin1_adj), "value_changed", G_CALLBACK(vi_spin1_changed), NULL);
         g_signal_connect(G_OBJECT(gui.vi_spin2_adj), "value_changed", G_CALLBACK(vi_spin2_changed), NULL);
         g_signal_connect(G_OBJECT(gui.vi_spin3_adj), "value_changed", G_CALLBACK(vi_spin3_changed), NULL);
      }

      gui.vi_check_video_only = gtk_check_button_new_with_label("Video only");
      gtk_table_attach_defaults(GTK_TABLE(gui.vi_table), gui.vi_check_video_only, 0, 1, switchpos, switchpos+1);
      gtk_widget_show(gui.vi_check_video_only);
      g_signal_connect(G_OBJECT(gui.vi_check_video_only), "toggled", G_CALLBACK (vi_check_video_only_toggled), NULL);

      gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.vi_check_video_only,
		"Display the video only, by disabling the graphics left and right of video. This saves bandwidth "
		"and helps on chips with low memory clocks and other hardware issues, such as the SiS760. "
		"This takes effect outside of sisctrl only.", "");

      if(gl.needswitchcrt) {

	 gui.vi_sw_table = gtk_table_new(1, 3, TRUE);
         gtk_table_set_row_spacings(GTK_TABLE(gui.vi_sw_table), 1);
         gtk_table_set_col_spacings(GTK_TABLE(gui.vi_sw_table), 1);
         gtk_container_set_border_width(GTK_CONTAINER(gui.vi_sw_table), 1);
         gtk_table_attach_defaults(GTK_TABLE(gui.vi_table), gui.vi_sw_table, 1, 3, switchpos, switchpos+1);
         gtk_widget_show(gui.vi_sw_table);

         gui.vi_label_sw = gtk_label_new("Video on");
         gtk_widget_show(gui.vi_label_sw);
         gtk_table_attach_defaults(GTK_TABLE(gui.vi_sw_table), gui.vi_label_sw, 0, 1, 0, 1);

         gui.vi_radio_crt1 = gtk_radio_button_new_with_label(NULL, "CRT1");
         gtk_table_attach_defaults(GTK_TABLE(gui.vi_sw_table), gui.vi_radio_crt1, 1, 2, 0, 1);
         gtk_widget_show(gui.vi_radio_crt1);

         gui.vi_group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(gui.vi_radio_crt1));
         gui.vi_radio_crt2 = gtk_radio_button_new_with_label(gui.vi_group, "CRT2");
         gtk_table_attach_defaults(GTK_TABLE(gui.vi_sw_table), gui.vi_radio_crt2, 2, 3, 0, 1);
         gtk_widget_show(gui.vi_radio_crt2);

         gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.vi_radio_crt1,
		 	"Show video overlay on first output (CRT1)", "");
         gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.vi_radio_crt2,
		 	"Show video overlay on second output (CRT2)", "");

         sisctrl_vi_set_switchcrt();

         g_signal_connect(G_OBJECT(gui.vi_radio_crt1), "toggled", G_CALLBACK(vi_radio_changed), NULL);
      }
      break;

    case VIDEO_PAGE_TYPE_NA_NOOVL:
       gui.vi_table = gtk_table_new(3, 3, TRUE);
       gtk_table_set_row_spacings(GTK_TABLE(gui.vi_table), 10);
       gtk_table_set_col_spacings(GTK_TABLE(gui.vi_table), 4);

       gtk_container_set_border_width(GTK_CONTAINER(gui.vi_table), 8);

       gtk_container_add(GTK_CONTAINER(gui.vi_frame), gui.vi_table);
       gtk_widget_show(gui.vi_table);

       gui.vi_label_sw = gtk_label_new("No video overlay available");
       gtk_widget_show(gui.vi_label_sw);
       gtk_table_attach_defaults(GTK_TABLE(gui.vi_table), gui.vi_label_sw, 0, 3, 1, 2);
       break;
    }

    gui.vi_label = xpm_label_box(icon_video, "Video", NULL, 1, "Video overlay (Xv) settings", "");
    gtk_notebook_append_page(GTK_NOTEBOOK(gui.notebook), gui.vi_frame, gui.vi_label);
}

/**************************************************
 *                 "Current" page                 *
 **************************************************/

void
gui_cf_clear_text(void)
{
   gtk_text_buffer_set_text(gui.cf_buffer, "", -1);
}

static GtkTextTag *
gui_cf_convert_int_to_tag(int texttag)
{
   GtkTextTag *tag = NULL;

   switch(texttag) {
      case CU_TT_BLACK:		tag = gui.black; break;
      case CU_TT_BLACK_BOLD:	tag = gui.blackb; break;
      case CU_TT_RED:		tag = gui.red; break;
      case CU_TT_RED_BOLD:	tag = gui.redb; break;
      case CU_TT_BLUE:		tag = gui.blue; break;
      case CU_TT_BLUE_BOLD:	tag = gui.blueb; break;
      default:
      case CU_TT_GREY: 		tag = gui.grey; break;
   }
   return tag;
}

void
gui_cf_insert_option(char *oname, char *ovalue, int texttag)
{
   GtkTextIter myiter, start, end;
   GtkTextTag *tag = gui_cf_convert_int_to_tag(texttag);
   char options[] = "\tOption \"";
   int curlen;

   curlen = gtk_text_buffer_get_char_count(gui.cf_buffer);

   gtk_text_buffer_get_end_iter(gui.cf_buffer, &myiter);
   gtk_text_buffer_insert(gui.cf_buffer, &myiter, options, -1);
   gtk_text_buffer_get_end_iter(gui.cf_buffer, &myiter);
   gtk_text_buffer_insert(gui.cf_buffer, &myiter, oname, -1);
   gtk_text_buffer_get_end_iter(gui.cf_buffer, &myiter);
   gtk_text_buffer_insert(gui.cf_buffer, &myiter, "\" \"", -1);
   gtk_text_buffer_get_end_iter(gui.cf_buffer, &myiter);
   gtk_text_buffer_insert(gui.cf_buffer, &myiter, ovalue, -1);
   gtk_text_buffer_get_end_iter(gui.cf_buffer, &myiter);
   gtk_text_buffer_insert(gui.cf_buffer, &myiter, "\"\n", -1);

   gtk_text_buffer_get_iter_at_offset(gui.cf_buffer, &start, curlen);
   gtk_text_buffer_get_end_iter(gui.cf_buffer, &end);
   if(tag) gtk_text_buffer_apply_tag(gui.cf_buffer, tag, &start, &end);
}

void
gui_cf_append_text(char *mytext, int texttag)
{
   GtkTextIter myiter, start, end;
   GtkTextTag *tag = gui_cf_convert_int_to_tag(texttag);;
   int curlen;

   curlen = gtk_text_buffer_get_char_count(gui.cf_buffer);
   gtk_text_buffer_get_end_iter(gui.cf_buffer, &myiter);
   gtk_text_buffer_insert(gui.cf_buffer, &myiter, mytext, -1);
   gtk_text_buffer_get_iter_at_offset(gui.cf_buffer, &start, curlen);
   gtk_text_buffer_get_end_iter(gui.cf_buffer, &end);
   if(tag) gtk_text_buffer_apply_tag(gui.cf_buffer, tag, &start, &end);
}

/* Signal handlers */

static void
gui_find_hint_type(void)
{
   if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(gui.cf_radio1))) {
      gui.showhints = CF_HINTS_GE;
   } else if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(gui.cf_radio2))) {
      gui.showhints = CF_HINTS_MFB;
   } else {
      gui.showhints = CF_HINTS_DHM;
   }
}

static void
cf_check_toggled(GtkWidget *widget, gpointer data)
{
   if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget))) {
      gtk_widget_set_sensitive(gui.cf_radio1, TRUE);
      if( (gl.dualhead) || (gl.mergedfb) || (!gl.havevb) ) {
         gtk_widget_set_sensitive(gui.cf_radio2, FALSE);
         gtk_widget_set_sensitive(gui.cf_radio3, FALSE);
         gui.showhints = CF_HINTS_GE;
      } else {
         gtk_widget_set_sensitive(gui.cf_radio2, TRUE);
         gtk_widget_set_sensitive(gui.cf_radio3, TRUE);
         gui_find_hint_type();
      }
   } else {
      gui.showhints = CF_HINTS_NO;
      gtk_widget_set_sensitive(gui.cf_radio1, FALSE);
      gtk_widget_set_sensitive(gui.cf_radio2, FALSE);
      gtk_widget_set_sensitive(gui.cf_radio3, FALSE);
   }
   sisctrl_gl_current_page_selected(gui.showhints);
}

static void
cf_radio_changed(GtkWidget *widget, gpointer data)
{
   if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget))) {
      gui_find_hint_type();
      sisctrl_gl_current_page_selected(gui.showhints);
   }
}

/* Build page */

static void
build_current_page(void)
{
   int myspace = 6;

   if(gl.config_page_type != CONFIG_PAGE_TYPE_NO_HINTS) myspace = 0;

   gui.cf_frame = gtk_frame_new("Current configuration");
   gtk_container_set_border_width(GTK_CONTAINER(gui.cf_frame), 8);
   gtk_widget_show(gui.cf_frame);

   /* Text window, text buffer */
   gui.cf_text_window = gtk_scrolled_window_new(NULL, NULL);
   gtk_container_set_border_width(GTK_CONTAINER(gui.cf_text_window), myspace);
   gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(gui.cf_text_window), GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);

   gui.cf_tview = gtk_text_view_new();
   gui.cf_buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(gui.cf_tview));

   gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(gui.cf_tview), GTK_WRAP_WORD);
   gtk_text_view_set_editable(GTK_TEXT_VIEW(gui.cf_tview), FALSE);
   gtk_text_buffer_set_text(gui.cf_buffer, "", -1);

   gtk_container_add(GTK_CONTAINER(gui.cf_text_window), gui.cf_tview);

   gtk_widget_show(gui.cf_tview);
   gtk_widget_show(gui.cf_text_window);

   if(gl.config_page_type == CONFIG_PAGE_TYPE_NO_HINTS) {

      gtk_container_add(GTK_CONTAINER(gui.cf_frame), gui.cf_text_window);

   } else {

      gui.cf_table = gtk_table_new(6, 4, FALSE);
      gtk_table_set_row_spacings(GTK_TABLE(gui.cf_table), 0);
      gtk_table_set_col_spacings(GTK_TABLE(gui.cf_table), 0);
      gtk_container_set_border_width(GTK_CONTAINER(gui.cf_table), 8);
      gtk_widget_show(gui.cf_table);

      gtk_table_attach_defaults(GTK_TABLE(gui.cf_table), gui.cf_text_window, 0, 4, 0, 5);

      gui.cf_check = gtk_check_button_new_with_label("Show hints");
      gtk_table_attach_defaults(GTK_TABLE(gui.cf_table), gui.cf_check, 0, 1, 5, 6);
      gtk_widget_show(gui.cf_check);

      gui.cf_radio1 = gtk_radio_button_new_with_label(NULL, "Generic");
      gtk_table_attach_defaults(GTK_TABLE(gui.cf_table), gui.cf_radio1, 1, 2, 5, 6);
      gtk_widget_show(gui.cf_radio1);

      /* gui.cf_group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(gui.cf_radio1)); */

      gui.cf_radio2 = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(gui.cf_radio1), "MergedFB");
      gtk_table_attach_defaults(GTK_TABLE(gui.cf_table), gui.cf_radio2, 2, 3, 5, 6);
      gtk_widget_show(gui.cf_radio2);

      gui.cf_radio3 = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(gui.cf_radio1), "Dual Head");
      gtk_table_attach_defaults(GTK_TABLE(gui.cf_table), gui.cf_radio3, 3, 4, 5, 6);
      gtk_widget_show(gui.cf_radio3);

      gui.showhints = CF_HINTS_NO;
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(gui.cf_check), FALSE);

      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(gui.cf_radio1), TRUE);

      gtk_widget_set_sensitive(gui.cf_radio1, FALSE);
      gtk_widget_set_sensitive(gui.cf_radio2, FALSE);
      gtk_widget_set_sensitive(gui.cf_radio3, FALSE);

      g_signal_connect(G_OBJECT(gui.cf_check), "toggled", G_CALLBACK(cf_check_toggled), NULL);

      g_signal_connect(G_OBJECT(gui.cf_radio1), "toggled", G_CALLBACK(cf_radio_changed), NULL);
      g_signal_connect(G_OBJECT(gui.cf_radio2), "toggled", G_CALLBACK(cf_radio_changed), NULL);
      g_signal_connect(G_OBJECT(gui.cf_radio3), "toggled", G_CALLBACK(cf_radio_changed), NULL);

      gtk_container_add(GTK_CONTAINER(gui.cf_frame), gui.cf_table);
   }

   gui.black = gtk_text_buffer_create_tag(gui.cf_buffer, "black_foreground", "foreground", "black", NULL);
   gui.grey = gtk_text_buffer_create_tag(gui.cf_buffer, "grey_foreground", "foreground", "grey32", NULL);
   gui.red = gtk_text_buffer_create_tag(gui.cf_buffer, "red_foreground", "foreground", "red", NULL);
   gui.blackb = gtk_text_buffer_create_tag(gui.cf_buffer, "black_foreground_bold", "foreground", "black", "weight", "bold", NULL);
   gui.redb = gtk_text_buffer_create_tag(gui.cf_buffer, "red_foreground_bold", "foreground", "red", "weight", "bold", NULL);
   gui.blue = gtk_text_buffer_create_tag(gui.cf_buffer, "blue_foreground", "foreground", "blue", NULL);
   gui.blueb = gtk_text_buffer_create_tag(gui.cf_buffer, "blue_foreground_bold", "foreground", "blue", "weight", "bold", NULL);

   gui.cf_label = xpm_label_box(icon_config, "Current", NULL, 1, "Show current configuration", "");
   gtk_notebook_append_page(GTK_NOTEBOOK(gui.notebook), gui.cf_frame, gui.cf_label);
}

/**************************************************
 *                   Info page                    *
 **************************************************/

void
gui_ab_clear_text(void)
{
   gtk_text_buffer_set_text(gui.ab_buffer, "", -1);
}

void
gui_ab_append_text(char *mytext, int texttag)
{
   GtkTextIter myiter, start, end;
   GtkTextTag *tag = NULL;
   int curlen;

   switch(texttag) {
   case AB_TT_BLACK:
      tag = gui.abblack;
      break;
   default:
   case AB_TT_BLACK_BOLD:
      tag = gui.abblackb;
      break;
   }

   curlen = gtk_text_buffer_get_char_count(gui.ab_buffer);
   gtk_text_buffer_get_end_iter(gui.ab_buffer, &myiter);
   gtk_text_buffer_insert(gui.ab_buffer, &myiter, mytext, -1);
   gtk_text_buffer_get_iter_at_offset(gui.ab_buffer, &start, curlen);
   gtk_text_buffer_get_end_iter(gui.ab_buffer, &end);
   if(tag) gtk_text_buffer_apply_tag(gui.ab_buffer, tag, &start, &end);
}

static gboolean
ab_layout_expose(GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
   int x, y, w, h, i, j, wmax, lastpos, hstepwidth, vstepwidth;
   int dynamicx = 0;
   gint lwidth = widget->allocation.width;
   gint lheight = widget->allocation.height;
   PangoLayout *playout;
   GtkAllocation alloc;
   GtkWidget *labels[] = { gui.ab_label_ab1,
			   gui.ab_label_ab2,
			   gui.ab_label_ab4 };

#define BW 2   /* border width */
#define BWT 8  /* border width for text area */

   if(!gui.ab_context) {
      gui.ab_context = gdk_gc_new(GTK_LAYOUT(gui.ab_layout)->bin_window);
   }

   gdk_gc_set_rgb_fg_color(gui.ab_context, &gui.ab_bgcolorb);
   gdk_draw_line(GTK_LAYOUT(widget)->bin_window,
		 gui.ab_context,
		 BW-1, BW-1, BW - 1 + lwidth - (BW * 2), BW - 1);
   gdk_draw_line(GTK_LAYOUT(widget)->bin_window,
		 gui.ab_context,
		 BW-1, BW-1, BW-1, BW - 1 + lheight - (BW * 2));

   gdk_gc_set_rgb_fg_color(gui.ab_context, &gui.ab_bgcolorw);
   gdk_draw_line(GTK_LAYOUT(widget)->bin_window,
		 gui.ab_context,
		 BW - 1,
		 BW - 1 + lheight - (BW * 2) + 1,
		 BW - 1 + lwidth - (BW * 2),
		 BW - 1 + lheight - (BW * 2) + 1);
   gdk_draw_line(GTK_LAYOUT(widget)->bin_window,
		 gui.ab_context,
		 BW - 1 + lwidth - (BW * 2) + 1,
		 BW - 1,
		 BW - 1 + lwidth - (BW * 2) + 1,
		 BW - 1 + lheight - (BW * 2) + 1);

   gdk_gc_set_rgb_fg_color(gui.ab_context, &gui.ab_bgcolor);
   gdk_draw_rectangle(GTK_LAYOUT(widget)->bin_window,
		      gui.ab_context,
		      TRUE,
		      BW, BW,
		      lwidth - (BW * 2), lheight - (BW * 2));

   gdk_draw_drawable(GTK_LAYOUT(widget)->bin_window,
		     gui.ab_context,
		     gui.ab_iconul,
		     0, 0,
		     BW, BW,
		     180, 130);

   hstepwidth = ((lwidth - (BWT * 2)) / 2);
   vstepwidth = ((lheight - 8 - (BWT * 2)) / 4);

   /* x = hstepwidth - (81 / 2) + BWT; */
   x = 180 + BWT;
   if((x + 81) > lwidth) x = hstepwidth - (81 / 2) + BWT;
   y = ((vstepwidth - (23 / 2)) / 2) + BWT;
   gdk_draw_drawable(GTK_LAYOUT(widget)->bin_window,
		     widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
		     gui.ab_icon,
		     0, 0, x, y, 81, 23);

   wmax = 0;
   for(i = 0; i < 3; i++) {
      playout = gtk_label_get_layout(GTK_LABEL(labels[i]));
      pango_layout_get_pixel_size(playout, &w, &h);
      if(wmax < w) wmax = w;
   }

   x = 180 + 20 + BWT;
   if((x + wmax) > lwidth) x = lwidth - wmax;
   if(x < 0) dynamicx = 1;

   for(i = 0; i < 3; i++) {
      playout = gtk_label_get_layout(GTK_LABEL(labels[i]));
      pango_layout_get_pixel_size(playout, &w, &h);
      lastpos = 8 + (vstepwidth * (i + 1)) + ((vstepwidth - (h / 2)) / 2);
      if(dynamicx) x = hstepwidth - (w / 2) + BWT;
      gdk_draw_layout(GTK_LAYOUT(widget)->bin_window,
                   labels[i]->style->fg_gc[GTK_WIDGET_STATE(labels[i])],
		   x, lastpos + BWT, playout);
   }

   return TRUE;
}

static void
build_info_page(void)
{
   gui.ab_bgcolor.red = 193 << 8;
   gui.ab_bgcolor.green = 193 << 8;
   gui.ab_bgcolor.blue = 217 << 8;

   gui.ab_bgcolorw.red = 65535;
   gui.ab_bgcolorw.green = 65535;
   gui.ab_bgcolorw.blue = 65535;

   gui.ab_bgcolorb.red = 0;
   gui.ab_bgcolorb.green = 0;
   gui.ab_bgcolorb.blue = 0;

   gui.ab_frame = gtk_frame_new("Info");
   gtk_container_set_border_width(GTK_CONTAINER(gui.ab_frame), 8);
   gtk_widget_show(gui.ab_frame);

   gui.ab_table = gtk_table_new(3, 1, FALSE);
   gtk_table_set_row_spacings(GTK_TABLE(gui.ab_table), 1);
   gtk_table_set_col_spacings(GTK_TABLE(gui.ab_table), 0);
   gtk_container_set_border_width(GTK_CONTAINER(gui.ab_table), 1);

   gui.ab_layout = gtk_layout_new(NULL, NULL);
   gtk_widget_set_size_request(gui.ab_layout, 0, 138);
   gtk_table_attach_defaults(GTK_TABLE(gui.ab_table), gui.ab_layout, 0, 1, 0, 2);
   gtk_widget_show(gui.ab_layout);

   gui.ab_iconul = gdk_pixmap_create_from_xpm_d(gui.window->window, &gui.ab_maskul, NULL, sisctrl_logo_xpm);
   gui.ab_icon = gdk_pixmap_create_from_xpm_d(gui.window->window, &gui.ab_mask, NULL, sisctrl2_xpm);

   /* So we get a fixed size - does not work. If table is inhomog., our size is entirely disregarded. */
#if 0
   gui.ab_image = gtk_image_new_from_pixmap(gui.ab_iconul, gui.ab_maskul);
   gtk_container_add(GTK_CONTAINER(gui.ab_layout), gui.ab_image);
   /* gtk_layout_put(GTK_LAYOUT(gui.ab_layout), gui.ab_image, 0, 0); */
   gtk_widget_show(gui.ab_image);
#endif

   gui.ab_label_ab1 = gtk_label_new("SiS Display Control Panel");
   gui.ab_label_ab2 = gtk_label_new("(C) 2003-2005 Thomas Winischhofer");
   gui.ab_label_ab4 = gtk_label_new("Licensed under the terms of the GPLv2");
   gtk_layout_put(GTK_LAYOUT(gui.ab_layout), gui.ab_label_ab1, 0, 0);
   gtk_layout_put(GTK_LAYOUT(gui.ab_layout), gui.ab_label_ab2, 0, 10);
   gtk_layout_put(GTK_LAYOUT(gui.ab_layout), gui.ab_label_ab4, 0, 20);
   gtk_widget_show(gui.ab_label_ab1);
   gtk_widget_show(gui.ab_label_ab2);
   gtk_widget_show(gui.ab_label_ab4);

   gui.ab_text_window = gtk_scrolled_window_new(NULL, NULL);
   gtk_container_set_border_width(GTK_CONTAINER(gui.ab_text_window), 0);
   gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(gui.ab_text_window), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

   gui.ab_tview = gtk_text_view_new();
   gui.ab_buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(gui.ab_tview));

   gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(gui.ab_tview), GTK_WRAP_WORD);
   gtk_text_view_set_editable(GTK_TEXT_VIEW(gui.ab_tview), FALSE);
   gtk_text_buffer_set_text(gui.ab_buffer, "", -1);

   gtk_container_add(GTK_CONTAINER(gui.ab_text_window), gui.ab_tview);
   gtk_table_attach_defaults(GTK_TABLE(gui.ab_table), gui.ab_text_window, 0, 1, 2, 3);

   gtk_widget_show(gui.ab_tview);
   gtk_widget_show(gui.ab_text_window);

   gui.abblack = gtk_text_buffer_create_tag(gui.ab_buffer, "black_foreground", "foreground", "black", NULL);
   gui.abblackb = gtk_text_buffer_create_tag(gui.ab_buffer, "black_bold", "foreground", "black",
							"weight", PANGO_WEIGHT_BOLD, NULL);

   sisctrl_ab_generate_current();

   gtk_container_add(GTK_CONTAINER(gui.ab_frame), gui.ab_table);

   gtk_widget_show(gui.ab_table);

   g_signal_connect(G_OBJECT(gui.ab_layout), "expose_event", G_CALLBACK(ab_layout_expose), NULL);

   gui.ab_label = xpm_label_box(icon_about, "Info", NULL, 1, "About SiSCtrl", "");
   gtk_notebook_append_page(GTK_NOTEBOOK(gui.notebook), gui.ab_frame, gui.ab_label);
}


/**************************************************
 *                  Global stuff                  *
 **************************************************/

#ifdef HAVE_RANDR
static void
xrandreventhandler(GdkScreen *screen, gpointer data)
{
   sisctrl_xrandreventhandler();
}
#endif

static gint
delete_event(GtkWidget *widget, GdkEvent *event, gpointer data)
{
    sisctrl_delete_event();
    return FALSE;
}

/* destroy callback */
static void
destroy(GtkWidget *widget, gpointer data)
{
    sisctrl_destroy_event();
    gtk_main_quit();
}

static void
apply_clicked(GtkWidget *widget, gpointer data)
{
   int nb_page = gtk_notebook_get_current_page(GTK_NOTEBOOK(gui.notebook));

   sisctrl_do_apply(nb_page);
}

static void
ok_clicked(GtkWidget *widget, gpointer data)
{
   int nb_page = gtk_notebook_get_current_page(GTK_NOTEBOOK(gui.notebook));

   sisctrl_ok_clicked(nb_page);
}

static void
close_clicked(GtkWidget *widget, gpointer data)
{
   sisctrl_close_clicked();
}

static void
revert_clicked(GtkWidget *widget, gpointer data)
{
   int nb_page = gtk_notebook_get_current_page(GTK_NOTEBOOK(gui.notebook));

   sisctrl_revert_clicked(nb_page);
}

/* Helper for enabling/disabling "Revert" button */

void
gui_enable_revert_button(int enable)
{
   gtk_widget_set_sensitive(gui.button4, enable ? TRUE : FALSE);
}

/* Notebook signal handler */

static gboolean
notebook_changed(GtkNotebook *nb, GtkNotebookPage *page, guint page_num, gpointer data)
{
   int i;
   gint temp;

   gui.notebook_page = page_num;

#ifdef USEXV
   if(!gl.noxvdemo) {
      if(page_num == 5) {
         showxvimage();
      } else {
         hidexvimage();
      }
   }
#endif

   if(page_num > 2)  gtk_widget_set_sensitive(gui.button2, FALSE);
   else              gtk_widget_set_sensitive(gui.button2, TRUE);

   switch(page_num) {
   case 0:
      sisctrl_gl_display_mode_page_selected();
      gui_enable_revert_button(0);
#ifdef HAVE_RANDR
      if(gl.do_randr) {
	 temp = gtk_option_menu_get_history(GTK_OPTION_MENU(gui.dm_omenu));
	 if(temp >= 2) {
	    gtk_option_menu_set_history(GTK_OPTION_MENU(gui.dm_omenu), 0);
	 }
      }
#endif
      break;
   case 1:
      sisctrl_gl_crt1_page_selected();
      gui_enable_revert_button(0);
      break;
   case 2:
      sisctrl_gl_crt2_page_selected();
      gui_enable_revert_button(0);
      break;
   case 3:
      sisctrl_gl_gamma_page_selected();
      gui_enable_revert_button(1);
      break;
   case 4:
      sisctrl_gl_tv_page_selected();
      break;
   case 5:
      sisctrl_gl_video_page_selected();
      break;
   case 6:
      sisctrl_gl_current_page_selected(gui.showhints);
      gui_enable_revert_button(0);
      break;
   default:
      gui_enable_revert_button(0);
   }

   return FALSE;
}


void
gui_init(int *argc, char ***argv)
{
   /* Clear gui data */
   memset(&gui, 0, sizeof(gui));

   /* Init gtk */
   gtk_init(argc, argv);
}

void
gui_build_gui(void)
{
    int i, j, k;

    /* create a new window */
    gui.window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_widget_realize(gui.window); /* realize here for gdk calls */

    /* set window title */
    gtk_window_set_title(GTK_WINDOW(gui.window), gl.windowtitle);

    /* set the border width of the window. */
    gtk_container_set_border_width(GTK_CONTAINER(gui.window), 8);

    g_signal_connect(G_OBJECT(gui.window), "delete_event", G_CALLBACK(delete_event), NULL);

    gui.table = gtk_table_new(2, 4, FALSE);
    gtk_table_set_row_spacings(GTK_TABLE(gui.table), 10);
    gtk_table_set_col_spacings(GTK_TABLE(gui.table), 10);

    gtk_container_add(GTK_CONTAINER(gui.window), gui.table);

    g_signal_connect(G_OBJECT(gui.window), "destroy", G_CALLBACK(destroy), NULL);

    if(gl.do_randr) {
#ifdef HAVE_RANDR
       GdkScreen *screen = gtk_window_get_screen(GTK_WINDOW(gui.window));
       g_signal_connect(screen, "size-changed", G_CALLBACK(xrandreventhandler), NULL);
#endif
    }

    /* Create tooltips */
    gui.mytooltips = gtk_tooltips_new();

    /* Create and attach notebook */
    gui.notebook = gtk_notebook_new();
    gtk_notebook_set_tab_pos(GTK_NOTEBOOK(gui.notebook), GTK_POS_TOP);
    gtk_table_attach_defaults(GTK_TABLE(gui.table), gui.notebook, 0, 4, 0, 1);
    gtk_notebook_set_scrollable(GTK_NOTEBOOK(gui.notebook), FALSE);

    gtk_widget_show(gui.notebook);

    if(gl.video_page_type != VIDEO_PAGE_TYPE_NORMAL) gl.noxvdemo = TRUE;

    /* test if we can make a child window */
    if(!gl.noxvdemo) {
       /* Need to flush here so that we have a valid window for testing */
       gdk_flush();
       if(!testxvwindow()) gl.noxvdemo = TRUE;
    }

    /* Create and attach notebook pages */

    build_display_mode_page();

    build_crt1_page();

    build_crt2_page();

    build_gamma_page();

    build_tv_page();

    build_video_page();

    build_current_page();

    build_info_page();

    /* Set current page of notebook */
    gtk_notebook_set_current_page(GTK_NOTEBOOK(gui.notebook), 0);

    g_signal_connect(G_OBJECT(gui.notebook), "switch-page", G_CALLBACK(notebook_changed), NULL);

    /* Create and attach buttons */
    gui.button1 = gtk_button_new();
    gui.button2 = gtk_button_new();
    gui.button3 = gtk_button_new();
    gui.button4 = gtk_button_new();

    gui.button1box = xpm_label_box(icon_ok, "_OK", gui.button1, 3, "", "");
    gui.button2box = xpm_label_box(icon_apply, "_Apply", gui.button2, 3, "", "");
    gui.button3box = xpm_label_box(icon_close, "_Close", gui.button3, 3, "", "");
    gui.button4box = xpm_label_box(icon_undo, "_Reset", gui.button4, 3, "", "");

    gtk_widget_show(gui.button1box);
    gtk_widget_show(gui.button2box);
    gtk_widget_show(gui.button3box);
    gtk_widget_show(gui.button4box);

    gtk_widget_set_sensitive(gui.button4, FALSE);

    gtk_container_add(GTK_CONTAINER(gui.button1), gui.button1box);
    gtk_container_add(GTK_CONTAINER(gui.button2), gui.button2box);
    gtk_container_add(GTK_CONTAINER(gui.button3), gui.button3box);
    gtk_container_add(GTK_CONTAINER(gui.button4), gui.button4box);

    g_signal_connect(G_OBJECT (gui.button1), "clicked", G_CALLBACK (ok_clicked), NULL);
    g_signal_connect(G_OBJECT (gui.button2), "clicked", G_CALLBACK (apply_clicked), NULL);
    g_signal_connect(G_OBJECT (gui.button3), "clicked", G_CALLBACK (close_clicked), NULL);
    g_signal_connect(G_OBJECT (gui.button4), "clicked", G_CALLBACK (revert_clicked), NULL);

    g_signal_connect_swapped(G_OBJECT(gui.button3), "clicked", G_CALLBACK (gtk_widget_destroy), gui.window);

    gtk_table_attach(GTK_TABLE(gui.table), gui.button1, 0, 1, 1, 2, GTK_FILL|GTK_EXPAND, 0, 0, 0);
    gtk_table_attach(GTK_TABLE(gui.table), gui.button2, 1, 2, 1, 2, GTK_FILL|GTK_EXPAND, 0, 0, 0);
    gtk_table_attach(GTK_TABLE(gui.table), gui.button3, 2, 3, 1, 2, GTK_FILL|GTK_EXPAND, 0, 0, 0);
    gtk_table_attach(GTK_TABLE(gui.table), gui.button4, 3, 4, 1, 2, GTK_FILL|GTK_EXPAND, 0, 0, 0);
    /*
    gtk_table_attach_defaults(GTK_TABLE(gui.table), gui.button1, 0, 1, 1, 2);
    gtk_table_attach_defaults(GTK_TABLE(gui.table), gui.button2, 1, 2, 1, 2);
    gtk_table_attach_defaults(GTK_TABLE(gui.table), gui.button3, 2, 3, 1, 2);
    gtk_table_attach_defaults(GTK_TABLE(gui.table), gui.button4, 3, 4, 1, 2);
    */

    gtk_widget_show(gui.button1);
    gtk_widget_show(gui.button2);
    gtk_widget_show(gui.button3);
    gtk_widget_show(gui.button4);

    gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.button1, "Apply changes and exit", "Apply the changes and exit SiSCtrl");
    gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.button2, "Apply changes", "Apply changes, but do not exit SiSCtrl");
    gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.button3, "Exit SiSCtrl", "Exit SiSCtrl without any changes");
    gtk_tooltips_set_tip(GTK_TOOLTIPS(gui.mytooltips), gui.button4, "Reset to previous settings", "Reset the controls to their previous setting");

    /* Show the table */
    gtk_widget_show(gui.table);

    /* Set window icons */
    {
       GdkPixbuf *icon16_pixbuf, *icon32_pixbuf, *icon48_pixbuf, *icon64_pixbuf;
       GList *icon_list = NULL;

       icon16_pixbuf = gdk_pixbuf_new_from_xpm_data((const char **)sisctrl_icon16);
       icon32_pixbuf = gdk_pixbuf_new_from_xpm_data((const char **)sisctrl_icon32);
       icon48_pixbuf = gdk_pixbuf_new_from_xpm_data((const char **)sisctrl_icon48);
       icon64_pixbuf = gdk_pixbuf_new_from_xpm_data((const char **)sisctrl_icon64);

       icon_list = g_list_append(icon_list, icon16_pixbuf);
       icon_list = g_list_append(icon_list, icon32_pixbuf);
       icon_list = g_list_append(icon_list, icon48_pixbuf);
       icon_list = g_list_append(icon_list, icon64_pixbuf);

       gtk_window_set_icon_list((GtkWindow *)gui.window, icon_list);
       gtk_window_set_default_icon_list(icon_list);

       g_list_free(icon_list);
    }

    /* Show the window */
    gtk_widget_show(gui.window);

#ifdef USEXV
    if(!gl.noxvdemo) {
       gdk_flush();
       if(makexvwindow()) {
          g_signal_connect(G_OBJECT(gui.vi_layout), "configure_event", G_CALLBACK(vi_layout_configure), NULL);
          g_signal_connect(G_OBJECT(gui.vi_layout), "expose_event", G_CALLBACK(vi_layout_expose), NULL);
          g_signal_connect(G_OBJECT(gui.window), "configure_event", G_CALLBACK(window_configure), NULL);
          g_signal_connect(G_OBJECT(gui.window), "expose_event", G_CALLBACK(window_expose), NULL);
          g_signal_connect(G_OBJECT(gui.window), "visibility_notify_event", G_CALLBACK(window_visibility), NULL);
          gtk_widget_add_events(gui.window, gtk_widget_get_events(gui.window) | GDK_VISIBILITY_NOTIFY_MASK);
          gui.xvinit = TRUE;
          /* Now show it once, otherwise - strangely- at the first time,
           * Xv will be re-initialized on every XVShmPutImage; seems to
           * relate to gtk internals
           */
          resizexvwindow();
          showxvimage();
          hidexvimage();
       } else {
          gl.noxvdemo = TRUE;
       }
    }
#endif

}

void
gui_main(void)
{
   gtk_main();
}







